@@ -2,7 +2,7 @@
 *                  Exim Monitor                  *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 
@@ -14,7 +14,7 @@ static Widget menushell;
 static Widget queue_text_sink;
 static Widget dialog_shell, dialog_widget;
 
-static Widget text_create(char *, int);
+static Widget text_create(uschar *, int);
 
 static int highlighted_start, highlighted_end, highlighted_x, highlighted_y;
 
@@ -105,9 +105,9 @@ static Arg item_13_arg[] = {
   { XtNfromVert,  (XtArgVal) NULL },        /* must be first */
   { "label",      (XtArgVal)" Edit sender" } };
 
-static Arg item_14_arg[] = {
+static Arg item_99_arg[] = {
   { XtNfromVert,  (XtArgVal) NULL },        /* must be first */
-  { "label",      (XtArgVal)" Edit body" } };
+  { "label",      (XtArgVal)" " } };
 
 
 
@@ -136,8 +136,8 @@ menu_is_up = FALSE;
 static void msglogAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 int i;
-char buffer[256];
-Widget text = text_create((char *)client_data, text_depth);
+uschar buffer[256];
+Widget text = text_create((uschar *)client_data, text_depth);
 FILE *f = NULL;
 
 w = w;      /* Keep picky compilers happy */
@@ -147,10 +147,10 @@ call_data = call_data;
 
 for (i = 0; i < (spool_is_split? 2:1); i++)
   {
-  message_subdir[0] = (i != 0)? ((char *)client_data)[5] : 0;
-  sprintf(buffer, "%s/msglog/%s/%s", spool_directory, message_subdir,
-    (char *)client_data);
-  f = fopen(buffer, "r");
+  message_subdir[0] = (i != 0)? ((uschar *)client_data)[5] : 0;
+  sprintf(CS buffer, "%s/msglog/%s/%s", spool_directory, message_subdir,
+    (uschar *)client_data);
+  f = fopen(CS buffer, "r");
   if (f != NULL) break;
   }
 
@@ -158,7 +158,7 @@ if (f == NULL)
   text_showf(text, "%s: %s\n", buffer, strerror(errno));
 else
   {
-  while (fgets(buffer, 256, f) != NULL) text_show(text, buffer);
+  while (Ufgets(buffer, 256, f) != NULL) text_show(text, buffer);
   fclose(f);
   }
 }
@@ -172,8 +172,8 @@ else
 static void bodyAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 int i;
-char buffer[256];
-Widget text = text_create((char *)client_data, text_depth);
+uschar buffer[256];
+Widget text = text_create((uschar *)client_data, text_depth);
 FILE *f = NULL;
 
 w = w;      /* Keep picky compilers happy */
@@ -181,10 +181,10 @@ call_data = call_data;
 
 for (i = 0; i < (spool_is_split? 2:1); i++)
   {
-  message_subdir[0] = (i != 0)? ((char *)client_data)[5] : 0;
-  sprintf(buffer, "%s/input/%s/%s-D", spool_directory, message_subdir,
-    (char *)client_data);
-  f = fopen(buffer, "r");
+  message_subdir[0] = (i != 0)? ((uschar *)client_data)[5] : 0;
+  sprintf(CS buffer, "%s/input/%s/%s-D", spool_directory, message_subdir,
+    (uschar *)client_data);
+  f = fopen(CS buffer, "r");
   if (f != NULL) break;
   }
 
@@ -193,13 +193,13 @@ if (f == NULL)
 else
   {
   int count = 0;
-  while (fgets(buffer, 256, f) != NULL)
+  while (Ufgets(buffer, 256, f) != NULL)
     {
     text_show(text, buffer);
-    count += (int)strlen(buffer);
+    count += Ustrlen(buffer);
     if (count > body_max)
       {
-      text_show(text, "\n*** Message length exceeds BODY_MAX ***\n");
+      text_show(text, US"\n*** Message length exceeds BODY_MAX ***\n");
       break;
       }
     }
@@ -218,56 +218,36 @@ unless action_output is set. We can't, however, tell until we have run
 the command whether we want the output or not, so the pipe has to be set up in
 all cases. */
 
-static void ActOnMessage(char *id, char *action, char *address_arg)
+static void ActOnMessage(uschar *id, uschar *action, uschar *address_arg)
 {
 int pid;
 int pipe_fd[2];
-int delivery = strcmp(action + (int)strlen(action) - 2, "-M") == 0;
-int editing = strcmp(action + (int)strlen(action) - 4, "-Meb") == 0;
-char *quote = "";
-char *at = "";
-char *qualify = "";
-char buffer[256];
+int delivery = Ustrcmp(action + Ustrlen(action) - 2, "-M") == 0;
+uschar *quote = US"";
+uschar *at = US"";
+uschar *qualify = US"";
+uschar buffer[256];
 queue_item *qq;
 Widget text = NULL;
 
-
-/* Editing is handled in a subprocess whose output is not messed about
-in any way. Don't wait for it - the SIGCHLD handler will clear it away
-when it is finished. */
-
-if (editing)
-  {
-  if ((pid = fork()) == 0)
-    {
-    sprintf(buffer, "xterm -n %s -title %s -e %s %s %s %s %s", id, id,
-      exim_path,
-      (alternate_config == NULL)? "" : "-C",
-      (alternate_config == NULL)? "" : alternate_config,
-      action, id);
-    system(buffer);
-    _exit(0);
-    }
-  return;
-  }
-
-
-/* The command for everything else is the same; if the address arg is not empty
-and does not contain @ and there is a qualify domain, qualify it. */
+/* If the address arg is not empty and does not contain @ and there is a
+qualify domain, qualify it. (But don't qualify '<>'.)*/
 
 if (address_arg[0] != 0)
   {
-  quote = "\'";
-  if (strchr(address_arg, '@') == NULL && qualify_domain != NULL &&
+  quote = US"\'";
+  if (Ustrchr(address_arg, '@') == NULL &&
+      Ustrcmp(address_arg, "<>") != 0 &&
+      qualify_domain != NULL &&
       qualify_domain[0] != 0)
     {
-    at = "@";
+    at = US"@";
     qualify = qualify_domain;
     }
   }
-sprintf(buffer, "%s %s %s %s %s %s%s%s%s%s", exim_path,
-  (alternate_config == NULL)? "" : "-C",
-  (alternate_config == NULL)? "" : alternate_config,
+sprintf(CS buffer, "%s %s %s %s %s %s%s%s%s%s", exim_path,
+  (alternate_config == NULL)? US"" : US"-C",
+  (alternate_config == NULL)? US"" : alternate_config,
   action, id, quote, address_arg, at, qualify, quote);
 
 /* If we know we are going to need the window, create it now. */
@@ -289,7 +269,7 @@ if (pipe(pipe_fd) != 0)
     text = text_create(id, text_depth);
     text_showf(text, "%s\n", buffer);
     }
-  text_show(text, "*** Failed to create pipe ***\n");
+  text_show(text, US"*** Failed to create pipe ***\n");
   return;
   }
 
@@ -315,7 +295,7 @@ if (!delivery)
   dup2(pipe_fd[1], 2);
   close(pipe_fd[1]);
 
-  rc = system(buffer);
+  rc = system(CS buffer);
 
   close(1);
   close(2);
@@ -344,14 +324,14 @@ if (!delivery)
   /* If action was to change the sender, and it succeeded, we have to
   update the in-store data. */
 
-  if (rc == 0 && strcmp(action + (int)strlen(action) - 4, "-Mes") == 0)
+  if (rc == 0 && Ustrcmp(action + Ustrlen(action) - 4, "-Mes") == 0)
     {
     queue_item *q = find_queue(id, queue_noop, 0);
     if (q != NULL)
       {
       if (q->sender != NULL) store_free(q->sender);
-      q->sender = store_malloc((int)strlen(address_arg) + 1);
-      strcpy(q->sender, address_arg);
+      q->sender = store_malloc(Ustrlen(address_arg) + 1);
+      Ustrcpy(q->sender, address_arg);
       }
     }
 
@@ -383,7 +363,7 @@ if ((pid = fork()) == 0)
   dup2(pipe_fd[1], 2);
   close(pipe_fd[1]);
 
-  system(buffer);
+  system(CS buffer);
 
   close(1);
   close(2);
@@ -399,7 +379,7 @@ if (pid < 0) text_showf(text, "Failed to fork: %s\n", strerror(pid)); else
 
   if (p == NULL)
     {
-    text_show(text, "Run out of store\n");
+    text_show(text, US"Run out of store\n");
     return;
     }
 
@@ -424,7 +404,7 @@ static void deliverAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-ActOnMessage((char *)client_data, "-v -M", "");
+ActOnMessage((uschar *)client_data, US"-v -M", US"");
 }
 
 
@@ -437,7 +417,7 @@ static void freezeAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-ActOnMessage((char *)client_data, "-Mf", "");
+ActOnMessage((uschar *)client_data, US"-Mf", US"");
 }
 
 
@@ -450,7 +430,7 @@ static void thawAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-ActOnMessage((char *)client_data, "-Mt", "");
+ActOnMessage((uschar *)client_data, US"-Mt", US"");
 }
 
 
@@ -465,7 +445,7 @@ start-up time. If the string is empty, do nothing. */
 
 XtActionProc dialogAction(Widget w, XEvent *event, String *ss, Cardinal *c)
 {
-char *s = XawDialogGetValueString(dialog_widget);
+uschar *s = US XawDialogGetValueString(dialog_widget);
 
 w = w;      /* Keep picky compilers happy */
 event = event;
@@ -474,7 +454,7 @@ c = c;
 
 XtPopdown((Widget)dialog_shell);
 XtDestroyWidget((Widget)dialog_shell);
-while (isspace((uschar)*s)) s++;
+while (isspace(*s)) s++;
 if (s[0] != 0)
   {
   if (actioned_message[0] != 0)
@@ -495,7 +475,7 @@ return NULL;
 be done to the application until the box is filled in. This
 function is also used by the Hide button handler. */
 
-void create_dialog(char *label, char *value)
+void create_dialog(uschar *label, uschar *value)
 {
 Arg warg[4];
 Dimension x, y, xx, yy;
@@ -509,9 +489,9 @@ get_pos_args[0].value = (XtArgVal)(&x);
 get_pos_args[1].value = (XtArgVal)(&y);
 XtGetValues(dialog_ref_widget, get_pos_args, 2);
 
-/* When this is not a message_specific thing, the position of the
-reference widget is relative to the window. Get the position of the top level
-widget and add to the position. */
+/* When this is not a message_specific thing, the position of the reference
+widget is relative to the window. Get the position of the top level widget and
+add to the position. */
 
 if (dialog_ref_widget != menushell)
   {
@@ -543,6 +523,7 @@ make it wider than the default, and override its translations to make Return
 call the dialog action function. */
 
 text = XtNameToWidget(dialog_widget, "value");
+XawTextSetInsertionPoint(text, Ustrlen(value));
 XtSetKeyboardFocus(dialog_widget, text);
 xs_SetValues(text, 1, "width", 200);
 pop_trans = XtParseTranslationTable(
@@ -563,17 +544,17 @@ XFlush(X_display);
 *        Cause a recipient to be added           *
 *************************************************/
 
-/*This just sets up the dialog box; the action happens when it is
-filled in. */
+/* This just sets up the dialog box; the action happens when it has been filled
+in. */
 
 static void addrecipAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-strcpy(actioned_message, (char *)client_data);
-action_required = "-Mar";
+Ustrcpy(actioned_message, (uschar *)client_data);
+action_required = US"-Mar";
 dialog_ref_widget = menushell;
-create_dialog("Recipient address to add?", "");
+create_dialog(US"Recipient address to add?", US"");
 }
 
 
@@ -586,10 +567,10 @@ static void markdelAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-strcpy(actioned_message, (char *)client_data);
-action_required = "-Mmd";
+Ustrcpy(actioned_message, (uschar *)client_data);
+action_required = US"-Mmd";
 dialog_ref_widget = menushell;
-create_dialog("Recipient address to mark delivered?", "");
+create_dialog(US"Recipient address to mark delivered?", US"");
 }
 
 
@@ -601,7 +582,7 @@ static void markalldelAction(Widget w, XtPointer client_data, XtPointer call_dat
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-ActOnMessage((char *)client_data, "-Mmad", "");
+ActOnMessage((uschar *)client_data, US"-Mmad", US"");
 }
 
 
@@ -613,32 +594,18 @@ static void editsenderAction(Widget w, XtPointer client_data,
   XtPointer call_data)
 {
 queue_item *q;
-char *sender;
+uschar *sender;
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-strcpy(actioned_message, (char *)client_data);
+Ustrcpy(actioned_message, (uschar *)client_data);
 q = find_queue(actioned_message, queue_noop, 0);
-sender = (q == NULL)? "" : q->sender;
-action_required = "-Mes";
+sender = (q == NULL)? US"" : (q->sender[0] == 0)? US"<>" : q->sender;
+action_required = US"-Mes";
 dialog_ref_widget = menushell;
-create_dialog("New sender address?", sender);
+create_dialog(US"New sender address?", sender);
 }
 
 
-
-/*************************************************
-*        Cause a message body to be edited       *
-*************************************************/
-
-static void editbodyAction(Widget w, XtPointer client_data, XtPointer call_data)
-{
-w = w;      /* Keep picky compilers happy */
-call_data = call_data;
-ActOnMessage((char *)client_data, "-Meb", "");
-}
-
-
-
 /*************************************************
 *    Cause a message to be returned to sender    *
 *************************************************/
@@ -647,7 +614,7 @@ static void giveupAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-ActOnMessage((char *)client_data, "-v -Mg", "");
+ActOnMessage((uschar *)client_data, US"-v -Mg", US"");
 }
 
 
@@ -660,7 +627,7 @@ static void removeAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
 w = w;      /* Keep picky compilers happy */
 call_data = call_data;
-ActOnMessage((char *)client_data, "-Mrm", "");
+ActOnMessage((uschar *)client_data, US"-Mrm", US"");
 }
 
 
@@ -671,9 +638,9 @@ ActOnMessage((char *)client_data, "-Mrm", "");
 
 static void headersAction(Widget w, XtPointer client_data, XtPointer call_data)
 {
-char buffer[256];
+uschar buffer[256];
 header_line *h, *next;
-Widget text = text_create((char *)client_data, text_depth);
+Widget text = text_create((uschar *)client_data, text_depth);
 void *reset_point;
 
 w = w;      /* Keep picky compilers happy */
@@ -684,14 +651,14 @@ Then use Exim's function to read the header. */
 
 reset_point = store_get(0);
 
-sprintf(buffer, "%s-H", (char *)client_data);
+sprintf(CS buffer, "%s-H", (uschar *)client_data);
 if (spool_read_header(buffer, TRUE, FALSE) != spool_read_OK)
   {
   if (errno == ERRNO_SPOOLFORMAT)
     {
     struct stat statbuf;
-    sprintf(big_buffer, "%s/input/%s", spool_directory, buffer);
-    if (stat(big_buffer, &statbuf) == 0)
+    sprintf(CS big_buffer, "%s/input/%s", spool_directory, buffer);
+    if (Ustat(big_buffer, &statbuf) == 0)
       text_showf(text, "Format error in spool file %s: size=%d\n", buffer,
         statbuf.st_size);
     else text_showf(text, "Format error in spool file %s\n", buffer);
@@ -710,14 +677,14 @@ if (sender_address != NULL)
 if (recipients_list != NULL)
   {
   int i;
-  text_show(text, "Recipients:\n");
+  text_show(text, US"Recipients:\n");
   for (i = 0; i < recipients_count; i++)
     {
     text_showf(text, "  %s %s\n",
       (tree_search(tree_nonrecipients, recipients_list[i].address) == NULL)?
         " ":"*", recipients_list[i].address);
     }
-  text_show(text, "\n");
+  text_show(text, US"\n");
   }
 
 for (h = header_list; h != NULL; h = next)
@@ -770,7 +737,7 @@ while (p != NULL)
 *             Set up popup text window           *
 *************************************************/
 
-static Widget text_create(char *name, int height)
+static Widget text_create(uschar *name, int height)
 {
 Widget textshell, form, text, button;
 
@@ -805,7 +772,7 @@ XawTextDisplayCaret(text, TRUE);
 
 if (queue_font != NULL)
   {
-  XFontStruct *f = XLoadQueryFont(X_display, queue_font);
+  XFontStruct *f = XLoadQueryFont(X_display, CS queue_font);
   if (f != NULL) xs_SetValues(text, 1, "font", f);
   }
 
@@ -835,11 +802,11 @@ void menu_create(Widget w, XEvent *event, String *actargs, Cardinal *count)
 {
 int line;
 int i;
-char *s;
+uschar *s;
 XawTextPosition p;
 Widget src, menu_line, item_1, item_2, item_3, item_4,
   item_5, item_6, item_7, item_8, item_9, item_10, item_11,
-  item_12, item_13, item_14;
+  item_12, item_13;
 XtTranslations menu_trans = XtParseTranslationTable(
   "<EnterWindow>:   highlight()\n\
    <LeaveWindow>:   unhighlight()\n\
@@ -871,7 +838,7 @@ while (i-- > 0)
   while (s[p] != 0 && s[p++] != '\n');
   }
 
-/* Now pointing either at 0 or 1st char after \n, or very 1st char.
+/* Now pointing either at 0 or 1st uschar after \n, or very 1st uschar.
 If 0, the click was beyond the end of the data; just set up a dummy
 menu. (Not easy to ignore as several actions are specified for the
 mouse click and it expects this one to set up a menu.) If on a
@@ -886,6 +853,18 @@ if (s[p] == 0)
   xs_SetValues(menushell, 2,
     "cursor",       XCreateFontCursor(X_display, XC_arrow),
     "translations", menu_trans);
+
+  /* To keep the widgets in XFree86 happy, we have to create at least one menu
+  item, it seems. (Openwindows doesn't mind a menu with no items.) Otherwise
+  there's a complaint about a zero width menu, and a crash. */
+
+  menu_line = XtCreateManagedWidget("line", smeLineObjectClass, menushell,
+    NULL, 0);
+
+  item_99_arg[0].value = (XtArgVal)menu_line;
+  (void)XtCreateManagedWidget("item99", smeBSBObjectClass, menushell,
+    item_99_arg, XtNumber(item_99_arg));
+
   highlighted_x = -1;
   return;
   }
@@ -899,7 +878,7 @@ while (p > 0 && s[p+11] == ' ')
 
 /* Now pointing at first character of a main line. */
 
-strncpy(message_id, s+p+11, MESSAGE_ID_LENGTH);
+Ustrncpy(message_id, s+p+11, MESSAGE_ID_LENGTH);
 message_id[MESSAGE_ID_LENGTH] = 0;
 
 /* Highlight the line being menued, and save its parameters so that it
@@ -993,11 +972,6 @@ item_13 = XtCreateManagedWidget("item13", smeBSBObjectClass, menushell,
   item_13_arg, XtNumber(item_13_arg));
 XtAddCallback(item_13, "callback",  editsenderAction, (XtPointer)message_id);
 
-item_14_arg[0].value = (XtArgVal)item_13;
-item_14 = XtCreateManagedWidget("item14", smeBSBObjectClass, menushell,
-  item_14_arg, XtNumber(item_14_arg));
-XtAddCallback(item_14, "callback",  editbodyAction, (XtPointer)message_id);
-
 /* Arrange that the menu pops up with the first item selected. */
 
 xs_SetValues(menushell, 1, "popupOnEntry", item_1);
