@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 #include "../exim.h"
@@ -21,23 +21,23 @@ Arguments:
 
 Returns:      OK on success
               BAD64 if response too large for buffer
-              FORCEFAIL if response is "*"
+              CANCELLED if response is "*"
 */
 
 int
-auth_get_data(char **aptr, char *challenge)
+auth_get_data(uschar **aptr, uschar *challenge)
 {
 int c;
 int p = 0;
-smtp_printf("334 %s\r\n", auth_b64encode(challenge, (int)strlen(challenge)));
-while ((c = accept_getc(smtp_in)) != '\n' && c != EOF)
+smtp_printf("334 %s\r\n", auth_b64encode(challenge, Ustrlen(challenge)));
+while ((c = receive_getc()) != '\n' && c != EOF)
   {
   if (p >= big_buffer_size - 1) return BAD64;
   big_buffer[p++] = c;
   }
 if (p > 0 && big_buffer[p-1] == '\r') p--;
 big_buffer[p] = 0;
-if (strcmp(big_buffer, "*") == 0) return FORCEFAIL;
+if (Ustrcmp(big_buffer, "*") == 0) return CANCELLED;
 *aptr = big_buffer;
 return OK;
 }
