@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 
@@ -26,10 +26,27 @@ are needed by any of the other headers, including system headers. */
 #define os_find_running_interfaces os_common_find_running_interfaces
 #endif
 
+/* If it didn't define the base for "base 62" numbers, we really do use 62.
+This is the case for all real Unix and Unix-like OS. It's only Cygwin and
+Darwin, with their case-insensitive file systems, that can't use base 62 for
+making unique names. */
+
+#ifndef BASE_62
+#define BASE_62 62
+#endif
+
+/* The maximum value of localhost_number depends on the base being used */
+
+#if BASE_62 == 62
+#define LOCALHOST_MAX  16
+#else
+#define LOCALHOST_MAX  10
+#endif
 
 /* ANSI C standard includes */
 
 #include <ctype.h>
+#include <locale.h>
 #include <signal.h>
 #include <stdarg.h>
 #include <stddef.h>
@@ -70,16 +87,31 @@ are needed by any of the other headers, including system headers. */
 #endif
 
 #include <sys/types.h>
+#include <sys/file.h>
 #include <dirent.h>
 #include <netdb.h>
 #include <pwd.h>
 #include <grp.h>
 #include <syslog.h>
 
+/* Not all systems have flock() available. Those that do must define LOCK_SH
+in sys/file.h. */
+
+#ifndef LOCK_SH
+#define NO_FLOCK
+#endif
+
 #ifndef NO_SYSEXITS        /* some OS don't have this */
 #include <sysexits.h>
 #endif
 
+/* A few OS don't have socklen_t; their os.h files define SOCKLEN_T to
+be size_t or whatever. */
+
+#ifndef SOCKLEN_T
+#define SOCKLEN_T socklen_t
+#endif
+
 /* Ensure that the sysexits we reference are defined */
 
 #ifndef EX_UNAVAILABLE
@@ -95,6 +127,11 @@ are needed by any of the other headers, including system headers. */
 #define EX_CONFIG      78        /* configuration error */
 #endif
 
+/* This one is not in any sysexits file that I've come across */
+
+#define EX_EXECFAILED 127        /* execve() failed */
+
+
 #include <sys/time.h>
 #include <sys/param.h>
 
@@ -157,6 +194,13 @@ or a macro with entries f_frsize and f_bsize. */
   #ifndef F_FAVAIL
   #define F_FAVAIL f_ffree
   #endif
+
+  /* All the systems I've been able to look at seem to have F_FILES */
+
+  #ifndef F_FILES
+  #define F_FILES  f_files
+  #endif
+
 #endif
 
 
@@ -191,10 +235,13 @@ present Exim doesn't actually use it. */
 #ifndef NO_NET_IF_H
 #include <net/if.h>
 #endif
+#include <sys/un.h>
 #include <netinet/in.h>
+#include <netinet/tcp.h>
 #include <arpa/inet.h>
 #include <arpa/nameser.h>
 
+
 /* If arpa/nameser.h defines a maximum name server packet size, use it,
 provided it is greater than 1024. Otherwise go for a default. PACKETSZ was used
 for this, but it seems that NS_PACKETSZ is coming into use. */
@@ -219,12 +266,16 @@ included in arpa/nameser.h. Fudge them here. */
 #endif
 
 /* Ancient systems (e.g. SunOS4) don't appear to have T_TXT defined in their
-header files. */
+header files. I don't suppose they have T_SRV either. */
 
 #ifndef T_TXT
 #define T_TXT 16
 #endif
 
+#ifndef T_SRV
+#define T_SRV 33
+#endif
+
 /* The resolv.h header defines __P(x) on some Solaris 2.5.1 systems (without
 checking that it is already defined, in fact). This conflicts with other
 headers that behave likewise (see below), leading to compiler warnings. Arrange
@@ -251,23 +302,18 @@ have netinet/ip_var.h, so there's a general macro to control its inclusion. */
 #include <netinet/ip_var.h>
 #endif
 
+/* Linux (and some others) uses a different type for the 2nd argument of
+iconv(). It's os.h file defines ICONV_ARG2_TYPE. For the rest, define a default
+here. */
 
-/* The ident.h header defines __P(x) if it is not already defined (for handling
-prototypes). Unfortunately, on some systems, db.h (which is included from
-Exim's dbfn.h for native Berkeley DB 1.85 use) also defines it, without
-checking for a previous definition. This leads to a warning message. However,
-on other systems, db.h doesn't define it, expecting it to be done by other
-headers. To avoid warnings and nevertheless keep things working in all cases,
-we arrange to undefine __P if and only if ident.h defines it. */
-
-#if defined(__P)
-#define __P_WAS_DEFINED_BEFORE_IDENT
+#ifndef ICONV_ARG2_TYPE
+#define ICONV_ARG2_TYPE const char **
 #endif
 
-#include "libident/ident.h"
+/* One OS uses a different type for the 5th argument of getsockopt */
 
-#if defined(__P) && ! defined (__P_WAS_DEFINED_BEFORE_IDENT)
-#undef __P
+#ifndef GETSOCKOPT_ARG5_TYPE
+#define GETSOCKOPT_ARG5_TYPE socklen_t *
 #endif
 
 /* One operating system uses a different type for the 2nd argument of select().
@@ -324,19 +370,32 @@ extern int ferror(FILE *);
 
 #include "pcre/pcre.h"
 
-/* Exim includes are in several files */
+/* Exim includes are in several files. Note that local_scan.h #includes
+mytypes.h and store.h, so we don't need to mention them explicitly. */
 
+#include "local_scan.h"
 #include "macros.h"
 #include "config.h"
-
-#include "dbfn.h"
+#include "dbstuff.h"
 #include "structs.h"
 #include "globals.h"
 #include "functions.h"
+#include "dbfunctions.h"
+#include "osfunctions.h"
 
 /* The following stuff must follow the inclusion of config.h because it
 requires various things that are set therein. */
 
+#if HAVE_ICONV             /* Not all OS have this */
+#include <iconv.h>
+#endif
+
+/* Backward compatibility; LOOKUP_LSEARCH now includes all three */
+
+#if (!defined LOOKUP_LSEARCH) && (defined LOOKUP_WILDLSEARCH || defined LOOKUP_NWILDLSEARCH)
+#define LOOKUP_LSEARCH yes
+#endif
+
 /* Define a union to hold either an IPv4 or an IPv6 sockaddr structure; this
 simplifies some of the coding. */
 
@@ -347,15 +406,19 @@ union sockaddr_46 {
   #endif
 };
 
-/* If SPOOL_DIRECTORY, STDERR_FILE, LOG_FILE_PATH or PID_FILE_PATH have not
-been defined, set them to the null string. */
+/* If SUPPORT_TLS is not defined, ensure that USE_GNUTLS is also not defined
+so that if USE_GNUTLS *is* set, we can assume SUPPORT_TLS is also set. */
+
+#ifndef SUPPORT_TLS
+#undef USE_GNUTLS
+#endif
+
+/* If SPOOL_DIRECTORY, LOG_FILE_PATH or PID_FILE_PATH have not been defined,
+set them to the null string. */
 
 #ifndef SPOOL_DIRECTORY
   #define SPOOL_DIRECTORY ""
 #endif
-#ifndef STDERR_FILE
-  #define STDERR_FILE ""
-#endif
 #ifndef LOG_FILE_PATH
   #define LOG_FILE_PATH ""
 #endif
@@ -367,7 +430,8 @@ been defined, set them to the null string. */
 There is a particular shambles in SunOS5, where it did not exist originally,
 but got installed with a particular patch for Solaris 2.4. There is a
 configuration variable for specifying what the system's "over quota" error is,
-but if it is not set, default to EDQUOT if it exists, otherwise ENOSPC. */
+which will end up in config.h if supplied in OS/Makefile-xxx. If it is not set,
+default to EDQUOT if it exists, otherwise ENOSPC. */
 
 #ifndef ERRNO_QUOTA
 #ifdef  EDQUOT
@@ -377,4 +441,15 @@ but if it is not set, default to EDQUOT if it exists, otherwise ENOSPC. */
 #endif
 #endif
 
+/* Ensure PATH_MAX is defined */
+
+#ifndef PATH_MAX
+  #ifdef MAXPATHLEN
+  #define PATH_MAX MAXPATHLEN
+  #else
+  #define PATH_MAX 1024
+  #endif
+#endif
+
+
 /* End of exim.h */
