@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 
@@ -16,8 +16,7 @@
 /* header_last points to the last header during message reception and delivery;
 otherwise it is NULL. We add new headers only when header_last is not NULL.
 The function may get called sometimes when it is NULL (e.g. during address
-verification where rewriting options exist). Headers are always added in the
-main storage pool.
+verification where rewriting options exist).
 
 Arguments:
   type      Exim header type character
@@ -30,25 +29,22 @@ Returns:    nothing
 void
 header_add(int type, char *format, ...)
 {
-int pool_reset = store_pool;
 header_line *new;
 va_list ap;
-char buffer[8192];
+uschar buffer[HEADER_ADD_BUFFER_SIZE];
 
 if (header_last == NULL) return;
 
 va_start(ap, format);
 if (!string_vformat(buffer, sizeof(buffer), format, ap))
-  log_write(0, LOG_PANIC_DIE, "string too long in header_add: %.100s ...",
+  log_write(0, LOG_MAIN|LOG_PANIC_DIE, "string too long in header_add: %.100s ...",
     buffer);
 va_end(ap);
 
-store_pool = POOL_MAIN;
 new = store_get(sizeof(header_line));
 new->text = string_copy(buffer);
-store_pool = pool_reset;
 
-new->slen = (int)strlen(buffer);
+new->slen = Ustrlen(buffer);
 new->next = NULL;
 new->type = type;
 
@@ -61,25 +57,45 @@ header_last = new;
 *          Check the name of a header            *
 *************************************************/
 
-/* This function is necessary because the colon is allowed to be separated from
-the name by white space. If it were not, a simple strncmpic would do.
+/* This function scans a table of header field names that Exim recognizes, and
+returns the identification of a match. If "resent" is true, the header is known
+to start with "resent-". In that case, the function matches only those fields
+that are allowed to appear with resent- in front of them.
 
 Arguments:
-  h         points to the header line
-  name      points to the header name to be checked, excluding the :
-  len       length of the name
+  h             points to the header line
+  is_resent     TRUE if the name starts "Resent-"
 
-Returns:    TRUE if the header starts with the name, followed by optional
-            white space and a colon
+Returns:        One of the htype_ enum values, identifying the header
 */
 
-BOOL
-header_checkname(header_line *h, char *name, int len)
+int
+header_checkname(header_line *h, BOOL is_resent)
 {
-char *text = h->text;
-while (len-- > 0) if (tolower(*text++) != tolower(*name++)) return FALSE;
-while (isspace((uschar)*text)) text++;
-return (*text == ':');
+uschar *text = h->text;
+header_name *bot = header_names;
+header_name *top = header_names + header_names_size;
+
+if (is_resent) text += 7;
+
+while (bot < top)
+  {
+  header_name *mid = bot + (top - bot)/2;
+  int c = strncmpic(text, mid->name, mid->len);
+
+  if (c == 0)
+    {
+    uschar *s = text + mid->len;
+    while (isspace(*s)) s++;
+    if (*s == ':')
+      return (!is_resent || mid->allow_resent)? mid->htype : htype_other;
+    c = 1;
+    }
+
+  if (c > 0) bot = mid + 1; else top = mid;
+  }
+
+return htype_other;
 }
 
 /* End of header.c */
