@@ -3,8 +3,6 @@
 *************************************************/
 
 /*
- * $Id: cdb.c,v 1.4.2.2 1998/05/29 16:33:12 cvs Exp $
- *
  * Exim - CDB database lookup module
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  *
@@ -15,6 +13,12 @@
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
  *
+ * --------------------------------------------------------------
+ * Modified by PH for Exim 4:
+ *   Changed over to using unsigned chars
+ *   Makes use of lf_check_file() for file checking
+ * --------------------------------------------------------------
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -49,6 +53,7 @@
 
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "cdb.h"
 
 #ifdef HAVE_MMAP
@@ -73,17 +78,15 @@
  */
 
 struct cdb_state {
-  int  fileno;
-  off_t filelen;
-  char *cdb_map;
-  char *cdb_offsets;
+  int     fileno;
+  off_t   filelen;
+  uschar *cdb_map;
+  uschar *cdb_offsets;
 };
 
 /* 32 bit unsigned type - this is an int on all modern machines */
 typedef unsigned int uint32;
 
-
-
 /*
  * cdb_hash()
  * Internal function to make hash value */
@@ -102,21 +105,19 @@ cdb_hash(uschar *buf, unsigned int len)
   return h;
 }
 
-
-
 /*
  * cdb_bread()
  * Internal function to read len bytes from disk, coping with oddities */
 
 static int
 cdb_bread(int fd,
-	 char *buf,
-	 int len)
+         uschar *buf,
+         int len)
 {
   int r;
   while (len > 0) {
     do
-      r = read(fd,buf,len);
+      r = Uread(fd,buf,len);
     while ((r == -1) && (errno == EINTR));
     if (r == -1) return -1;
     if (r == 0) { errno = EIO; return -1; }
@@ -126,40 +127,31 @@ cdb_bread(int fd,
   return 0;
 }
 
-
-
 /*
  * cdb_bread()
  * Internal function to parse 4 byte number (endian independant) */
 
-/* Argument changed from unsigned char * to char * by PH, in order to stop
-complaints from Sun's compiler since all calls to this function pass over
-char * arguments. Do the casting to unsigned inside the function. */
-
 static uint32
-cdb_unpack(char *buf)
+cdb_unpack(uschar *buf)
 {
   uint32 num;
-  uschar *ubuf = (uschar *)buf;
-  num = ubuf[3]; num <<= 8;
-  num += ubuf[2]; num <<= 8;
-  num += ubuf[1]; num <<= 8;
-  num += ubuf[0];
+  num =  buf[3]; num <<= 8;
+  num += buf[2]; num <<= 8;
+  num += buf[1]; num <<= 8;
+  num += buf[0];
   return num;
 }
 
-
-
 void *
-cdb_open(char *filename,
-	char **errmsg)
+cdb_open(uschar *filename,
+         uschar **errmsg)
 {
   int fileno;
   struct cdb_state *cdbp;
   struct stat statbuf;
   void * mapbuf;
 
-  fileno = open(filename, O_RDONLY);
+  fileno = Uopen(filename, O_RDONLY, 0);
   if (fileno == -1) {
     int save_errno = errno;
     *errmsg = string_open_failed(errno, "%s for cdb lookup", filename);
@@ -173,16 +165,16 @@ cdb_open(char *filename,
     if (statbuf.st_size < CDB_HASH_TABLE) {
       int save_errno = errno;
       *errmsg = string_open_failed(errno,
-				  "%s to short for cdb lookup",
-				  filename);
+                                  "%s to short for cdb lookup",
+                                  filename);
       errno = save_errno;
       return NULL;
     }
   } else {
     int save_errno = errno;
     *errmsg = string_open_failed(errno,
-				"fstat(%s) failed - cannot do cdb lookup",
-				filename);
+                                "fstat(%s) failed - cannot do cdb lookup",
+                                filename);
     errno = save_errno;
     return NULL;
   }
@@ -199,11 +191,11 @@ cdb_open(char *filename,
   /* if we are allowed to we use mmap here.... */
 #ifdef HAVE_MMAP
   mapbuf = mmap(NULL,
-	       statbuf.st_size,
-	       PROT_READ,
-	       MAP_SHARED,
-	       fileno,
-	       0);
+               statbuf.st_size,
+               PROT_READ,
+               MAP_SHARED,
+               fileno,
+               0);
   if (mapbuf != MAP_FAILED) {
     /* We have an mmap-ed section.  Now we can just use it */
     cdbp->cdb_map = mapbuf;
@@ -219,7 +211,7 @@ cdb_open(char *filename,
      * this since we fall back to slower methods....
      * However lets debug log it...
      */
-    DEBUG(1) debug_printf("cdb mmap failed - %d\n", errno);
+    DEBUG(D_lookup) debug_printf("cdb mmap failed - %d\n", errno);
   }
 #endif /* HAVE_MMAP */
 
@@ -235,8 +227,8 @@ cdb_open(char *filename,
      * time to give up I think....
      * call the close routine (deallocs the memory), and return NULL */
     *errmsg = string_open_failed(errno,
-				"cannot read header from %s for cdb lookup",
-				filename);
+                                "cannot read header from %s for cdb lookup",
+                                filename);
     cdb_close(cdbp);
     return NULL;
   }
@@ -253,20 +245,21 @@ cdb_open(char *filename,
 
 BOOL
 cdb_check(void *handle,
-	 char *filename,
-	 int modemask,
-	 int *owners,
-	 int *owngroups,
-	 char **errmsg)
+         uschar *filename,
+         int modemask,
+         uid_t *owners,
+         gid_t *owngroups,
+         uschar **errmsg)
 {
   struct cdb_state * cdbp = handle;
-  return search_check_file(cdbp->fileno,
-			  filename,
-			  modemask,
-			  owners,
-			  owngroups,
-			  "cdb",
-			  errmsg) == 0;
+  return lf_check_file(cdbp->fileno,
+                       filename,
+                       S_IFREG,
+                       modemask,
+                       owners,
+                       owngroups,
+                       "cdb",
+                       errmsg) == 0;
 }
 
 
@@ -277,11 +270,11 @@ cdb_check(void *handle,
 
 int
 cdb_find(void *handle,
-	char *filename,
-	char *keystring,
-	int  key_len,
-	char **result,
-	char **errmsg)
+        uschar *filename,
+        uschar *keystring,
+        int  key_len,
+        uschar **result,
+        uschar **errmsg)
 {
   struct cdb_state * cdbp = handle;
   uint32 item_key_len,
@@ -314,8 +307,8 @@ cdb_find(void *handle,
    * than the file, then we have problems.... */
   if ((hash_offset + (hash_offlen * CDB_HASH_ENTRY)) > cdbp->filelen) {
     *errmsg = string_sprintf("cdb: corrupt cdb file %s (too short)",
-			    filename);
-    DEBUG(3) debug_printf("%s\n", *errmsg);
+                            filename);
+    DEBUG(D_lookup) debug_printf("%s\n", *errmsg);
     return DEFER;
   }
 
@@ -325,8 +318,8 @@ cdb_find(void *handle,
 #ifdef HAVE_MMAP
   /* make sure the mmap was OK */
   if (cdbp->cdb_map != NULL) {
-    char * cur_pos = cur_offset + cdbp->cdb_map;
-    char * end_pos = end_offset + cdbp->cdb_map;
+    uschar * cur_pos = cur_offset + cdbp->cdb_map;
+    uschar * end_pos = end_offset + cdbp->cdb_map;
     for (loop = 0; (loop < hash_offlen); ++loop) {
       item_hash = cdb_unpack(cur_pos);
       cur_pos += 4;
@@ -338,24 +331,24 @@ cdb_find(void *handle,
 
       if (item_hash == key_hash) {
        /* matching hash value */
-       char * item_ptr = cdbp->cdb_map + item_posn;
+       uschar * item_ptr = cdbp->cdb_map + item_posn;
        item_key_len = cdb_unpack(item_ptr);
        item_ptr += 4;
        item_dat_len = cdb_unpack(item_ptr);
        item_ptr += 4;
        /* check key length matches */
        if (item_key_len == key_len) {
-	 /* finally check if key matches */
-	 if (strncmp(keystring, item_ptr, key_len) == 0) {
-	   /* we have a match....
-	    * make item_ptr point to data */
-	   item_ptr += item_key_len;
-	   /* ... and the returned result */
-	   *result = store_malloc(item_dat_len + 1);
-	   memcpy(*result, item_ptr, item_dat_len);
-	   (*result)[item_dat_len] = 0;
-	   return OK;
-	 }
+         /* finally check if key matches */
+         if (Ustrncmp(keystring, item_ptr, key_len) == 0) {
+           /* we have a match....
+            * make item_ptr point to data */
+           item_ptr += item_key_len;
+           /* ... and the returned result */
+           *result = store_malloc(item_dat_len + 1);
+           memcpy(*result, item_ptr, item_dat_len);
+           (*result)[item_dat_len] = 0;
+           return OK;
+         }
        }
       }
       /* handle warp round of table */
@@ -367,7 +360,7 @@ cdb_find(void *handle,
   }
 #endif /* HAVE_MMAP */
   for (loop = 0; (loop < hash_offlen); ++loop) {
-    char packbuf[8];
+    uschar packbuf[8];
     if (lseek(cdbp->fileno, (off_t) cur_offset,SEEK_SET) == -1) return DEFER;
     if (cdb_bread(cdbp->fileno, packbuf,8) == -1) return DEFER;
     item_hash = cdb_unpack(packbuf);
@@ -384,17 +377,17 @@ cdb_find(void *handle,
       /* check key length matches */
       if (item_key_len == key_len) {
        /* finally check if key matches */
-       char * item_key = store_malloc(key_len);
+       uschar * item_key = store_malloc(key_len);
        if (cdb_bread(cdbp->fileno, item_key, key_len) == -1) return DEFER;
-       if (strncmp(keystring, item_key, key_len) == 0) {
-	 /* matches - get data length */
-	 item_dat_len = cdb_unpack(packbuf + 4);
-	 /* then we build a new result string */
-	 *result = store_malloc(item_dat_len + 1);
-	 if (cdb_bread(cdbp->fileno, *result, item_dat_len) == -1)
-	   return DEFER;
-	 (*result)[item_dat_len] = 0;
-	 return OK;
+       if (Ustrncmp(keystring, item_key, key_len) == 0) {
+         /* matches - get data length */
+         item_dat_len = cdb_unpack(packbuf + 4);
+         /* then we build a new result string */
+         *result = store_malloc(item_dat_len + 1);
+         if (cdb_bread(cdbp->fileno, *result, item_dat_len) == -1)
+           return DEFER;
+         (*result)[item_dat_len] = 0;
+         return OK;
        }
       }
     }
@@ -422,7 +415,7 @@ struct cdb_state * cdbp = handle;
 
 #ifdef HAVE_MMAP
  if (cdbp->cdb_map) {
-   munmap(cdbp->cdb_map, cdbp->filelen);
+   munmap(CS cdbp->cdb_map, cdbp->filelen);
    if (cdbp->cdb_map == cdbp->cdb_offsets)
      cdbp->cdb_offsets = NULL;
  }
