@@ -2,10 +2,11 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "dbmdb.h"
 
 
@@ -16,7 +17,7 @@
 /* See local README for interface description */
 
 void *
-dbmdb_open(char *filename, char **errmsg)
+dbmdb_open(uschar *filename, uschar **errmsg)
 {
 EXIM_DB *yield;
 EXIM_DBOPEN(filename, O_RDONLY, 0, &yield);
@@ -38,34 +39,35 @@ return yield;
 /* This needs to know more about the underlying files than is good for it!
 We need to know what the real file names are in order to check the owners and
 modes. If USE_DB is set, we know it is Berkeley DB, which uses an unmodified
-file name. If USE_TDB is set, we know it is tdb, which does the same.
-Otherwise, for safety, we have to check for x.db or x.dir and x.pag. */
+file name. If USE_TDB or USE_GDBM is set, we know it is tdb or gdbm, which do
+the same. Otherwise, for safety, we have to check for x.db or x.dir and x.pag.
+*/
 
 BOOL
-dbmdb_check(void *handle, char *filename, int modemask, uid_t *owners,
-  gid_t *owngroups, char **errmsg)
+dbmdb_check(void *handle, uschar *filename, int modemask, uid_t *owners,
+  gid_t *owngroups, uschar **errmsg)
 {
 int rc;
 handle = handle;    /* Keep picky compilers happy */
 
-#if defined(USE_DB) || defined(USE_TDB)
-rc = search_check_file(-1, filename, modemask, owners, owngroups,
+#if defined(USE_DB) || defined(USE_TDB) || defined(USE_GDBM)
+rc = lf_check_file(-1, filename, S_IFREG, modemask, owners, owngroups,
   "dbm", errmsg);
 #else
   {
-  char filebuffer[256];
-  sprintf(filebuffer, "%.250s.db", filename);
-  rc = search_check_file(-1, filebuffer, modemask, owners, owngroups,
+  uschar filebuffer[256];
+  sprintf(CS filebuffer, "%.250s.db", filename);
+  rc = lf_check_file(-1, filebuffer, S_IFREG, modemask, owners, owngroups,
     "dbm", errmsg);
   if (rc < 0)        /* stat() failed */
     {
-    sprintf(filebuffer, "%.250s.dir", filename);
-    rc = search_check_file(-1, filebuffer, modemask, owners, owngroups,
+    sprintf(CS filebuffer, "%.250s.dir", filename);
+    rc = lf_check_file(-1, filebuffer, S_IFREG, modemask, owners, owngroups,
       "dbm", errmsg);
     if (rc == 0)     /* x.dir was OK */
       {
-      sprintf(filebuffer, "%.250s.pag", filename);
-      rc = search_check_file(-1, filebuffer, modemask, owners, owngroups,
+      sprintf(CS filebuffer, "%.250s.pag", filename);
+      rc = lf_check_file(-1, filebuffer, S_IFREG, modemask, owners, owngroups,
         "dbm", errmsg);
       }
     }
@@ -85,8 +87,8 @@ return rc == 0;
 the keylength in order to include the terminating zero. */
 
 int
-dbmdb_find(void *handle, char *filename, char *keystring, int length,
-  char **result, char **errmsg)
+dbmdb_find(void *handle, uschar *filename, uschar *keystring, int length,
+  uschar **result, uschar **errmsg)
 {
 EXIM_DB *d = (EXIM_DB *)handle;
 EXIM_DATUM key, data;
@@ -96,12 +98,12 @@ errmsg = errmsg;
 
 EXIM_DATUM_INIT(key);               /* Some DBM libraries require datums to */
 EXIM_DATUM_INIT(data);              /* be cleared before use. */
-EXIM_DATUM_DATA(key) = keystring;
+EXIM_DATUM_DATA(key) = CS keystring;
 EXIM_DATUM_SIZE(key) = length + 1;
 
 if (EXIM_DBGET(d, key, data))
   {
-  *result = string_copyn(EXIM_DATUM_DATA(data), EXIM_DATUM_SIZE(data));
+  *result = string_copyn(US EXIM_DATUM_DATA(data), EXIM_DATUM_SIZE(data));
   EXIM_DATUM_FREE(data);            /* Some DBM libraries need a free() call */
   return OK;
   }
@@ -117,8 +119,8 @@ return FAIL;
 /* See local README for interface description */
 
 int
-dbmnz_find(void *handle, char *filename, char *keystring, int length,
-  char **result, char **errmsg)
+dbmnz_find(void *handle, uschar *filename, uschar *keystring, int length,
+  uschar **result, uschar **errmsg)
 {
 return dbmdb_find(handle, filename, keystring, length-1, result, errmsg);
 }
