@@ -2,10 +2,11 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "dnsdb.h"
 
 
@@ -55,7 +56,7 @@ static int type_values[] = {
 /* See local README for interface description. */
 
 void *
-dnsdb_open(char *filename, char **errmsg)
+dnsdb_open(uschar *filename, uschar **errmsg)
 {
 filename = filename;   /* Keep picky compilers happy */
 errmsg = errmsg;       /* Ditto */
@@ -71,16 +72,16 @@ return (void *)(-1);   /* Any non-0 value */
 /* See local README for interface description. */
 
 int
-dnsdb_find(void *handle, char *filename, char *keystring, int length,
-  char **result, char **errmsg)
+dnsdb_find(void *handle, uschar *filename, uschar *keystring, int length,
+  uschar **result, uschar **errmsg)
 {
 int rc;
 int size = 0;
 int ptr = 0;
 int type = T_TXT;
-char *yield = NULL;
-char *equals = strchr(keystring, '=');
-char buffer[256];
+uschar *yield = NULL;
+uschar *equals = Ustrchr(keystring, '=');
+uschar buffer[256];
 
 dns_record *rr;
 dns_answer dnsa;
@@ -96,104 +97,33 @@ if (equals != NULL)
   {
   int i;
   int len = equals - keystring;
-  for (i = 0; i < sizeof(type_names)/sizeof(char *); i++)
+  for (i = 0; i < sizeof(type_names)/sizeof(uschar *); i++)
     {
-    if (len == (int)strlen(type_names[i]) &&
-        strncmpic(keystring, type_names[i], len) == 0)
+    if (len == Ustrlen(type_names[i]) &&
+        strncmpic(keystring, US type_names[i], len) == 0)
       {
       type = type_values[i];
       break;
       }
     }
-  if (i >= sizeof(type_names)/sizeof(char *))
+  if (i >= sizeof(type_names)/sizeof(uschar *))
     {
-    *errmsg = "unsupported DNS record type";
+    *errmsg = US"unsupported DNS record type";
     return DEFER;
     }
   keystring += len + 1;
   }
 
 /* If the type is PTR, we have to construct the relevant magic lookup
-key. */
+key. This code is now in a separate function. */
 
 if (type == T_PTR)
   {
-  char *p = keystring + (int)strlen(keystring);
-  char *pp = buffer;
-
-  /* Handle IPv4 address */
-
-  #if HAVE_IPV6
-  if (strchr(keystring, ':') == NULL)
-  #endif
-    {
-    int i;
-    for (i = 0; i < 4; i++)
-      {
-      char *ppp = p;
-      while (ppp > keystring && ppp[-1] != '.') ppp--;
-      strncpy(pp, ppp, p - ppp);
-      pp += p - ppp;
-      *pp++ = '.';
-      p = ppp - 1;
-      }
-    strcpy(pp, "in-addr.arpa");
-    }
-
-  /* Handle IPv6 address; convert to binary so as to fill out any
-  abbreviation in the textual form. */
-
-  #if HAVE_IPV6
-  else
-    {
-    int i;
-    int v6[4];
-    (void)host_aton(keystring, v6);
-
-    /* The original specification for IPv6 reverse lookup, to go with
-    AAAA records, was to invert each nibble, and look in the ip6.int
-    domain. This code does that. */
-
-    for (i = 3; i >= 0; i--)
-      {
-      int j;
-      for (j = 0; j < 32; j += 4)
-        {
-        sprintf(pp, "%x.", (v6[i] >> j) & 15);
-        pp += 2;
-        }
-      }
-    strcpy(pp, "ip6.int.");
-
-    /* The revised way of doind IPv6 reverse lookups is to construct
-    a binary key, and look in ip6.arpa. This code does that, but I have
-    not been able to make it work on Solairs 8. The resolver seems to
-    lose the initial backslash somehow. */
-
-    /* For the moment, therefore, just leave this out. It's a minor point -
-    how many people actually do PTR lookups using dnsdb? */
-
-    /**************************************************
-    strcpy(pp, "\\[x");
-    pp += 3;
-
-    for (i = 0; i < 4; i++)
-      {
-      sprintf(pp, "%08X", v6[i]);
-      pp += 8;
-      }
-    strcpy(pp, "].ip6.arpa.");
-    **************************************************/
-
-    }
-  #endif
-
-  /* Point at constructed key */
-
+  dns_build_reverse(keystring, buffer);
   keystring = buffer;
   }
 
-DEBUG(9) debug_printf("dnsdb key: %s\n", keystring);
+DEBUG(D_lookup) debug_printf("dnsdb key: %s\n", keystring);
 
 /* Initialize the resolver, in case this is the first time it is used
 in this run. Then do the lookup and sort out the result. */
@@ -201,7 +131,7 @@ in this run. Then do the lookup and sort out the result. */
 dns_init(FALSE, FALSE);
 rc = dns_lookup(&dnsa, keystring, type, NULL);
 
-if (rc == DNS_NOMATCH) return FAIL;
+if (rc == DNS_NOMATCH || rc == DNS_NODATA) return FAIL;
 if (rc != DNS_SUCCEED) return DEFER;
 
 for (rr = dns_next_rr(&dnsa, &dnss, RESET_ANSWERS);
@@ -209,7 +139,7 @@ for (rr = dns_next_rr(&dnsa, &dnss, RESET_ANSWERS);
      rr = dns_next_rr(&dnsa, &dnss, RESET_NEXT))
   {
   int len = 0;        /* Stop picky compilers warning */
-  char *s = NULL;
+  uschar *s = NULL;
 
   if (rr->type != type) continue;
 
@@ -223,22 +153,19 @@ for (rr = dns_next_rr(&dnsa, &dnss, RESET_ANSWERS);
       type == T_AAAA)
     {
     dns_address *da;
-
-    /* This loop runs only once for A and AAAA records */
-
     for (da = dns_address_from_rr(&dnsa, rr); da != NULL; da = da->next)
       {
-      len = (int)strlen(CS da->address);
+      len = Ustrlen(da->address);
       if (yield == NULL)
         {
-        yield = CS da->address;
+        yield = da->address;
         ptr = len;
         size = ptr + 1;
         }
       else
         {
-        yield = string_cat(yield, &size, &ptr, "\n", 1);
-        yield = string_cat(yield, &size, &ptr, CS da->address, len);
+        yield = string_cat(yield, &size, &ptr, US"\n", 1);
+        yield = string_cat(yield, &size, &ptr, da->address, len);
         }
       }
     continue;
@@ -249,7 +176,7 @@ for (rr = dns_next_rr(&dnsa, &dnss, RESET_ANSWERS);
   if (type == T_TXT)
     {
     len = (rr->data)[0];
-    s = string_copyn((char *)(rr->data+1), len);
+    s = string_copyn((uschar *)(rr->data+1), len);
     }
   else   /* T_CNAME, T_MX, T_NS, T_PTR */
     {
@@ -260,12 +187,12 @@ for (rr = dns_next_rr(&dnsa, &dnss, RESET_ANSWERS);
       {
       int number;
       GETSHORT(number, p);      /* pointer is advanced */
-      sprintf(s, "%d ", number);
-      len = (int)strlen(s);
+      sprintf(CS s, "%d ", number);
+      len = Ustrlen(s);
       }
     p += dn_expand(dnsa.answer, dnsa.answer + dnsa.answerlen, p,
       (DN_EXPAND_ARG4_TYPE)(s + len), ssize - len);
-    len += (int)strlen(s+len);
+    len += Ustrlen(s+len);
     store_reset(s + len + 1);
     }
 
@@ -277,7 +204,7 @@ for (rr = dns_next_rr(&dnsa, &dnss, RESET_ANSWERS);
     }
   else
     {
-    yield = string_cat(yield, &size, &ptr, "\n", 1);
+    yield = string_cat(yield, &size, &ptr, US"\n", 1);
     yield = string_cat(yield, &size, &ptr, s, len);
     }
   }
