@@ -2,12 +2,12 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Header for the dnsdb lookup */
 
-extern void *dnsdb_open(char *, char **);
-extern int   dnsdb_find(void *, char *, char *, int, char **, char **);
+extern void *dnsdb_open(uschar *, uschar **);
+extern int   dnsdb_find(void *, uschar *, uschar *, int, uschar **, uschar **);
 
 /* End of lookups/dnsdb.h */
