@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Thanks to Paul Kelly for contributing the original code for these
@@ -10,6 +10,7 @@ functions. */
 
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "mysql.h"       /* The local header */
 
 
@@ -34,7 +35,7 @@ static void dummy(int x) { dummy(x-1); }
 
 typedef struct mysql_connection {
   struct mysql_connection *next;
-  char  *server;
+  uschar  *server;
   MYSQL *handle;
 } mysql_connection;
 
@@ -49,7 +50,7 @@ static mysql_connection *mysql_connections = NULL;
 /* See local README for interface description. */
 
 void *
-mysql_open(char *filename, char **errmsg)
+mysql_open(uschar *filename, uschar **errmsg)
 {
 return (void *)(1);    /* Just return something non-null */
 }
@@ -69,7 +70,7 @@ mysql_connection *cn;
 while ((cn = mysql_connections) != NULL)
   {
   mysql_connections = cn->next;
-  DEBUG(9) debug_printf("close MYSQL connection: %s\n", cn->server);
+  DEBUG(D_lookup) debug_printf("close MYSQL connection: %s\n", cn->server);
   mysql_close(cn->handle);
   store_free(cn->server);
   store_free(cn->handle);
@@ -94,14 +95,15 @@ Arguments:
   defer_break  TRUE if no more servers are to be tried after DEFER
 
 The server string is of the form "host/dbname/user/password". The host can be
-host:port.
+host:port. This string is in a nextinlist temporary buffer, so can be
+overwritten.
 
 Returns:       OK, FAIL, or DEFER
 */
 
 static int
-perform_mysql_search(char *query, char *server, char **resultptr, char **errmsg,
-  BOOL *defer_break)
+perform_mysql_search(uschar *query, uschar *server, uschar **resultptr,
+  uschar **errmsg, BOOL *defer_break)
 {
 MYSQL *mysql_handle = NULL;        /* Keep compilers happy */
 MYSQL_RES *mysql_result = NULL;
@@ -113,55 +115,75 @@ int ssize = 0;
 int offset = 0;
 int yield = DEFER;
 unsigned int num_fields;
-char *result = NULL;
+uschar *result = NULL;
 mysql_connection *cn;
+uschar *server_copy = NULL;
+uschar *sdata[4];
+
+/* Disaggregate the parameters from the server argument. The order is host,
+database, user, password. We can write to the string, since it is in a
+nextinlist temporary buffer. The copy of the string that is used for caching
+has the password removed. This copy is also used for debugging output. */
+
+for (i = 3; i > 0; i--)
+  {
+  uschar *pp = Ustrrchr(server, '/');
+  if (pp == NULL)
+    {
+    *errmsg = string_sprintf("incomplete MySQL server data: %s",
+      (i == 3)? server : server_copy);
+    *defer_break = TRUE;
+    return DEFER;
+    }
+  *pp++ = 0;
+  sdata[i] = pp;
+  if (i == 3) server_copy = string_copy_malloc(server);  /* sans password */
+  }
+sdata[0] = server;   /* What's left at the start */
 
 /* See if we have a cached connection to the server */
 
 for (cn = mysql_connections; cn != NULL; cn = cn->next)
   {
-  if (strcmp(cn->server, server) == 0)
+  if (Ustrcmp(cn->server, server_copy) == 0)
     {
     mysql_handle = cn->handle;
     break;
     }
   }
 
-/* If no cached connection, we must set one up */
+/* If no cached connection, we must set one up. Mysql allows for a host name
+and port to be specified. It also allows the name of a Unix socket to be used.
+Unfortunately, this contains slashes, but its use is expected to be rare, so
+the rather cumbersome syntax shouldn't inconvenience too many people. We use
+this:  host:port(socket)  where all the parts are optional. */
 
 if (cn == NULL)
   {
+  uschar *p;
+  uschar *socket = NULL;
   int port = 0;
-  char *p = server;
-  char *server_copy = string_copy_malloc(server);
-  char *sdata[4];
 
-  /* Disaggregate the parameters from the server argument. The order is
-  host, database, user, password. We can write to the string, since it is in an
-  isinlist temporary buffer. */
-
-  for (i = 0; i < 3; i++)
-    {
-    char *pp = strchr(p, '/');
-    if (pp == NULL)
-      {
-      *errmsg = string_sprintf("incomplete MYSQL server data: %s", server_copy);
-      store_free(server_copy);
-      *defer_break = TRUE;
-      return DEFER;
-      }
-    sdata[i] = p;
-    *pp++ = 0;
-    p = pp;
-    }
-  sdata[3] = p;    /* Final thing (password) is all the rest */
-
-  /* If host contains a colon, sort out the port number. */
-
-  if ((p = strchr(sdata[0], ':')) != NULL)
+  if ((p = Ustrchr(sdata[0], '(')) != NULL)
     {
     *p++ = 0;
-    port = atoi(p);
+    socket = p;
+    while (*p != 0 && *p != ')') p++;
+    *p = 0;
+    }
+
+  if ((p = Ustrchr(sdata[0], ':')) != NULL)
+    {
+    *p++ = 0;
+    port = Uatoi(p);
+    }
+
+  if (Ustrchr(sdata[0], '/') != NULL)
+    {
+    *errmsg = string_sprintf("unexpected slash in MySQL server hostname: %s",
+      sdata[0]);
+    *defer_break = TRUE;
+    return DEFER;
     }
 
   /* If the database is the empty string, set it NULL - the query must then
@@ -169,16 +191,18 @@ if (cn == NULL)
 
   if (sdata[1][0] == 0) sdata[1] = NULL;
 
-  DEBUG(9) debug_printf("MYSQL new connection: host=%s port=%d database=%s "
-    "user=%s password=%s\n", sdata[0], port, sdata[1], sdata[2], sdata[3]);
+  DEBUG(D_lookup)
+    debug_printf("MYSQL new connection: host=%s port=%d socket=%s "
+      "database=%s user=%s\n", sdata[0], port, socket, sdata[1], sdata[2]);
 
   /* Get store for a new handle, initialize it, and connect to the server */
 
   mysql_handle = store_malloc(sizeof(MYSQL));
   mysql_init(mysql_handle);
-                                    /*  host       user    passwd    database */
-  if (mysql_real_connect(mysql_handle, sdata[0], sdata[2], sdata[3], sdata[1],
-      port, NULL, 0) == NULL)
+  if (mysql_real_connect(mysql_handle,
+      /*  host        user         passwd     database */
+      CS sdata[0], CS sdata[2], CS sdata[3], CS sdata[1],
+      port, CS socket, 0) == NULL)
     {
     *errmsg = string_sprintf("MYSQL connection failed: %s",
       mysql_error(mysql_handle));
@@ -201,12 +225,13 @@ if (cn == NULL)
 
 else
   {
-  DEBUG(9) debug_printf("MYSQL using cached connection for %s\n", server);
+  DEBUG(D_lookup)
+    debug_printf("MYSQL using cached connection for %s\n", server_copy);
   }
 
 /* Run the query */
 
-if (mysql_query(mysql_handle, query) != 0)
+if (mysql_query(mysql_handle, CS query) != 0)
   {
   *errmsg = string_sprintf("MYSQL: query failed: %s\n",
     mysql_error(mysql_handle));
@@ -214,10 +239,20 @@ if (mysql_query(mysql_handle, query) != 0)
   goto MYSQL_EXIT;
   }
 
-/* Pick up the result */
+/* Pick up the result. If the query was not of the type that returns data,
+namely INSERT, UPDATE, or DELETE, an error occurs here. However, this situation
+can be detected by calling mysql_field_count(). If its result is zero, no data
+was expected (this is all explained clearly in the MySQL manual). In this case,
+we return the number of rows affected by the command. */
 
 if ((mysql_result = mysql_use_result(mysql_handle)) == NULL)
   {
+  if ( mysql_field_count(mysql_handle) == 0 )
+    {
+    DEBUG(D_lookup) debug_printf("MYSQL: query was not one that returns data\n");
+    result = string_sprintf("%d", mysql_affected_rows(mysql_handle));
+    goto MYSQL_EXIT;
+    }
   *errmsg = string_sprintf("MYSQL: lookup result failed: %s\n",
     mysql_error(mysql_handle));
   *defer_break = FALSE;
@@ -239,40 +274,47 @@ while ((mysql_row_data = mysql_fetch_row(mysql_result)) != NULL)
   unsigned long *lengths = mysql_fetch_lengths(mysql_result);
 
   if (result != NULL)
-      result = string_cat(result, &ssize, &offset, "\n", 1);
+      result = string_cat(result, &ssize, &offset, US"\n", 1);
 
   if (num_fields == 1)
     {
-    result = string_cat(result ,&ssize, &offset, mysql_row_data[0],
-      lengths[0]);
+    if (mysql_row_data[0] != NULL)    /* NULL value yields nothing */
+      result = string_cat(result, &ssize, &offset, US mysql_row_data[0],
+        lengths[0]);
     }
 
   else for (i = 0; i < num_fields; i++)
     {
-    result = string_cat(result, &ssize, &offset, fields[i].name,
-      strlen(fields[i].name));
-    result = string_cat(result, &ssize, &offset, "=", 1);
+    result = string_cat(result, &ssize, &offset, US fields[i].name,
+      Ustrlen(fields[i].name));
+    result = string_cat(result, &ssize, &offset, US"=", 1);
 
     /* Quote the value if it contains spaces or is empty */
 
-    if (mysql_row_data[i][0] == 0 || strchr(mysql_row_data[i], ' ') != NULL)
+    if (mysql_row_data[i] == NULL)    /* NULL value */
+      {
+      result = string_cat(result, &ssize, &offset, US"\"\"", 2);
+      }
+
+    else if (mysql_row_data[i][0] == 0 ||
+             Ustrchr(mysql_row_data[i], ' ') != NULL)
       {
       int j;
-      result = string_cat(result, &ssize, &offset, "\"", 1);
+      result = string_cat(result, &ssize, &offset, US"\"", 1);
       for (j = 0; j < lengths[i]; j++)
         {
-	if (mysql_row_data[i][j] == '\"' || mysql_row_data[i][j] == '\\')
-	  result = string_cat(result, &ssize, &offset, "\\", 1);
-	result = string_cat(result, &ssize, &offset, mysql_row_data[i]+j, 1);
+        if (mysql_row_data[i][j] == '\"' || mysql_row_data[i][j] == '\\')
+          result = string_cat(result, &ssize, &offset, US"\\", 1);
+        result = string_cat(result, &ssize, &offset, US mysql_row_data[i]+j, 1);
         }
-      result = string_cat(result, &ssize, &offset, "\"", 1);
+      result = string_cat(result, &ssize, &offset, US"\"", 1);
       }
     else
       {
-      result = string_cat(result ,&ssize, &offset, mysql_row_data[i],
+      result = string_cat(result, &ssize, &offset, US mysql_row_data[i],
         lengths[i]);
       }
-    result = string_cat(result, &ssize, &offset, " ", 1);
+    result = string_cat(result, &ssize, &offset, US" ", 1);
     }
   }
 
@@ -283,7 +325,7 @@ always leaves enough room for a terminating zero. */
 if (result == NULL)
   {
   yield = FAIL;
-  *errmsg = "MYSQL: no data found";
+  *errmsg = US"MYSQL: no data found";
   }
 else result[offset] = 0;
 
@@ -305,7 +347,7 @@ if(result != NULL)
   }
 else
   {
-  DEBUG(9) debug_printf("%s\n", *errmsg);
+  DEBUG(D_lookup) debug_printf("%s\n", *errmsg);
   return yield;      /* FAIL or DEFER */
   }
 }
@@ -322,25 +364,25 @@ arguments are not used. Loop through a list of servers while the query is
 deferred with a retryable error. */
 
 int
-mysql_find(void *handle, char *filename, char *query, int length,
-  char **result, char **errmsg)
+mysql_find(void *handle, uschar *filename, uschar *query, int length,
+  uschar **result, uschar **errmsg)
 {
 int sep = 0;
-char *server;
-char *list = mysql_servers;
-char buffer[512];
+uschar *server;
+uschar *list = mysql_servers;
+uschar buffer[512];
 
-DEBUG(9) debug_printf("MYSQL query: %s\n", query);
+DEBUG(D_lookup) debug_printf("MYSQL query: %s\n", query);
 
 while ((server = string_nextinlist(&list, &sep, buffer, sizeof(buffer))) != NULL)
   {
-  BOOL defer_break;
+  BOOL defer_break = FALSE;
   int rc = perform_mysql_search(query, server, result, errmsg, &defer_break);
   if (rc != DEFER || defer_break) return rc;
   }
 
 if (mysql_servers == NULL)
-  *errmsg = "no MYSQL servers defined (mysql_servers option)";
+  *errmsg = US"no MYSQL servers defined (mysql_servers option)";
 
 return DEFER;
 }
@@ -360,27 +402,32 @@ Sadly, MySQL doesn't seem to behave like other programs. If you use something
 like "where id="ab\%cd" it does not treat the string as "ab%cd". So you really
 can't quote "on spec".
 
-Argument: the string to be quoted
-Returns:  the processed string
+Arguments:
+  s          the string to be quoted
+  opt        additional option text or NULL if none
+
+Returns:     the processed string or NULL for a bad option
 */
 
-char *
-mysql_quote(char *s)
+uschar *
+mysql_quote(uschar *s, uschar *opt)
 {
 register int c;
 int count = 0;
-char *t = s;
-char *quoted;
+uschar *t = s;
+uschar *quoted;
+
+if (opt != NULL) return NULL;     /* No options recognized */
 
 while ((c = *t++) != 0)
-  if (strchr("\n\t\r\b\'\"\\", c) != NULL) count++;
+  if (Ustrchr("\n\t\r\b\'\"\\", c) != NULL) count++;
 
 if (count == 0) return s;
-t = quoted = store_get((int)strlen(s) + count + 1);
+t = quoted = store_get(Ustrlen(s) + count + 1);
 
 while ((c = *s++) != 0)
   {
-  if (strchr("\n\t\r\b\'\"\\", c) != NULL)
+  if (Ustrchr("\n\t\r\b\'\"\\", c) != NULL)
     {
     *t++ = '\\';
     switch(c)
