@@ -2,10 +2,11 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "nis.h"
 
 /* We can't just compile this code and allow the library mechanism to omit the
@@ -31,7 +32,7 @@ static void dummy(int x) { dummy(x-1); }
 the "nis" and "nis0" lookup types. */
 
 void *
-nis_open(char *filename, char **errmsg)
+nis_open(uschar *filename, uschar **errmsg)
 {
 char *nis_domain;
 if (yp_get_default_domain(&nis_domain) != 0)
@@ -53,14 +54,14 @@ for nis0 because they are so short it isn't worth trying to use any common
 code. */
 
 int
-nis_find(void *handle, char *filename, char *keystring, int length,
-  char **result, char **errmsg)
+nis_find(void *handle, uschar *filename, uschar *keystring, int length,
+  uschar **result, uschar **errmsg)
 {
 int rc;
-char *nis_data;
+uschar *nis_data;
 int nis_data_length;
-if ((rc = yp_match((char *)handle, filename, keystring, length,
-    &nis_data, &nis_data_length)) == 0)
+if ((rc = yp_match(CS handle, CS filename, CS keystring, length,
+    CSS &nis_data, &nis_data_length)) == 0)
   {
   *result = string_copy(nis_data);
   (*result)[nis_data_length] = 0;    /* remove final '\n' */
@@ -78,14 +79,14 @@ return (rc == YPERR_KEY || rc == YPERR_MAP)? FAIL : DEFER;
 /* See local README for interface description. */
 
 int
-nis0_find(void *handle, char *filename, char *keystring, int length,
-  char **result, char **errmsg)
+nis0_find(void *handle, uschar *filename, uschar *keystring, int length,
+  uschar **result, uschar **errmsg)
 {
 int rc;
-char *nis_data;
+uschar *nis_data;
 int nis_data_length;
-if ((rc = yp_match((char *)handle, filename, keystring, length + 1,
-    &nis_data, &nis_data_length)) == 0)
+if ((rc = yp_match(CS handle, CS filename, CS keystring, length + 1,
+    CSS &nis_data, &nis_data_length)) == 0)
   {
   *result = string_copy(nis_data);
   (*result)[nis_data_length] = 0;    /* remove final '\n' */
