@@ -2,13 +2,13 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Header for the nis and nis0 lookups */
 
-extern void *nis_open(char *, char **);
-extern int   nis_find(void *, char *, char *, int, char **, char **);
-extern int   nis0_find(void *, char *, char *, int, char **, char **);
+extern void *nis_open(uschar *, uschar **);
+extern int   nis_find(void *, uschar *, uschar *, int, uschar **, uschar **);
+extern int   nis0_find(void *, uschar *, uschar *, int, uschar **, uschar **);
 
 /* End of lookups/nis.h */
