@@ -2,10 +2,11 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "nisplus.h"
 
 /* We can't just compile this code and allow the library mechanism to omit the
@@ -32,7 +33,7 @@ static void dummy(int x) { dummy(x-1); }
 /* See local README for interface description. */
 
 void *
-nisplus_open(char *filename, char **errmsg)
+nisplus_open(uschar *filename, uschar **errmsg)
 {
 return (void *)(1);    /* Just return something non-null */
 }
@@ -56,21 +57,21 @@ yield is the concatenation of all the fields, preceded by their names and an
 equals sign. */
 
 int
-nisplus_find(void *handle, char *filename, char *query, int length,
-  char **result, char **errmsg)
+nisplus_find(void *handle, uschar *filename, uschar *query, int length,
+  uschar **result, uschar **errmsg)
 {
 int i;
 int ssize = 0;
 int offset = 0;
 int error_error = FAIL;
-char *field_name = NULL;
+uschar *field_name = NULL;
 nis_result *nrt = NULL;
 nis_result *nre = NULL;
 nis_object *tno, *eno;
 struct entry_obj *eo;
 struct table_obj *ta;
-char *p = query + length;
-char *yield = NULL;
+uschar *p = query + length;
+uschar *yield = NULL;
 
 /* Search backwards for a colon to see if a result field name
 has been given. */
@@ -90,7 +91,7 @@ table name. */
 while (p > query && p[-1] != ',') p--;
 if (p <= query)
   {
-  *errmsg = "NIS+ query malformed";
+  *errmsg = US"NIS+ query malformed";
   error_error = DEFER;
   goto NISPLUS_EXIT;
   }
@@ -99,7 +100,7 @@ if (p <= query)
 check that we got back a table, and set up pointers so the field
 names can be scanned. */
 
-nrt = nis_lookup(p, EXPAND_NAME | NO_CACHE);
+nrt = nis_lookup(CS p, EXPAND_NAME | NO_CACHE);
 if (nrt->status != NIS_SUCCESS)
   {
   *errmsg = string_sprintf("NIS+ error accessing %s table: %s", p,
@@ -119,7 +120,7 @@ ta = &(tno->zo_data.objdata_u.ta_data);
 /* Now look up the entry in the table, check that we got precisely one
 object and that it is a table entry. */
 
-nre = nis_list(query, EXPAND_NAME, NULL, NULL);
+nre = nis_list(CS query, EXPAND_NAME, NULL, NULL);
 if (nre->status != NIS_SUCCESS)
   {
   *errmsg = string_sprintf("NIS+ error accessing entry %s: %s",
@@ -154,46 +155,46 @@ for (i = 0; i < eo->en_cols.en_cols_len; i++)
   table_col *tc = ta->ta_cols.ta_cols_val + i;
   entry_col *ec = eo->en_cols.en_cols_val + i;
   int len = ec->ec_value.ec_value_len;
-  char *value = ec->ec_value.ec_value_val;
+  uschar *value = US ec->ec_value.ec_value_val;
 
   /* The value may be NULL for a zero-length field. Turn this into an
   empty string for consistency. Remove trailing whitespace and zero
   bytes. */
 
-  if (value == NULL) value = ""; else
-    while (len > 0 && (value[len-1] == 0 || isspace((uschar)value[len-1])))
+  if (value == NULL) value = US""; else
+    while (len > 0 && (value[len-1] == 0 || isspace(value[len-1])))
       len--;
 
   /* Concatenate all fields if no specific one selected */
 
   if (field_name == NULL)
     {
-    yield = string_cat(yield, &ssize, &offset, tc->tc_name,
-      (int)strlen(tc->tc_name));
-    yield = string_cat(yield, &ssize, &offset, "=", 1);
+    yield = string_cat(yield, &ssize, &offset,US  tc->tc_name,
+      Ustrlen(tc->tc_name));
+    yield = string_cat(yield, &ssize, &offset, US"=", 1);
 
     /* Quote the value if it contains spaces or is empty */
 
-    if (value[0] == 0 || strchr(value, ' ') != NULL)
+    if (value[0] == 0 || Ustrchr(value, ' ') != NULL)
       {
       int j;
-      yield = string_cat(yield, &ssize, &offset, "\"", 1);
+      yield = string_cat(yield, &ssize, &offset, US"\"", 1);
       for (j = 0; j < len; j++)
         {
         if (value[j] == '\"' || value[j] == '\\')
-          yield = string_cat(yield, &ssize, &offset, "\\", 1);
+          yield = string_cat(yield, &ssize, &offset, US"\\", 1);
         yield = string_cat(yield, &ssize, &offset, value+j, 1);
         }
-      yield = string_cat(yield, &ssize, &offset, "\"", 1);
+      yield = string_cat(yield, &ssize, &offset, US"\"", 1);
       }
     else yield = string_cat(yield, &ssize, &offset, value, len);
 
-    yield = string_cat(yield, &ssize, &offset, " ", 1);
+    yield = string_cat(yield, &ssize, &offset, US" ", 1);
     }
 
   /* When the specified field is found, grab its data and finish */
 
-  else if (strcmp(field_name, tc->tc_name) == 0)
+  else if (Ustrcmp(field_name, tc->tc_name) == 0)
     {
     yield = string_copyn(value, len);
     goto NISPLUS_EXIT;
@@ -232,22 +233,28 @@ return error_error;      /* FAIL or DEFER */
 *************************************************/
 
 /* The only quoting that is necessary for NIS+ is to double any doublequote
-characters.
+characters. No options are recognized.
 
-Argument: the string to be quoted
-Returns:  the processed string
+Arguments:
+  s          the string to be quoted
+  opt        additional option text or NULL if none
+
+Returns:     the processed string or NULL for a bad option
 */
 
-char *
-nisplus_quote(char *s)
+uschar *
+nisplus_quote(uschar *s, uschar *opt)
 {
 int count = 0;
-char *quoted;
-char *t = s;
+uschar *quoted;
+uschar *t = s;
+
+if (opt != NULL) return NULL;    /* No options recognized */
+
 while (*t != 0) if (*t++ == '\"') count++;
 if (count == 0) return s;
 
-t = quoted = store_get((int)strlen(s) + count + 1);
+t = quoted = store_get(Ustrlen(s) + count + 1);
 
 while (*s != 0)
   {
