@@ -2,13 +2,15 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Thanks to Petr Cech for contributing the original code for these
-functions. */
+functions. Thanks to Joachim Wieland for the initial patch for the Unix domain
+socket extension. */
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "pgsql.h"       /* The local header */
 
 /* We can't just compile this code and allow the library mechanism to omit the
@@ -31,7 +33,7 @@ static void dummy(int x) { dummy(x-1); }
 
 typedef struct pgsql_connection {
   struct pgsql_connection *next;
-  char   *server;
+  uschar *server;
   PGconn *handle;
 } pgsql_connection;
 
@@ -46,7 +48,7 @@ static pgsql_connection *pgsql_connections = NULL;
 /* See local README for interface description. */
 
 void *
-pgsql_open(char *filename, char **errmsg)
+pgsql_open(uschar *filename, uschar **errmsg)
 {
 return (void *)(1);    /* Just return something non-null */
 }
@@ -66,7 +68,7 @@ pgsql_connection *cn;
 while ((cn = pgsql_connections) != NULL)
   {
   pgsql_connections = cn->next;
-  DEBUG(9) debug_printf("close PGSQL connection: %s\n", cn->server);
+  DEBUG(D_lookup) debug_printf("close PGSQL connection: %s\n", cn->server);
   PQfinish(cn->handle);
   store_free(cn->server);
   store_free(cn);
@@ -80,8 +82,20 @@ while ((cn = pgsql_connections) != NULL)
 *************************************************/
 
 /* This function is called from the find entry point to do the search for a
-single server. The server string is of the form "host/dbname/user/password".
-The host can be host:port.
+single server. The server string is of the form "server/dbname/user/password".
+
+PostgreSQL supports connections through Unix domain sockets. This is usually
+faster and costs less cpu time than a TCP/IP connection. However it can only be
+used if the mail server runs on the same machine as the database server. A
+configuration line for PostgreSQL via Unix domain sockets looks like this:
+
+hide pgsql_servers = (/tmp/.s.PGSQL.5432)/db/user/password[:<nextserver>]
+
+We enclose the path name in parentheses so that its slashes aren't visually
+confused with the delimeters for the other pgsql_server settings.
+
+For TCP/IP connections, the server is a host name and optional port (with a
+colon separator).
 
 NOTE:
  1) All three '/' must be present.
@@ -89,7 +103,7 @@ NOTE:
 
 Arguments:
   query        the query string
-  server       the server string
+  server       the server string; this is in dynamic memory and can be updated
   resultptr    where to store the result
   errmsg       where to point an error message
   defer_break  TRUE if no more servers are to be tried after DEFER
@@ -98,8 +112,8 @@ Returns:       OK, FAIL, or DEFER
 */
 
 static int
-perform_pgsql_search(char *query, char *server, char **resultptr, char **errmsg,
-  BOOL *defer_break)
+perform_pgsql_search(uschar *query, uschar *server, uschar **resultptr,
+  uschar **errmsg, BOOL *defer_break)
 {
 PGconn *pg_conn = NULL;
 PGresult *pg_result = NULL;
@@ -109,71 +123,118 @@ int ssize = 0;
 int offset = 0;
 int yield = DEFER;
 unsigned int num_fields, num_tuples;
-char *result = NULL;
+uschar *result = NULL;
 pgsql_connection *cn;
+uschar *server_copy = NULL;
+uschar *sdata[3];
 
-/* See if we have a cached connection to the server */
+/* Disaggregate the parameters from the server argument. The order is host or
+path, database, user, password. We can write to the string, since it is in a
+nextinlist temporary buffer. The copy of the string that is used for caching
+has the password removed. This copy is also used for debugging output. */
+
+for (i = 2; i >= 0; i--)
+  {
+  uschar *pp = Ustrrchr(server, '/');
+  if (pp == NULL)
+    {
+    *errmsg = string_sprintf("incomplete pgSQL server data: %s",
+      (i == 2)? server : server_copy);
+    *defer_break = TRUE;
+    return DEFER;
+    }
+  *pp++ = 0;
+  sdata[i] = pp;
+  if (i == 2) server_copy = string_copy_malloc(server);  /* sans password */
+  }
+
+/* The total server string has now been truncated so that what is left at the
+start is the identification of the server (host or path). See if we have a
+cached connection to the server. */
 
 for (cn = pgsql_connections; cn != NULL; cn = cn->next)
   {
-  if (strcmp(cn->server, server) == 0)
+  if (Ustrcmp(cn->server, server_copy) == 0)
     {
     pg_conn = cn->handle;
     break;
     }
   }
 
-/* If no cached connection, we must set one up */
+/* If there is no cached connection, we must set one up. */
 
 if (cn == NULL)
   {
-  int port = 0;
-  char *p = server;
-  char *server_copy = string_copy_malloc(server);
-  char *sdata[5];
+  uschar *port = US"";
 
-  /* Disaggregate the parameters from the server argument. The order is
-  host, database, user, password. We can write to the string, since it is in an
-  isinlist temporary buffer. */
+  /* For a Unix domain socket connection, the path is in parentheses */
 
-  for (i = 0; i < 3; i++)
+  if (*server == '(')
     {
-    char *pp = strchr(p, '/');
-    if (pp == NULL)
+    uschar *last_slash, *last_dot, *p;
+
+    p = ++server;
+    while (*p != 0 && *p != ')') p++;
+    *p = 0;
+
+    last_slash = Ustrrchr(server, '/');
+    last_dot = Ustrrchr(server, '.');
+
+    DEBUG(D_lookup) debug_printf("PGSQL new connection: socket=%s "
+      "database=%s user=%s\n", server, sdata[0], sdata[1]);
+
+    /* A valid socket name looks like this: /var/run/postgresql/.s.PGSQL.5432
+    We have to call PQsetdbLogin with '/var/run/postgresql' as the hostname
+    argument and put '5432' into the port variable. */
+
+    if (last_slash == NULL || last_dot == NULL)
       {
-      *errmsg = string_sprintf("incomplete PGSQL server data: %s", server_copy);
-      store_free(server_copy);
+      *errmsg = string_sprintf("PGSQL invalid filename for socket: %s",
+        server);
       *defer_break = TRUE;
       return DEFER;
       }
-    sdata[i] = p;
-    *pp++ = 0;
-    p = pp;
-    }
-  sdata[3] = p;    /* Final thing (password) is all the rest */
 
-  /* If host contains a colon, sort out the port number. */
+    /* Terminate the path name and set up the port: we'll have something like
+    server = "/var/run/postgresql" and port = "5432". */
 
-  if ((p = strchr(sdata[0], ':')) != NULL)
-    {
-    *p++ = 0;
-    sdata[4] = p;
+    *last_slash = 0;
+    port = last_dot + 1;
     }
+
+  /* Host connection; sort out the port */
+
   else
-    sdata[4]="";
+    {
+    uschar *p;
+    if ((p = Ustrchr(server, ':')) != NULL)
+      {
+      *p++ = 0;
+      port = p;
+      }
+
+    if (Ustrchr(server, '/') != NULL)
+      {
+      *errmsg = string_sprintf("unexpected slash in pgSQL server hostname: %s",
+        server);
+      *defer_break = TRUE;
+      return DEFER;
+      }
+
+    DEBUG(D_lookup) debug_printf("PGSQL new connection: host=%s port=%s "
+      "database=%s user=%s\n", server, port, sdata[0], sdata[1]);
+    }
 
   /* If the database is the empty string, set it NULL - the query must then
   define it. */
 
-  if (sdata[1][0] == 0) sdata[1] = NULL;
-
-  DEBUG(9) debug_printf("PGSQL new connection: host=%s port=%d database=%s "
-    "user=%s password=%s\n", sdata[0], port, sdata[1], sdata[2], sdata[3]);
+  if (sdata[0][0] == 0) sdata[0] = NULL;
 
   /* Get store for a new handle, initialize it, and connect to the server */
 
-                    /*  host       port  options tty   database  user      passwd */
-  pg_conn=PQsetdbLogin(sdata[0], sdata[4], NULL, NULL, sdata[1], sdata[2], sdata[3]);
+  pg_conn=PQsetdbLogin(
+    /*  host      port  options tty   database       user       passwd */
+    CS server, CS port,  NULL, NULL, CS sdata[0], CS sdata[1], CS sdata[2]);
 
   if(PQstatus(pg_conn) == CONNECTION_BAD)
     {
@@ -198,16 +259,24 @@ if (cn == NULL)
 
 else
   {
-  DEBUG(9) debug_printf("PGSQL using cached connection for %s\n", server);
+  DEBUG(D_lookup) debug_printf("PGSQL using cached connection for %s\n",
+    server_copy);
   }
 
 /* Run the query */
 
-  pg_result = PQexec(pg_conn, query);
+  pg_result = PQexec(pg_conn, CS query);
   switch(PQresultStatus(pg_result))
     {
     case PGRES_EMPTY_QUERY:
     case PGRES_COMMAND_OK:
+    /* The command was successful but did not return any data since it was
+     * not SELECT but either an INSERT, UPDATE or DELETE statement. */
+    result = string_copy(US PQcmdTuples(pg_result));
+    offset = Ustrlen(result);
+    DEBUG(D_lookup) debug_printf("PGSQL: command does not return any data "
+      "but was successful. Rows affected: %s\n", result);
+
     case PGRES_TUPLES_OK:
     break;
 
@@ -219,7 +288,9 @@ else
     }
 
 /* Result is in pg_result. Find the number of fields returned. If this is one,
-we don't add field names to the data. Otherwise we do. */
+we don't add field names to the data. Otherwise we do. If the query did not
+return anything we skip the for loop; this also applies to the case
+PGRES_COMMAND_OK. */
 
 num_fields = PQnfields(pg_result);
 num_tuples = PQntuples(pg_result);
@@ -230,12 +301,12 @@ row, we insert '\n' between them. */
 for (i = 0; i < num_tuples; i++)
   {
   if (result != NULL)
-    result = string_cat(result, &ssize, &offset, "\n", 1);
+    result = string_cat(result, &ssize, &offset, US"\n", 1);
 
    if (num_fields == 1)
     {
     result = string_cat(result, &ssize, &offset,
-      PQgetvalue(pg_result, i, 0), PQgetlength(pg_result, i, 0));
+      US PQgetvalue(pg_result, i, 0), PQgetlength(pg_result, i, 0));
     }
 
    else
@@ -243,30 +314,30 @@ for (i = 0; i < num_tuples; i++)
     int j;
     for (j = 0; j < num_fields; j++)
       {
-      char *tmp = PQfname(pg_result, j);
-      result = string_cat(result, &ssize, &offset, tmp, (int)strlen(tmp));
-      result = string_cat(result, &ssize, &offset, "=", 1);
+      uschar *tmp = US PQfname(pg_result, j);
+      result = string_cat(result, &ssize, &offset, tmp, Ustrlen(tmp));
+      result = string_cat(result, &ssize, &offset, US"=", 1);
 
       /* Quote the value if it contains spaces or is empty */
 
-      tmp = PQgetvalue(pg_result, i, j);
-      if (*tmp == 0 || (strchr(tmp, ' ') != NULL))
-	{
-	int k, t_len = (int)strlen(tmp);
-	result = string_cat(result, &ssize, &offset, "\"", 1);
-	for(k = 0; k < t_len; k++)
-	  {
-	  if (tmp[k] == '\"' || tmp[k] == '\\')
-	     result = string_cat(result, &ssize, &offset, "\\", 1);
-	  result = string_cat(result, &ssize, &offset, tmp+k, 1);
-	  }
-	result = string_cat(result, &ssize, &offset, "\"", 1);
-	}
+      tmp = US PQgetvalue(pg_result, i, j);
+      if (*tmp == 0 || (Ustrchr(tmp, ' ') != NULL))
+        {
+        int k, t_len = Ustrlen(tmp);
+        result = string_cat(result, &ssize, &offset, US"\"", 1);
+        for(k = 0; k < t_len; k++)
+          {
+          if (tmp[k] == '\"' || tmp[k] == '\\')
+             result = string_cat(result, &ssize, &offset, US"\\", 1);
+          result = string_cat(result, &ssize, &offset, tmp+k, 1);
+          }
+        result = string_cat(result, &ssize, &offset, US"\"", 1);
+        }
       else
-	{
-	result = string_cat(result, &ssize, &offset, tmp, (int)strlen(tmp));
-	}
-      result = string_cat(result, &ssize, &offset, " ", 1);
+        {
+        result = string_cat(result, &ssize, &offset, tmp, Ustrlen(tmp));
+        }
+      result = string_cat(result, &ssize, &offset, US" ", 1);
       }
     }
   }
@@ -278,7 +349,7 @@ always leaves enough room for a terminating zero. */
 if (result == NULL)
   {
   yield = FAIL;
-  *errmsg = "PGSQL: no data found";
+  *errmsg = US"PGSQL: no data found";
   }
 else result[offset] = 0;
 
@@ -300,7 +371,7 @@ if (result != NULL)
   }
 else
   {
-  DEBUG(9) debug_printf("%s\n", *errmsg);
+  DEBUG(D_lookup) debug_printf("%s\n", *errmsg);
   return yield;      /* FAIL or DEFER */
   }
 }
@@ -317,15 +388,15 @@ arguments are not used. Loop through a list of servers while the query is
 deferred with a retryable error. */
 
 int
-pgsql_find(void *handle, char *filename, char *query, int length,
-  char **result, char **errmsg)
+pgsql_find(void *handle, uschar *filename, uschar *query, int length,
+  uschar **result, uschar **errmsg)
 {
 int sep = 0;
-char *server;
-char *list = pgsql_servers;
-char buffer[512];
+uschar *server;
+uschar *list = pgsql_servers;
+uschar buffer[512];
 
-DEBUG(9) debug_printf("PGSQL query: %s\n", query);
+DEBUG(D_lookup) debug_printf("PGSQL query: %s\n", query);
 
 while ((server = string_nextinlist(&list, &sep, buffer, sizeof(buffer)))
         != NULL)
@@ -336,7 +407,7 @@ while ((server = string_nextinlist(&list, &sep, buffer, sizeof(buffer)))
   }
 
 if (pgsql_servers == NULL)
-  *errmsg = "no PGSQL servers defined (pgsql_servers option)";
+  *errmsg = US"no PGSQL servers defined (pgsql_servers option)";
 
 return DEFER;
 }
@@ -356,27 +427,32 @@ to allow escaping "on spec". If you use something like "where id="ab\%cd" it
 does treat the string as "ab%cd". So we can safely quote percent and
 underscore. [This is different to MySQL, where you can't do this.]
 
-Argument: the string to be quoted
-Returns:  the processed string
+Arguments:
+  s          the string to be quoted
+  opt        additional option text or NULL if none
+
+Returns:     the processed string or NULL for a bad option
 */
 
-char *
-pgsql_quote(char *s)
+uschar *
+pgsql_quote(uschar *s, uschar *opt)
 {
 register int c;
 int count = 0;
-char *t = s;
-char *quoted;
+uschar *t = s;
+uschar *quoted;
+
+if (opt != NULL) return NULL;     /* No options recognized */
 
 while ((c = *t++) != 0)
-  if (strchr("\n\t\r\b\'\"\\%_", c) != NULL) count++;
+  if (Ustrchr("\n\t\r\b\'\"\\%_", c) != NULL) count++;
 
 if (count == 0) return s;
-t = quoted = store_get((int)strlen(s) + count + 1);
+t = quoted = store_get(Ustrlen(s) + count + 1);
 
 while ((c = *s++) != 0)
   {
-  if (strchr("\n\t\r\b\'\"\\%_", c) != NULL)
+  if (Ustrchr("\n\t\r\b\'\"\\%_", c) != NULL)
     {
     *t++ = '\\';
     switch(c)
