@@ -2,10 +2,11 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 1999 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 #include "../exim.h"
+#include "lf_functions.h"
 #include "testdb.h"
 
 
@@ -21,7 +22,7 @@ the find function. */
 /* See local README for interface description. */
 
 void *
-testdb_open(char *filename, char **errmsg)
+testdb_open(uschar *filename, uschar **errmsg)
 {
 filename = filename;   /* Keep picky compilers happy */
 errmsg = errmsg;
@@ -37,21 +38,23 @@ return (void *)(1);    /* Just return something non-null */
 /* See local README for interface description. */
 
 int
-testdb_find(void *handle, char *filename, char *query, int length,
-  char **result, char **errmsg)
+testdb_find(void *handle, uschar *filename, uschar *query, int length,
+  uschar **result, uschar **errmsg)
 {
 handle = handle;          /* Keep picky compilers happy */
 filename = filename;
 length = length;
 
-if (strcmp(query, "fail") == 0)
+if (Ustrcmp(query, "fail") == 0)
   {
-  *errmsg = "testdb lookup forced FAIL";
+  *errmsg = US"testdb lookup forced FAIL";
+  DEBUG(D_lookup) debug_printf("%s\n", *errmsg);
   return FAIL;
   }
-if (strcmp(query, "defer") == 0)
+if (Ustrcmp(query, "defer") == 0)
   {
-  *errmsg = "testdb lookup forced DEFER";
+  *errmsg = US"testdb lookup forced DEFER";
+  DEBUG(D_lookup) debug_printf("%s\n", *errmsg);
   return DEFER;
   }
 *result = string_copy(query);
