@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Functions concerned with rewriting headers */
@@ -23,6 +23,28 @@ static char *rrname[] = {
   "  env-to"
 };
 
+/* Structure and table for finding source of address for debug printing */
+
+typedef struct where_list_block {
+  int bit;
+  uschar *string;
+} where_list_block;
+
+static where_list_block where_list[] = {
+  { rewrite_sender,  US"sender:" },
+  { rewrite_from,    US"from:" },
+  { rewrite_to,      US"to:" },
+  { rewrite_cc,      US"cc:" },
+  { rewrite_bcc,     US"bcc:" },
+  { rewrite_replyto, US"reply-to:" },
+  { rewrite_envfrom, US"env-from" },
+  { rewrite_envto,   US"env-to" },
+  { rewrite_smtp,    US"smtp recipient" },
+  { rewrite_smtp|rewrite_smtp_sender, US"smtp sender" }
+};
+
+static int where_list_size = sizeof(where_list)/sizeof(where_list_block);
+
 
 
 /*************************************************
@@ -30,7 +52,6 @@ static char *rrname[] = {
 *************************************************/
 
 /*
-
 Arguments:
   s              address to check
   is_recipient   TRUE if a recipient address; FALSE if a sender address
@@ -38,8 +59,8 @@ Arguments:
 Returns:         fully-qualified address
 */
 
-char *
-rewrite_address_qualify(char *s, BOOL is_recipient)
+uschar *
+rewrite_address_qualify(uschar *s, BOOL is_recipient)
 {
 return (parse_find_at(s) != NULL)? s :
   string_sprintf("%s@%s", s,
@@ -60,8 +81,8 @@ expanding a replacement, in case that involves file lookups.
 
 Arguments:
   s              address to rewrite
-  flag           indicates which header this address comes from; it must
-                   match the flags in the rewriting rule
+  flag           indicates where this address comes from; it must match the
+                   flags in the rewriting rule
   whole          if not NULL, set TRUE if any rewriting rule contained the
                    "whole" bit and it is a header that is being rewritten
   add_header     if TRUE and rewriting occurs, add an "X-rewrote-xxx" header
@@ -75,21 +96,16 @@ Returns:         new address if rewritten; the input address if no change;
                  rewritten address is returned, not just the active bit.
 */
 
-char *
-rewrite_one(char *s, int flag, BOOL *whole, BOOL add_header, char *name,
+uschar *
+rewrite_one(uschar *s, int flag, BOOL *whole, BOOL add_header, uschar *name,
   rewrite_rule *rewrite_rules)
 {
 rewrite_rule *rule;
-char *yield = s;
-char *subject = s;
-char *domain = NULL;
-BOOL uid_checked = FALSE;
-BOOL uid_set = FALSE;
+uschar *yield = s;
+uschar *subject = s;
+uschar *domain = NULL;
 BOOL done = FALSE;
-uid_t saved_euid = 0;
-gid_t saved_egid = 0;
 int rule_number = 1;
-int sep = UCHAR_MAX + 1;
 int yield_start = 0, yield_end = 0;
 
 if (whole != NULL) *whole = FALSE;
@@ -102,8 +118,8 @@ for (rule = rewrite_rules;
   {
   int start, end, pdomain;
   int count = 0;
-  char *save_localpart, *save_domain;
-  char *error, *new, *newparsed;
+  uschar *save_localpart, *save_domain;
+  uschar *error, *new, *newparsed;
 
   /* Ensure that the flag matches the flags in the rule. */
 
@@ -116,11 +132,22 @@ for (rule = rewrite_rules;
 
   /* If this is an SMTP-time rewrite, the pattern must be a regex and
   the subject may have any structure. No local part or domain variables
-  can be set for the expansion. */
+  can be set for the expansion. We expand the pattern in order to be consistent
+  with the other kinds of rewrite, where expansion happens inside
+  match_address_list(). */
 
   if ((flag & rewrite_smtp) != 0)
     {
-    if (!match_check_string(subject, rule->key, 0, TRUE, FALSE, NULL)) continue;
+    uschar *key = expand_string(rule->key);
+    if (key == NULL)
+      {
+      if (!expand_string_forcedfail)
+        log_write(0, LOG_MAIN|LOG_PANIC, "failed to expand \"%s\" while "
+          "checking for SMTP rewriting: %s", rule->key, expand_string_message);
+      continue;
+      }
+    if (match_check_string(subject, key, 0, TRUE, FALSE, FALSE, NULL) != OK)
+      continue;
     new = expand_string(rule->replacement);
     }
 
@@ -130,41 +157,25 @@ for (rule = rewrite_rules;
 
   else
     {
-    if (domain == NULL) domain = strrchr(subject, '@') + 1;
+    if (domain == NULL) domain = Ustrrchr(subject, '@') + 1;
 
     /* Use the general function for matching an address against a list (here
-    just one item, so set the separator to 0). */
+    just one item, so use the "impossible value" separator UCHAR_MAX+1). */
 
-    if (!match_address_list(subject, domain-subject, FALSE, &(rule->key), 0,
-      &sep, FALSE)) continue;
+    if (match_address_list(subject, FALSE, TRUE, &(rule->key), NULL, 0,
+        UCHAR_MAX + 1, NULL) != OK)
+      continue;
 
     /* The source address matches, and numerical variables have been
     set up. If the replacement string consists of precisely "*" then no
     rewriting is required for this address - the behaviour is as for "fail"
     in the replacement expansion, but assuming the quit flag. */
 
-    if (strcmp(rule->replacement, "*") == 0) break;
+    if (Ustrcmp(rule->replacement, "*") == 0) break;
 
-    /* Otherwise, expand the replacement string, changing the uid/gid first if
-    necessary. While doing the expansion, set $local_part and $domain to the
-    appropriate values, restoring whatever value they previously had afterwards.
-    */
-
-    if (!uid_checked)
-      {
-      uid_t uid = exim_uid_set? exim_uid : root_uid;
-      uid_checked = TRUE;
-      saved_euid = geteuid();
-      saved_egid = getegid();
-      if (saved_euid != uid)
-        {
-        if (saved_euid != root_uid) mac_seteuid(root_uid);
-        if (exim_gid_set) mac_setegid(exim_gid);
-        mac_seteuid(uid);
-        uid_set = TRUE;
-        DEBUG(2) debug_print_ids("rewrite_one");
-        }
-      }
+    /* Otherwise, expand the replacement string. Set $local_part and $domain to
+    the appropriate values, restoring whatever value they previously had
+    afterwards. */
 
     save_localpart = deliver_localpart;
     save_domain = deliver_domain;
@@ -198,97 +209,148 @@ for (rule = rewrite_rules;
     break;
     }
 
-  /* Validate what has been generated, qualifying if permitted to do so.
-  Only envelope from is permitted to be rewritten as <>.*/
-
-  if ((rule->flags & rewrite_qualify) != 0)
-    new = rewrite_address_qualify(new, TRUE);
+  /* Check the what has been generated is a valid RFC 2822 address. Only
+  envelope from or SMTP sender is permitted to be rewritten as <>.*/
 
   newparsed = parse_extract_address(new, &error, &start, &end, &pdomain,
-    flag == rewrite_envfrom);
+    flag == rewrite_envfrom || flag == (rewrite_smtp|rewrite_smtp_sender));
 
   if (newparsed == NULL)
     {
     log_write(0, LOG_MAIN|LOG_PANIC, "Rewrite of %s yielded unparseable "
       "address: %s in address %s", subject, error, new);
+    break;   /* Give up on this address */
     }
-  else if (pdomain == 0 && (flag != rewrite_envfrom || *newparsed != 0))
+
+  /* A non-null unqualified address can be qualified if requested. Otherwise,
+  this is an error unless it's the empty address in circumstances where that is
+  permitted. */
+
+  if (pdomain == 0 && (*newparsed != 0 ||
+      (flag != rewrite_envfrom && flag != (rewrite_smtp|rewrite_smtp_sender))))
     {
-    log_write(0, LOG_MAIN|LOG_PANIC, "Rewrite of %s yielded unqualified "
-      "address \"%s\"", subject, new);
-    }
-  else
-    {
-    DEBUG(2) debug_printf("\"%s\" rewritten as ", yield);
-
-    if (log_rewrites)
-      log_write(0, LOG_MAIN, "\"%s\" rewritten as \"%s\" by rule %d",
-        yield, new, rule_number);
-
-    /* A header will only actually be added if header_last is non-NULL,
-    i.e. during message reception or delivery, but add_header should not
-    be set TRUE during delivery, as otherwise multiple instances of the header
-    can fill up the -H file and make it embarrassingly large. We don't need
-    to set header_changed because the -H file always gets written at the end of
-    message reception. */
-
-    if (add_header)
-      header_add(htype_old, "X-rewrote-%s: %s\n", name, subject);
-
-    /* Handle the case when replacement of the whole address is possible.
-    This happens only when whole is not NULL and we are rewriting a header.
-    If *whole is already TRUE it means that a previous rule had the w
-    flag set and so we must preserve the non-active portion of the current
-    subject unless the current rule also has the w flag set. */
-
-    if (whole != NULL && (flag & rewrite_all_headers) != 0)
+    if ((rule->flags & rewrite_qualify) != 0)
       {
-      /* Current rule has the w flag set */
+      newparsed = rewrite_address_qualify(newparsed, TRUE);
+      new = string_sprintf("%.*s%s%.*s", start, new, newparsed,
+        Ustrlen(new) - end, new + end);
+      end = start + Ustrlen(newparsed);
+      }
+    else
+      {
+      log_write(0, LOG_MAIN|LOG_PANIC, "Rewrite of %s yielded unqualified "
+        "address \"%s\"", subject, new);
+      break;   /* Give up on this address */
+      }
+    }
 
-      if ((rule->flags & rewrite_whole) != 0)
+  /* We have a validly rewritten address */
+
+  if ((log_write_selector & L_address_rewrite) != 0 ||
+      (debug_selector & D_rewrite) != 0)
+    {
+    int i;
+    uschar *where = US"?";
+
+    for (i = 0; i < where_list_size; i++)
+      {
+      if (flag == where_list[i].bit)
         {
-        yield = new;
-        yield_start = start;
-        yield_end = end;
-        subject = newparsed;
-        *whole = TRUE;
-        }
-
-      /* Current rule does not have the w flag set; if not previously
-      done any whole rewriting, behave in non-whole manner. */
-
-      else if (!*whole) goto NEVER_WHOLE;
-
-      /* Current rule does not have the w flag set, but a previous
-      rule did rewrite the whole address. Thus yield and subject will be
-      different. Preserve the previous non-active part of the address. */
-
-      else
-        {
-        subject = newparsed;
-        new = string_sprintf("%.*s%s%n%s",
-           yield_start, yield, subject, &end, yield + yield_end);
-        yield_end = end;
-        yield = new;
+        where = where_list[i].string;
+        break;
         }
       }
+    log_write(L_address_rewrite,
+           LOG_MAIN, "\"%s\" from %s rewritten as \"%s\" by rule %d",
+           yield, where, new, rule_number);
+    }
 
-    /* Rule just rewrites active part, or handling an envelope. This
-    code is obeyed only when all rules so far have not done "whole"
-    replacement. */
+  /* A header will only actually be added if header_last is non-NULL,
+  i.e. during message reception or delivery, but add_header should not
+  be set TRUE during delivery, as otherwise multiple instances of the header
+  can fill up the -H file and make it embarrassingly large. We don't need
+  to set header_rewritten because the -H file always gets written at the end
+  of message reception. */
+
+  if (add_header)
+    header_add(htype_old, "X-rewrote-%s: %s\n", name, subject);
+
+  /* Handle the case when replacement of the whole address is possible.
+  This happens only when whole is not NULL and we are rewriting a header.
+  If *whole is already TRUE it means that a previous rule had the w
+  flag set and so we must preserve the non-active portion of the current
+  subject unless the current rule also has the w flag set. */
+
+  if (whole != NULL && (flag & rewrite_all_headers) != 0)
+    {
+    /* Current rule has the w flag set. We must ensure the phrase parts
+    are syntactically valid if they are present. */
+
+    if ((rule->flags & rewrite_whole) != 0)
+      {
+      if (start > 0 && new[start-1] == '<')
+        {
+        uschar *p1 = new + start - 1;
+        uschar *p2 = new + end + 1;
+        uschar *pf1, *pf2;
+        uschar buff1[256], buff2[256];
+
+        while (p1 > new && p1[-1] == ' ') p1--;
+        pf1 = parse_fix_phrase(new, p1 - new, buff1, sizeof(buff1));
+        while (*p2 == ' ') p2++;
+        pf2 = parse_fix_phrase(p2, Ustrlen(p2), buff2, sizeof(buff2));
+
+        /* Note that pf1 and pf2 are NOT necessarily buff1 and buff2. For
+        a non-RFC 2047 phrase that does not need to be RFC 2822 quoted, they
+        will be buff1+1 and buff2+1. */
+
+        start = Ustrlen(pf1) + start + new - p1;
+        end = start + Ustrlen(newparsed);
+        new = string_sprintf("%s%.*s%s", pf1, p2 - p1, p1, pf2);
+        }
+
+      /* Now accept the whole thing */
+
+      yield = new;
+      yield_start = start;
+      yield_end = end;
+      subject = newparsed;
+      *whole = TRUE;
+      }
+
+    /* Current rule does not have the w flag set; if not previously
+    done any whole rewriting, behave in non-whole manner. */
+
+    else if (!*whole) goto NEVER_WHOLE;
+
+    /* Current rule does not have the w flag set, but a previous
+    rule did rewrite the whole address. Thus yield and subject will be
+    different. Preserve the previous non-active part of the address. */
 
     else
       {
-      NEVER_WHOLE:
-      subject = yield = newparsed;
+      subject = newparsed;
+      new = string_sprintf("%.*s%s%n%s",
+         yield_start, yield, subject, &end, yield + yield_end);
+      yield_end = end;
+      yield = new;
       }
-
-    DEBUG(2) debug_printf("\"%s\"\n", yield);
-    domain = NULL;    /* Reset for next rule */
     }
 
-  /* If no further rewrites are to be done, set the done flag. This does
-  allow repeats of the current rule if configured. */
+  /* Rule just rewrites active part, or handling an envelope. This
+  code is obeyed only when all rules so far have not done "whole"
+  replacement. */
+
+  else
+    {
+    NEVER_WHOLE:
+    subject = yield = newparsed;
+    }
+
+  domain = NULL;    /* Reset for next rule */
+
+  /* If no further rewrites are to be done, set the done flag. This allows
+  repeats of the current rule if configured before breaking the loop. */
 
   if ((rule->flags & rewrite_quit) != 0) done = TRUE;
 
@@ -303,16 +365,8 @@ for (rule = rewrite_rules;
     }
   }
 
-/* Reset the uid/gid if necessary; free the original subject if it
-has been rewritten; unset expansion numeric variables, and that's it. */
+/* Unset expansion numeric variables, and that's it. */
 
-if (uid_set)
-  {
-  mac_seteuid(root_uid);
-  mac_setegid(saved_egid);
-  mac_seteuid(saved_euid);
-  DEBUG(2) debug_print_ids("restored");
-  }
 expand_nmax = -1;
 return yield;
 }
@@ -342,16 +396,16 @@ Arguments:
 Returns:         possibly rewritten address
 */
 
-char *
-rewrite_address(char *s, BOOL is_recipient, BOOL add_header,
+uschar *
+rewrite_address(uschar *s, BOOL is_recipient, BOOL add_header,
   rewrite_rule *rewrite_rules, int existflags)
 {
 int flag = is_recipient? rewrite_envto : rewrite_envfrom;
 s = rewrite_address_qualify(s, is_recipient);
 if ((existflags & flag) != 0)
   {
-  char *new = rewrite_one(s, flag, NULL, add_header, is_recipient?
-    "original-recipient" : "sender", rewrite_rules);
+  uschar *new = rewrite_one(s, flag, NULL, add_header, is_recipient?
+    US"original-recipient" : US"sender", rewrite_rules);
   if (new != s) s = new;
   }
 return s;
@@ -378,7 +432,7 @@ qualified addresses, but there are options to permit unqualified ones from
 certain hosts. For those hosts only, addresses in headers can also be
 qualified. For other hosts, unqualified addresses in headers do not get touched
 in any way. For locally sourced messages, unqualified addresses always get
-qualified.
+qualified, except when -bnq is used to explicitly suppress this.
 
 Arguments:
   h              pointer to header line block
@@ -395,29 +449,33 @@ Returns:         NULL if header unchanged; otherwise the rewritten header
 */
 
 static header_line *
-rewrite_one_header(header_line *h, int flag, char *routed_old,
-  char *routed_new, rewrite_rule *rewrite_rules, int existflags, BOOL replace)
+rewrite_one_header(header_line *h, int flag, uschar *routed_old,
+  uschar *routed_new, rewrite_rule *rewrite_rules, int existflags, BOOL replace)
 {
 int lastnewline = 0;
 header_line *newh = NULL;
-char *s = strchr(h->text, ':') + 1;
-while (isspace((uschar)*s)) s++;
+void *function_reset_point = store_get(0);
+uschar *s = Ustrchr(h->text, ':') + 1;
+while (isspace(*s)) s++;
 
-DEBUG(5) debug_printf("rewrite_one_header: type=%c:\n  %s", h->type, h->text);
+DEBUG(D_rewrite)
+  debug_printf("rewrite_one_header: type=%c:\n  %s", h->type, h->text);
 
 parse_allow_group = TRUE;     /* Allow group syntax */
 
 /* Loop for multiple addresses in the header. We have to go through them all
 in case any need qualifying, even if there's no rewriting. Pathological headers
 may have thousands of addresses in them, so cause the store to be reset for
-any that don't actually get rewritten. */
+any that don't actually get rewritten. We also play silly games for those that
+_are_ rewritten so as to avoid runaway store usage for these kinds of header.
+We want to avoid keeping store for any intermediate versions. */
 
 while (*s != 0)
   {
-  char *sprev;
-  char *ss = parse_find_address_end(s, FALSE);
-  char *recipient, *new, *errmess;
-  void *reset_point = store_get(0);
+  uschar *sprev;
+  uschar *ss = parse_find_address_end(s, FALSE);
+  uschar *recipient, *new, *errmess;
+  void *loop_reset_point = store_get(0);
   BOOL changed = FALSE;
   int terminator = *ss;
   int start, end, domain;
@@ -431,12 +489,16 @@ while (*s != 0)
   *ss = terminator;
   sprev = s;
   s = ss + (terminator? 1:0);
-  while (isspace((uschar)*s)) s++;
+  while (isspace(*s)) s++;
 
   /* There isn't much we can do for syntactic disasters at this stage.
   Pro tem (possibly for ever) ignore them. */
 
-  if (recipient == NULL) continue;
+  if (recipient == NULL)
+    {
+    store_reset(loop_reset_point);
+    continue;
+    }
 
   /* If routed_old is not NULL, this is a rewrite caused by a router,
   consisting of changing routed_old into routed_new, and applying to all
@@ -453,7 +515,7 @@ while (*s != 0)
     if (domain <= 0 || strcmpic(recipient+domain, routed_old) != 0) continue;
     recipient[domain-1] = 0;
     new = string_sprintf("%s@%s", recipient, routed_new);
-    DEBUG(2)
+    DEBUG(D_rewrite)
       {
       recipient[domain-1] = '@';
       debug_printf("%s rewritten by router as %s\n", recipient, new);
@@ -480,7 +542,10 @@ while (*s != 0)
 
     if (changed && ((is_recipient && !allow_unqualified_recipient) ||
                     (!is_recipient && !allow_unqualified_sender)))
+      {
+      store_reset(loop_reset_point);
       continue;
+      }
     }
 
   /* If there are rewrite rules for this type of header, apply
@@ -505,24 +570,34 @@ while (*s != 0)
       }
     }
 
+  /* If nothing has changed, lose all dynamic store obtained in this loop, and
+  move on to the next address. We can't reset to the function start store
+  point, because we may have a rewritten line from a previous time round the
+  loop. */
+
+  if (!changed) store_reset(loop_reset_point);
+
   /* If the address has changed, create a new header containing the
   rewritten address. We do not need to set the chain pointers at this
-  stage. */
+  stage. We want to avoid using more and more memory if the header is very long
+  and contains lots and lots of rewritten addresses. Therefore, we build the
+  new text string in malloc store, then at the end we reset dynamic store
+  before copying the new header to a new block (and then freeing the malloc
+  block). The header must end up in dynamic store so that it's freed at the end
+  of receiving a message. */
 
-  if (changed)
+  else
     {
     int remlen;
-    int newlen = (int)strlen(new);
+    int newlen = Ustrlen(new);
     int oldlen = end - start;
 
     header_line *prev = (newh == NULL)? h : newh;
-    header_line *newnewh =
-      store_get(sizeof(header_line) + prev->slen - oldlen + newlen + 4);
-    char *newt = store_get(prev->slen - oldlen + newlen + 4);
+    uschar *newt = store_malloc(prev->slen - oldlen + newlen + 4);
+    uschar *newtstart = newt;
 
-    newnewh->text = newt;
-    newnewh->type = prev->type;
-    newnewh->slen = prev->slen - oldlen + newlen;
+    int type = prev->type;
+    int slen = prev->slen - oldlen + newlen;
 
     /* Build the new header text by copying the old and putting in the
     replacement. This process may make the header substantially longer
@@ -533,15 +608,15 @@ while (*s != 0)
     newlines in the header, but it doesn't seem worth going to that
     amount of trouble. */
 
-    strncpy(newt, prev->text, sprev - prev->text + start);
+    Ustrncpy(newt, prev->text, sprev - prev->text + start);
     newt += sprev - prev->text + start;
     *newt = 0;
-    strcat(newt, new);
+    Ustrcat(newt, new);
     newt += newlen;
     remlen = s - (sprev + end);
     if (remlen > 0)
       {
-      strncpy(newt, sprev + end, remlen);
+      Ustrncpy(newt, sprev + end, remlen);
       newt += remlen;
       *newt = 0;
       }
@@ -551,36 +626,47 @@ while (*s != 0)
     another one! The pointer s has been skipped over white space, so just
     look back to see if the last non-space-or-tab was a newline. */
 
-    if (newlen > oldlen && newt - newnewh->text - lastnewline > 40)
+    if (newlen > oldlen && newt - newtstart - lastnewline > 40)
       {
-      char *p = s - 1;
+      uschar *p = s - 1;
       while (p >= prev->text && (*p == ' ' || *p == '\t')) p--;
       if (*p != '\n')
         {
-        lastnewline = newt - newnewh->text;
-        strcat(newt, "\n\t");
-        newnewh->slen += 2;
+        lastnewline = newt - newtstart;
+        Ustrcat(newt, "\n\t");
+        slen += 2;
         }
       }
 
     /* Finally, the remaining unprocessed addresses, if any. */
 
-    strcat(newt, s);
+    Ustrcat(newt, s);
 
-    DEBUG(5) debug_printf("newlen=%d newtype=%c newtext:\n%s",
-      newnewh->slen, newnewh->type, newnewh->text);
+    DEBUG(D_rewrite) debug_printf("newlen=%d newtype=%c newtext:\n%s",
+      slen, type, newtstart);
 
-    s = newnewh->text + (s - prev->text) - oldlen + newlen;
+    /* Compute the length of the rest of the header line before we possibly
+    flatten a previously rewritten copy. */
 
-    DEBUG(5) debug_printf("remainder: %s", (*s == 0)? "\n" : s);
+    remlen = (s - prev->text) - oldlen + newlen;
 
-    newh = newnewh;
+    /* We have the new text in a malloc block. That enables us to release all
+    the memory that has been used, back to the point at which the function was
+    entered. Then set up a new header in dynamic store. This will override a
+    rewritten copy from a previous time round this loop. */
+
+    store_reset(function_reset_point);
+    newh = store_get(sizeof(header_line));
+    newh->type = type;
+    newh->slen = slen;
+    newh->text = string_copyn(newtstart, slen);
+    store_free(newtstart);
+
+    /* Set up for scanning the rest of the header */
+
+    s = newh->text + remlen;
+    DEBUG(D_rewrite) debug_printf("remainder: %s", (*s == 0)? US"\n" : s);
     }
-
-  /* If the address was not changed, reset so as to reuse the store it
-  was extracted into. */
-
-  else store_reset(reset_point);
   }
 
 parse_allow_group = FALSE;  /* Reset group flags */
@@ -631,7 +717,7 @@ Returns:         NULL if header unchanged; otherwise the rewritten header
 */
 
 header_line *
-rewrite_header(header_line *h, char *routed_old, char *routed_new,
+rewrite_header(header_line *h, uschar *routed_old, uschar *routed_new,
   rewrite_rule *rewrite_rules, int existflags, BOOL replace)
 {
 switch (h->type)
@@ -656,7 +742,7 @@ switch (h->type)
   return rewrite_one_header(h, rewrite_bcc, routed_old, routed_new,
     rewrite_rules, existflags, replace);
 
-  case htype_replyto:
+  case htype_reply_to:
   return rewrite_one_header(h, rewrite_replyto, routed_old, routed_new,
     rewrite_rules, existflags, replace);
   }
@@ -677,9 +763,9 @@ Argument: the address to test
 Returns:  nothing
 */
 
-void rewrite_test(char *s)
+void rewrite_test(uschar *s)
 {
-char *recipient, *error;
+uschar *recipient, *error;
 int i, start, end, domain;
 BOOL done_smtp = FALSE;
 
@@ -689,22 +775,26 @@ if (rewrite_existflags == 0)
   return;
   }
 
-/* Do SMTP rewrite only if a rule with the S flag exists. */
+/* Do SMTP rewrite only if a rule with the S flag exists. Allow <> by
+pretending it is a sender. */
 
 if ((rewrite_existflags & rewrite_smtp) != 0)
   {
-  char *new = rewrite_one(s, rewrite_smtp, NULL, FALSE, "",
-    global_rewrite_rules);
+  uschar *new = rewrite_one(s, rewrite_smtp|rewrite_smtp_sender, NULL, FALSE,
+    US"", global_rewrite_rules);
   if (new != s)
     {
-    printf("    SMTP: %s\n", new);
+    if (*new == 0)
+      printf("    SMTP: <>\n");
+    else
+      printf("    SMTP: %s\n", new);
     done_smtp = TRUE;
     }
   }
 
 /* Do the other rewrites only if a rule without the S flag exists */
 
-if ((rewrite_existflags & ~(rewrite_smtp|rewrite_X)) == 0) return;
+if ((rewrite_existflags & ~rewrite_smtp) == 0) return;
 
 /* Qualify if necessary before extracting the address */
 
@@ -724,20 +814,14 @@ for (i = 0; i < 8; i++)
   {
   BOOL whole = FALSE;
   int flag = 1 << i;
-  char *new = rewrite_one(recipient, flag, &whole, FALSE, "",
+  uschar *new = rewrite_one(recipient, flag, &whole, FALSE, US"",
     global_rewrite_rules);
   printf("%s: ", rrname[i]);
-  if (whole || (flag & rewrite_all_headers) == 0) printf("%s\n", new);
-    else printf("%.*s%s%s\n", start, s, new, s+end);
-  }
-
-/* Only show the X case if there is a rewrite rule for it. */
-
-if ((rewrite_existflags & rewrite_X) != 0)
-  {
-  char *new = rewrite_one(recipient, rewrite_X, NULL, FALSE, "",
-    global_rewrite_rules);
-  if (new != recipient) printf("       X: %s\n", new);
+  if (*new == 0)
+    printf("<>\n");
+  else if (whole || (flag & rewrite_all_headers) == 0)
+    printf("%s\n", CS new);
+  else printf("%.*s%s%s\n", start, s, new, s+end);
   }
 }
 
