@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* A set of functions to search databases in various formats. An open
@@ -16,8 +16,10 @@ However, if too many files are opened, some of those that are not in use have
 to be closed. Those open items that use real files are kept on a LRU chain to
 help with this.
 
-All the data is held in malloc'd store so as to be independent of the stacking
-store pools. */
+All the data is held in permanent store so as to be independent of the stacking
+pool that is reset from time to time. In fact, we use malloc'd store so that it
+can be freed when the caches are tidied up. It isn't actually clear whether
+this is a benefit or not, to be honest. */
 
 #include "exim.h"
 
@@ -40,130 +42,156 @@ static int open_filecount = 0;
 
 
 /*************************************************
-*        Validate a lookup type name             *
+*      Validate a plain lookup type name         *
 *************************************************/
 
 /* Only those names that are recognized and whose code is included in the
-binary give an OK response.
+binary give an OK response. Use a binary chop search now that the list has got
+so long.
 
 Arguments:
-  name       lookup type name
-  errmsg     where to put an error message
+  name       lookup type name - not necessarily zero terminated (e.g. dbm*)
+  len        length of the name
 
 Returns:     +ve => valid lookup name; value is offset in lookup_list
-             -ve => invalid name; message in errmsg.
+             -ve => invalid name; message in search_error_message.
 */
 
 int
-search_findtype(char *name, char **errmsg)
+search_findtype(uschar *name, int len)
 {
-int n = 0;
-lookup_info *li;
-
-for (li = lookup_list; li->name[0] != 0; n++,li++)
+int bot = 0;
+int top = lookup_list_count;
+while (top > bot)
   {
-  if (strcmp(name, li->name) == 0)
+  int mid = (top + bot)/2;
+  int c = Ustrncmp(name, lookup_list[mid].name, len);
+
+  /* If c == 0 we have matched the incoming name with the start of the search
+  type name. However, some search types are substrings of others (e.g. nis and
+  nisplus) so we need to check that the lengths are the same. The length of the
+  type name cannot be shorter (else c would not be 0); if it is not equal it
+  must be longer, and in that case, the incoming name comes before the name we
+  are testing. By leaving c == 0 when the lengths are different, and doing a
+  > 0 test below, this all falls out correctly. */
+
+  if (c == 0 && Ustrlen(lookup_list[mid].name) == len)
     {
-    if (li->find != NULL) return n;
-    *errmsg = string_sprintf("lookup type \"%s\" is not available "
-    "(not in the binary - check buildtime LOOKUP configuration)", name);
+    if (lookup_list[mid].find != NULL) return mid;
+    search_error_message  = string_sprintf("lookup type \"%.*s\" is not "
+      "available (not in the binary - check buildtime LOOKUP configuration)",
+      len, name);
     return -1;
     }
+
+  if (c > 0) bot = mid + 1; else top = mid;
   }
 
-*errmsg = string_sprintf("unknown lookup type \"%s\"", name);
+search_error_message = string_sprintf("unknown lookup type \"%.*s\"",len,name);
 return -1;
 }
 
 
 
 /*************************************************
-*         Check a file's credentials             *
+*       Validate a full lookup type name         *
 *************************************************/
 
-/* fstat can normally be expected to work on an open file, but there are some
-NFS states where it may not.
+/* This function recognizes the "partial-" prefix and also terminating * and *@
+suffixes.
 
 Arguments:
-  fd         an open file descriptor or -1
-  filename   a file name if fd is -1
-  modemask   a mask specifying mode bits that must *not* be set
-  owners     NULL or a list of of allowable uids, count in the first item
-  owngroups  NULL or a list of allowable gids, count in the first item
-  type       name of lookup type for putting in error message
-  errmsg     where to point an error message
+  name         the full lookup type name
+  ptypeptr     where to put the partial type
+                 after subtraction of 1024 or 2048:
+                   negative     => no partial matching
+                   non-negative => minimum number of non-wild components
+  ptypeaff     where to put a pointer to the affix
+                 the affix is within name if supplied therein
+                 otherwise it's a literal string
+  afflen       the length of the affix
+  starflags    where to put the SEARCH_STAR and SEARCH_STARAT flags
 
-Returns:     -1 stat() or fstat() failed
-              0 OK
-             +1 something didn't match
-
-Side effect: sets errno to ERRNO_BADUGID, ERRNO_NOTREGULAR or ERRNO_BADMODE for
-             bad uid/gid, not a regular file, or bad mode; otherwise leaves it
-             to what fstat set it to.
+Returns:     +ve => valid lookup name; value is offset in lookup_list
+             -ve => invalid name; message in search_error_message.
 */
 
 int
-search_check_file(int fd, char *filename, int modemask, uid_t *owners,
-  gid_t *owngroups, char *type, char **errmsg)
+search_findtype_partial(uschar *name, int *ptypeptr, uschar **ptypeaff,
+  int *afflen, int *starflags)
 {
-int i;
-struct stat statbuf;
+int len, stype;
+int pv = -1;
+uschar *ss = name;
 
-if ((fd >= 0 && fstat(fd, &statbuf) != 0) ||
-    (fd  < 0 && stat(filename, &statbuf) != 0))
+*starflags = 0;
+*ptypeaff = NULL;
+
+/* Check for a partial matching type. It must start with "partial", optionally
+followed by a sequence of digits. If this is followed by "-", the affix is the
+default "*." string. Otherwise we expect an affix in parentheses. Affixes are a
+limited number of characters, not including parens. */
+
+if (Ustrncmp(name, "partial", 7) == 0)
   {
-  int save_errno = errno;
-  *errmsg = string_sprintf("%s: stat failed", filename);
-  errno = save_errno;
+  ss += 7;
+  if (isdigit (*ss))
+    {
+    pv = 0;
+    while (isdigit(*ss)) pv = pv*10 + *ss++ - '0';
+    }
+  else pv = 2;         /* Default number of wild components */
+
+  if (*ss == '(')
+    {
+    *ptypeaff = ++ss;
+    while (ispunct(*ss) && *ss != ')') ss++;
+    if (*ss != ')') goto BAD_TYPE;
+    *afflen = ss++ - *ptypeaff;
+    }
+  else if (*ss++ == '-')
+    {
+    *ptypeaff = US "*.";
+    *afflen = 2;
+    }
+  else
+    {
+    BAD_TYPE:
+    search_error_message = string_sprintf("format error in lookup type \"%s\"",
+      name);
+    return -1;
+    }
+  }
+
+/* Now we are left with a lookup name, possibly followed by * or *@. */
+
+len = Ustrlen(ss);
+if (len >= 2 && Ustrncmp(ss + len - 2, "*@", 2) == 0)
+  {
+  *starflags |= SEARCH_STARAT;
+  len -= 2;
+  }
+else if (len >= 1 && ss[len-1]  == '*')
+  {
+  *starflags |= SEARCH_STAR;
+  len--;
+  }
+
+/* Check for the individual search type. Only those that are actually in the
+binary are valid. For query-style types, "partial" is an error. */
+
+stype = search_findtype(ss, len);
+if (pv >= 0 && mac_islookup(stype, lookup_querystyle))
+  {
+  search_error_message = string_sprintf("\"partial\" is not permitted "
+    "for lookup type \"%s\"", ss);
   return -1;
   }
 
-if ((statbuf.st_mode & S_IFMT) != S_IFREG)
-  {
-  *errmsg = string_sprintf("%s is not a regular file (%s lookup)",
-    filename, type);
-  errno = ERRNO_NOTREGULAR;
-  return +1;
-  }
+/* All is well; pass back the partial type and return the lookup type. */
 
-if ((statbuf.st_mode & modemask) != 0)
-  {
-  *errmsg = string_sprintf("%s (%s lookup): file mode %.4o should not contain "
-    "%.4o", filename, type,  statbuf.st_mode & 07777,
-    statbuf.st_mode & modemask);
-  errno = ERRNO_BADMODE;
-  return +1;
-  }
-
-if (owners != NULL)
-  {
-  BOOL uid_ok = FALSE;
-  for (i = 1; i <= (int)owners[0]; i++)
-    if (owners[i] == statbuf.st_uid) { uid_ok = TRUE; break; }
-  if (!uid_ok)
-    {
-    *errmsg = string_sprintf("%s (%s lookup): file has wrong owner", filename,
-      type);
-    errno = ERRNO_BADUGID;
-    return +1;
-    }
-  }
-
-if (owngroups != NULL)
-  {
-  BOOL gid_ok = FALSE;
-  for (i = 1; i <= (int)owngroups[0]; i++)
-    if (owngroups[i] == statbuf.st_gid) { gid_ok = TRUE; break; }
-  if (!gid_ok)
-    {
-    *errmsg = string_sprintf("%s (%s lookup): file has wrong group", filename,
-      type);
-    errno = ERRNO_BADUGID;
-    return +1;
-    }
-  }
-
-return 0;
+*ptypeptr = pv;
+return stype;
 }
 
 
@@ -215,9 +243,9 @@ Returns:  nothing
 void
 search_tidyup(void)
 {
-lookup_info *li;
+int i;
 
-DEBUG(9) debug_printf("search_tidyup called\n");
+DEBUG(D_lookup) debug_printf("search_tidyup called\n");
 
 /* Close individually each cached open file. */
 
@@ -231,8 +259,8 @@ open_filecount = 0;
 
 /* Call the general tidyup entry for any drivers that have one. */
 
-for (li = lookup_list; li->name[0] != 0; li++)
-  if (li->tidy != NULL) (li->tidy)();
+for (i = 0; i < lookup_list_count; i++)
+  if (lookup_list[i].tidy != NULL) (lookup_list[i].tidy)();
 }
 
 
@@ -279,44 +307,43 @@ Arguments:
                  owners of the file; otherwise it is ignored
   owngroups      if a real single file is used, this specifies the possible
                  group owners of the file; otherwise it is ignored
-  message        points to a char * which can be set to point to an error
-                 message when something goes wrong
 
 Returns:         an identifying handle for the open database;
                  this is the pointer to the tree block in the
-                 cache of open files; return NULL on open failure
+                 cache of open files; return NULL on open failure, with
+                 a message in search_error_message
 */
 
 void *
-search_open(char *filename, int search_type, int modemask, uid_t *owners,
-  gid_t *owngroups, char **message)
+search_open(uschar *filename, int search_type, int modemask, uid_t *owners,
+  gid_t *owngroups)
 {
 void *handle;
 tree_node *t;
 search_cache *c;
 lookup_info *lk = lookup_list + search_type;
-char keybuffer[256];
+uschar keybuffer[256];
 
-DEBUG(9) debug_printf("search_open: %s \"%s\"\n", lk->name,
-  (filename == NULL)? "NULL" : filename);
+DEBUG(D_lookup) debug_printf("search_open: %s \"%s\"\n", lk->name,
+  (filename == NULL)? US"NULL" : filename);
 
 /* See if we already have this open for this type of search, and if so,
 pass back the tree block as the handle. The key for the tree node is the search
 type plus '0' concatenated with the file name. There may be entries in the tree
 with closed files if a lot of files have been opened. */
 
-sprintf(keybuffer, "%c%.254s", search_type + '0',
-  (filename == NULL)? "" : filename);
+sprintf(CS keybuffer, "%c%.254s", search_type + '0',
+  (filename == NULL)? US"" : filename);
 
 if ((t = tree_search(search_tree, keybuffer)) != NULL)
   {
   c = (search_cache *)(t->data.ptr);
   if (c->handle != NULL)
     {
-    DEBUG(9) debug_printf("  cached open\n");
+    DEBUG(D_lookup) debug_printf("  cached open\n");
     return t;
     }
-  DEBUG(9) debug_printf("  cached closed\n");
+  DEBUG(D_lookup) debug_printf("  cached closed\n");
   }
 
 /* Otherwise, we need to open the file or database - each search type has its
@@ -333,7 +360,7 @@ if (lk->type == lookup_absfile && open_filecount >= lookup_open_max)
   else
     {
     search_cache *c = (search_cache *)(open_bot->data.ptr);
-    DEBUG(9) debug_printf("Too many lookup files open\n  closing %s\n",
+    DEBUG(D_lookup) debug_printf("Too many lookup files open\n  closing %s\n",
       open_bot->name);
     open_bot = c->up;
     if (open_bot != NULL)
@@ -349,11 +376,12 @@ if (lk->type == lookup_absfile && open_filecount >= lookup_open_max)
 /* If opening is successful, call the file-checking function if there is one,
 and if all is still well, enter the open database into the tree. */
 
-handle = lk->open(filename, message);
+handle = lk->open(filename, &search_error_message);
 if (handle == NULL) return handle;
 
 if (lk->check != NULL &&
-   !lk->check(handle, filename, modemask, owners, owngroups, message))
+   !lk->check(handle, filename, modemask, owners, owngroups,
+     &search_error_message))
   {
   lk->close(handle);
   return NULL;
@@ -369,11 +397,11 @@ count alone. */
 
 if (t == NULL)
   {
-  t = store_malloc(sizeof(tree_node) + (int)strlen(keybuffer));
+  t = store_malloc(sizeof(tree_node) + Ustrlen(keybuffer));
   t->data.ptr = c = store_malloc(sizeof(search_cache));
   c->key = NULL;
   c->data = NULL;
-  strcpy(t->name, keybuffer);
+  Ustrcpy(t->name, keybuffer);
   tree_insertnode(&search_tree, t);
   }
 else c = t->data.ptr;
@@ -393,11 +421,8 @@ return t;
 *  Internal function: Find one item in database  *
 *************************************************/
 
-/* The answer is always put into dynamic store. If the key contains a colon,
-then it is treated as a double key: the first part is the key for the record in
-the file, and the remainder is a subkey that is used to extract a subfield from
-the main data. Subfields are specified as subkey=value in the records. The last
-lookup for each handle is cached.
+/* The answer is always put into dynamic store. The last lookup for each handle
+is cached.
 
 Arguments:
   handle       the handle from search_open; points to tree node
@@ -405,42 +430,28 @@ Arguments:
                NULL for query-style searches
   keystring    the keystring for single-key+file lookups, or
                the querystring for query-style lookups
-  errmsg       somewhere to point to an error message
 
 Returns:       a pointer to a dynamic string containing the answer,
                or NULL if the query failed or was deferred; in the
-               latter case, search_find_defer is set TRUE
+               latter case, search_find_defer is set TRUE; after an unusual
+               failure, there may be a message in search_error_message.
 */
 
-static char *
-internal_search_find(void *handle, char *filename, char *keystring,
-  char **errmsg)
+static uschar *
+internal_search_find(void *handle, uschar *filename, uschar *keystring)
 {
 tree_node *t = (tree_node *)handle;
 search_cache *c = (search_cache *)(t->data.ptr);
 int search_type = t->name[0] - '0';
 
-char *subkey = NULL;
-char *colon = NULL;
+/* Lookups that return DEFER may not always set an error message. So that
+the callers don't have to test for NULL, set an empty string. */
 
-*errmsg = "";              /* indicates no error, or ordinary failed to find */
+search_error_message = US"";
 search_find_defer = FALSE;
 
-/* If there's a colon in a key for a single-key+file lookup, temporarily
-terminate the main key, and set up the subkey. */
-
-if (filename != NULL)
-  {
-  colon = strchr(keystring, ':');
-  if (colon != NULL)
-    {
-    subkey = colon + 1;
-    *colon = 0;
-    }
-  }
-
-DEBUG(9) debug_printf("internal_search_find: file=\"%s\"\n  "
-  "type=%s key=\"%s\"\n", (filename == NULL)? "NULL" : filename,
+DEBUG(D_lookup) debug_printf("internal_search_find: file=\"%s\"\n  "
+  "type=%s key=\"%s\"\n", filename,
   lookup_list[search_type].name, keystring);
 
 /* Insurance. If the keystring is empty, just fail. */
@@ -450,19 +461,16 @@ if (keystring[0] == 0) return NULL;
 /* Look up the data for the key, unless it is already in the cache for this
 file. */
 
-if (c->key == NULL || strcmp(c->key, keystring) != 0)
+if (c->key == NULL || Ustrcmp(c->key, keystring) != 0)
   {
-  char *data = NULL;
-  int keylength = (int)strlen(keystring);
+  uschar *data = NULL;
+  int keylength = Ustrlen(keystring);
 
-  DEBUG(9)
+  DEBUG(D_lookup)
     {
     if (filename != NULL)
-      debug_printf("file lookup required for %s%s%s\n  in %s\n",
-        keystring,
-        (subkey == NULL)? "" : ":",
-        (subkey == NULL)? "" : subkey,
-        filename);
+      debug_printf("file lookup required for %s\n  in %s\n",
+        keystring, filename);
     else
       debug_printf("database lookup required for %s\n", keystring);
     }
@@ -478,7 +486,7 @@ if (c->key == NULL || strcmp(c->key, keystring) != 0)
   distinguish if necessary. */
 
   if (lookup_list[search_type].find(c->handle, filename, keystring, keylength,
-      &data, errmsg) == DEFER)
+      &data, &search_error_message) == DEFER)
     {
     search_find_defer = TRUE;
     }
@@ -493,42 +501,36 @@ if (c->key == NULL || strcmp(c->key, keystring) != 0)
     memcpy(c->key, keystring, len);
     if (data != NULL)
       {
-      len = (int)strlen(data) + 1;
+      len = Ustrlen(data) + 1;
       c->data = store_malloc(len);
       memcpy(c->data, data, len);
       }
     }
   }
 
-else DEBUG(9) debug_printf("cached data used for lookup of %s%s%s%s%s\n",
+/* Data was in the cache already */
+
+else DEBUG(D_lookup) debug_printf("cached data used for lookup of %s%s%s\n",
   keystring,
-  (subkey == NULL)? "" : ":",
-  (subkey == NULL)? "" : subkey,
-  (filename == NULL)? "" : "\n  in ",
-  (filename == NULL)? "" : filename);
+  (filename == NULL)? US"" : US"\n  in ",
+  (filename == NULL)? US"" : filename);
 
-/* Debug: output the main record */
+/* Debug: output the answer */
 
-DEBUG(9)
+DEBUG(D_lookup)
   {
   if (c->data == NULL)
     {
-    if (search_find_defer) debug_printf("lookup deferred\n");
-      else debug_printf("lookup failed\n");
+    if (search_find_defer) debug_printf("lookup deferred: %s\n",
+      search_error_message);
+    else debug_printf("lookup failed\n");
     }
   else debug_printf("lookup yielded: %s\n", c->data);
   }
 
-/* Put back the colon if it was overwritten */
+/* Return it in new dynamic store */
 
-if (colon != NULL) *colon = ':';
-
-/* If we have found data, pick out the subfield if required. Otherwise
-return the whole cached string. */
-
-return (c->data == NULL)? NULL :
-  (subkey != NULL)?
-    expand_getkeyed(subkey, c->data) : string_copy(c->data);
+return (c->data == NULL)? NULL : string_copy(c->data);
 }
 
 
@@ -542,39 +544,40 @@ return (c->data == NULL)? NULL :
 is no partial matching, but repeatedly when partial matching is requested.
 
 Arguments:
-  handle       the handle from search_open
-  filename     the filename that was handed to search_open, or
-               NULL for query-style searches
-  keystring    the keystring for single-key+file lookups, or
-               the querystring for query-style lookups
-  partial      if the value is greater than 2000, subtract 2048 and
-               set the "try * to the left of @" flag; then
-               if the value is greater than 1000, subtract 1024 to
-               get the real value, and set the "try '*' if all else
-               fails" flag; after this,
-               -1 means no partial matching;
-               otherwise it's the minimum number of components;
-  expand_setup pointer to offset for setting up expansion strings;
-               don't do any if < 0
-  errmsg       somewhere to point to an error message
+  handle         the handle from search_open
+  filename       the filename that was handed to search_open, or
+                   NULL for query-style searches
+  keystring      the keystring for single-key+file lookups, or
+                   the querystring for query-style lookups
+  partial        -1 means no partial matching;
+                   otherwise it's the minimum number of components;
+  affix          the affix string for partial matching
+  affixlen       the length of the affix string
+  starflags      SEARCH_STAR and SEARCH_STARAT flags
+  expand_setup   pointer to offset for setting up expansion strings;
+                 don't do any if < 0
 
-Returns:       a pointer to a dynamic string containing the answer,
-               or NULL if the query failed or was deferred; in the
-               latter case, search_find_defer is set TRUE
+Returns:         a pointer to a dynamic string containing the answer,
+                 or NULL if the query failed or was deferred; in the
+                 latter case, search_find_defer is set TRUE
 */
 
-char *
-search_find(void *handle, char *filename, char *keystring, int partial,
-  int *expand_setup, char **errmsg)
+uschar *
+search_find(void *handle, uschar *filename, uschar *keystring, int partial,
+  uschar *affix, int affixlen, int starflags, int *expand_setup)
 {
 tree_node *t = (tree_node *)handle;
 BOOL set_null_wild = FALSE;
-BOOL do_star = FALSE;
-BOOL do_at = FALSE;
-char *yield;
+uschar *yield;
 
-DEBUG(9) debug_printf("search_find: file=\"%s\"\n  key=\"%s\" partial=%d\n",
-  (filename == NULL)? "NULL" : filename, keystring, partial);
+DEBUG(D_lookup)
+  {
+  if (partial < 0) affixlen = 99;   /* So that "NULL" prints */
+  debug_printf("search_find: file=\"%s\"\n  key=\"%s\" "
+    "partial=%d affix=%.*s starflags=%x\n",
+    (filename == NULL)? US"NULL" : filename,
+    keystring, partial, affixlen, affix, starflags);
+  }
 
 /* Arrange to put this database at the top of the LRU chain if it is a type
 that opens real files. */
@@ -607,7 +610,7 @@ if (open_top != (tree_node *)handle &&
   open_top = t;
   }
 
-DEBUG(9)
+DEBUG(D_lookup)
   {
   tree_node *t = open_top;
   debug_printf("LRU list:\n");
@@ -620,49 +623,35 @@ DEBUG(9)
     }
   }
 
-
-/* Determine whether independent "*" or "*@" matching is wanted and adjust the
-partial count. If there is no filename, partial matching can never happen. */
-
-if (filename != NULL)
-  {
-  if (partial > 2000)
-    {
-    do_at = TRUE;
-    partial -= 2048;
-    }
-  if (partial > 1000)
-    {
-    do_star = TRUE;
-    partial -= 1024;
-    }
-  }
-else partial = -1;
-
 /* First of all, try to match the key string verbatim. If matched a complete
 entry but could have been partial, flag to set up variables. */
 
-yield = internal_search_find(handle, filename, keystring, errmsg);
+yield = internal_search_find(handle, filename, keystring);
 if (search_find_defer) return NULL;
 if (yield != NULL) { if (partial >= 0) set_null_wild = TRUE; }
 
 /* Not matched a complete entry; handle partial lookups, but only if the full
 search didn't defer. Don't use string_sprintf() to construct the initial key,
 just in case the original key is too long for the string_sprintf() buffer (it
-*has* happened!). */
+*has* happened!). The case of a zero-length affix has to be treated specially.
+*/
 
 else if (partial >= 0)
   {
-  int len = (int)strlen(keystring);
-  char *keystring2 = store_get(len + 3);
+  int len = Ustrlen(keystring);
+  uschar *keystring2;
 
-  strcpy(keystring2 + 2, keystring);
-  keystring2[0] = '*';
-  keystring2[1] = '.';
+  /* Try with the affix on the front, except for a zero-length affix */
 
-  DEBUG(9) debug_printf("trying partial match %s\n", keystring2);
-  yield = internal_search_find(handle, filename, keystring2, errmsg);
-  if (search_find_defer) return NULL;
+  if (affixlen == 0) keystring2 = keystring; else
+    {
+    keystring2 = store_get(len + affixlen + 1);
+    Ustrncpy(keystring2, affix, affixlen);
+    Ustrcpy(keystring2 + affixlen, keystring);
+    DEBUG(D_lookup) debug_printf("trying partial match %s\n", keystring2);
+    yield = internal_search_find(handle, filename, keystring2);
+    if (search_find_defer) return NULL;
+    }
 
   /* The key in its entirety did not match a wild entry; try chopping off
   leading components. */
@@ -670,16 +659,35 @@ else if (partial >= 0)
   if (yield == NULL)
     {
     int dotcount = 0;
-    char *keystring3 = keystring2 + 2;
-    char *s = keystring3;
+    uschar *keystring3 = keystring2 + affixlen;
+    uschar *s = keystring3;
     while (*s != 0) if (*s++ == '.') dotcount++;
 
     while (dotcount-- >= partial)
       {
-      while (keystring3[1] != '.' && keystring3[1] != 0) keystring3++;
-      *keystring3 = '*';
-      DEBUG(9) debug_printf("trying partial match %s\n", keystring3);
-      yield = internal_search_find(handle, filename, keystring3, errmsg);
+      while (*keystring3 != 0 && *keystring3 != '.') keystring3++;
+
+      /* If we get right to the end of the string (which will be the last time
+      through this loop), we've failed if the affix is null. Otherwise do one
+      last lookup for the affix itself, but if it is longer than 1 character,
+      remove the last character if it is ".". */
+
+      if (*keystring3 == 0)
+        {
+        if (affixlen < 1) break;
+        if (affixlen > 1 && affix[affixlen-1] == '.') affixlen--;
+        Ustrncpy(keystring2, affix, affixlen);
+        keystring2[affixlen] = 0;
+        keystring3 = keystring2;
+        }
+      else
+        {
+        keystring3 -= affixlen - 1;
+        if (affixlen > 0) Ustrncpy(keystring3, affix, affixlen);
+        }
+
+      DEBUG(D_lookup) debug_printf("trying partial match %s\n", keystring3);
+      yield = internal_search_find(handle, filename, keystring3);
       if (search_find_defer) return NULL;
       if (yield != NULL)
         {
@@ -688,8 +696,8 @@ else if (partial >= 0)
 
         if (expand_setup != NULL && *expand_setup >= 0)
           {
-          int fixedlength = (int)strlen(keystring3) - 2;
-          int wildlength = (int)strlen(keystring) - fixedlength - 1;
+          int fixedlength = Ustrlen(keystring3) - affixlen;
+          int wildlength = Ustrlen(keystring) - fixedlength - 1;
           *expand_setup += 1;
           expand_nstring[*expand_setup] = keystring;
           expand_nlength[*expand_setup] = wildlength;
@@ -699,7 +707,7 @@ else if (partial >= 0)
           }
         break;
         }
-      keystring3++;    /* Must work for null components */
+      keystring3 += affixlen;
       }
     }
 
@@ -710,17 +718,17 @@ else if (partial >= 0)
 replacing everthing to the left of @ by *. After a match, the wild part
 is set to the string to the left of the @. */
 
-if (yield == NULL && do_at)
+if (yield == NULL && (starflags & SEARCH_STARAT) != 0)
   {
-  char *atat = strrchr(keystring, '@');
+  uschar *atat = Ustrrchr(keystring, '@');
   if (atat != NULL && atat > keystring)
     {
     int savechar;
     savechar = *(--atat);
     *atat = '*';
 
-    DEBUG(9) debug_printf("trying default match %s\n", atat);
-    yield = internal_search_find(handle, filename, atat, errmsg);
+    DEBUG(D_lookup) debug_printf("trying default match %s\n", atat);
+    yield = internal_search_find(handle, filename, atat);
     *atat = savechar;
     if (search_find_defer) return NULL;
 
@@ -736,20 +744,19 @@ if (yield == NULL && do_at)
     }
   }
 
-/* If we still haven't matched anything, and the option to look for "*" is
-set, try that, but not if partial == 0, because in that case it will already
-have been searched for. If we do match, the first variable (the wild part)
-is the whole key, and the second is empty. */
+/* If we still haven't matched anything, and the option to look for "*" is set,
+try that. If we do match, the first variable (the wild part) is the whole key,
+and the second is empty. */
 
-if (yield == NULL && do_star && partial != 0)
+if (yield == NULL && (starflags & (SEARCH_STAR|SEARCH_STARAT)) != 0)
   {
-  DEBUG(9) debug_printf("trying to match *\n");
-  yield = internal_search_find(handle, filename, "*", errmsg);
+  DEBUG(D_lookup) debug_printf("trying to match *\n");
+  yield = internal_search_find(handle, filename, US"*");
   if (yield != NULL && expand_setup != NULL && *expand_setup >= 0)
     {
     *expand_setup += 1;
     expand_nstring[*expand_setup] = keystring;
-    expand_nlength[*expand_setup] = (int)strlen(keystring);
+    expand_nlength[*expand_setup] = Ustrlen(keystring);
     *expand_setup += 1;
     expand_nstring[*expand_setup] = keystring;
     expand_nlength[*expand_setup] = 0;
@@ -770,7 +777,7 @@ if (set_null_wild && expand_setup != NULL && *expand_setup >= 0)
   expand_nlength[*expand_setup] = 0;
   *expand_setup += 1;
   expand_nstring[*expand_setup] = keystring;
-  expand_nlength[*expand_setup] = (int)strlen(keystring);
+  expand_nlength[*expand_setup] = Ustrlen(keystring);
   }
 
 return yield;
