@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* A number of functions for driving outgoing SMTP calls. */
@@ -12,6 +12,137 @@
 
 
 
+/*************************************************
+*           Find an outgoing interface           *
+*************************************************/
+
+/* This function is called from the smtp transport and also from the callout
+code in verify.c. Its job is to expand a string to get a list of interfaces,
+and choose a suitable one (IPv4 or IPv6) for the outgoing address.
+
+Arguments:
+  istring    string interface setting, may be NULL, meaning "any", in
+               which case the function does nothing
+  host_af    AF_INET or AF_INET6 for the outgoing IP address
+  addr       the mail address being handled (for setting errors)
+  changed    if not NULL, set TRUE if expansion actually changed istring
+  interface  point this to the interface
+  msg        to add to any error message
+
+Returns:     TRUE on success, FALSE on failure, with error message
+               set in addr and transport_return set to PANIC
+*/
+
+BOOL
+smtp_get_interface(uschar *istring, int host_af, address_item *addr,
+  BOOL *changed, uschar **interface, uschar *msg)
+{
+uschar *expint;
+uschar *iface;
+int sep = 0;
+
+if (istring == NULL) return TRUE;
+
+expint = expand_string(istring);
+if (expint == NULL)
+  {
+  if (expand_string_forcedfail) return TRUE;
+  addr->transport_return = PANIC;
+  addr->message = string_sprintf("failed to expand \"interface\" "
+      "option for %s: %s", msg, expand_string_message);
+  return FALSE;
+  }
+
+if (changed != NULL) *changed = expint != istring;
+
+while (isspace(*expint)) expint++;
+if (*expint == 0) return TRUE;
+
+while ((iface = string_nextinlist(&expint, &sep, big_buffer,
+          big_buffer_size)) != NULL)
+  {
+  if (string_is_ip_address(iface, NULL) == 0)
+    {
+    addr->transport_return = PANIC;
+    addr->message = string_sprintf("\"%s\" is not a valid IP "
+      "address for the \"interface\" option for %s",
+      iface, msg);
+    return FALSE;
+    }
+
+  if (((Ustrchr(iface, ':') == NULL)? AF_INET:AF_INET6) == host_af)
+    break;
+  }
+
+if (iface != NULL) *interface = string_copy(iface);
+return TRUE;
+}
+
+
+
+/*************************************************
+*           Find an outgoing port                *
+*************************************************/
+
+/* This function is called from the smtp transport and also from the callout
+code in verify.c. Its job is to find a port number. Note that getservbyname()
+produces the number in network byte order.
+
+Arguments:
+  rstring     raw (unexpanded) string representation of the port
+  addr        the mail address being handled (for setting errors)
+  port        stick the port in here
+  msg         for adding to error message
+
+Returns:      TRUE on success, FALSE on failure, with error message set
+                in addr, and transport_return set to PANIC
+*/
+
+BOOL
+smtp_get_port(uschar *rstring, address_item *addr, int *port, uschar *msg)
+{
+uschar *pstring = expand_string(rstring);
+
+if (pstring == NULL)
+  {
+  addr->transport_return = PANIC;
+  addr->message = string_sprintf("failed to expand \"%s\" (\"port\" option) "
+    "for %s: %s", msg, expand_string_message);
+  return FALSE;
+  }
+
+if (isdigit(*pstring))
+  {
+  uschar *end;
+  *port = Ustrtol(pstring, &end, 0);
+  if (end != pstring + Ustrlen(pstring))
+    {
+    addr->transport_return = PANIC;
+    addr->message = string_sprintf("invalid port number for %s: %s", msg,
+      pstring);
+    return FALSE;
+    }
+  }
+
+else
+  {
+  struct servent *smtp_service = getservbyname(CS pstring, "tcp");
+  if (smtp_service == NULL)
+    {
+    addr->transport_return = PANIC;
+    addr->message = string_sprintf("TCP port \"%s\" is not defined for %s",
+      pstring, msg);
+    return FALSE;
+    }
+  *port = ntohs(smtp_service->s_port);
+  }
+
+return TRUE;
+}
+
+
+
+
 /*************************************************
 *           Connect to remote host               *
 *************************************************/
@@ -21,9 +152,14 @@ detected by checking for a colon in the address. AF_INET6 is defined even on
 non-IPv6 systems, to enable the code to be less messy. However, on such systems
 host->address will always be an IPv4 address.
 
+The port field in the host item is used if it is set (usually router from SRV
+records). In other cases, the default passed as an argument is used.
+
 Arguments:
-  host        host item containing name and address
-  port        remote port to connect to, in network byte order
+  host        host item containing name and address (and sometimes port)
+  host_af     AF_INET or AF_INET6
+  port        default, remote port to connect to, in host byte order for those
+                hosts whose port setting is PORT_NONE
   interface   outgoing interface address or NULL
   timeout     timeout value or 0
   keepalive   TRUE to use keepalive
@@ -32,165 +168,73 @@ Returns:      connected socket number, or -1 with errno set
 */
 
 int
-smtp_connect(host_item *host, int port, char *interface, int timeout,
-  BOOL keepalive)
+smtp_connect(host_item *host, int host_af, int port, uschar *interface,
+  int timeout, BOOL keepalive)
 {
-struct sockaddr *s_ptr;
-struct sockaddr_in s_in4;
-int s_len, rc, sock, save_errno;
-int host_af = (strchr(host->address, ':') != NULL)? AF_INET6 : AF_INET;
+int on = 1;
+int save_errno = 0;
+int sock;
 
-#if HAVE_IPV6
-struct sockaddr_in6 s_in6;
-
-/* For an IPv6 address, use an IPv6 sockaddr structure. */
-
-if (host_af == AF_INET6)
+if (host->port != PORT_NONE)
   {
-  s_ptr = (struct sockaddr *)&s_in6;
-  s_len = sizeof(s_in6);
-  }
-else
-#endif
-
-/* For an IPv4 address, use an IPv4 sockaddr structure,
-even on an IPv6 system. */
-
-  {
-  s_ptr = (struct sockaddr *)&s_in4;
-  s_len = sizeof(s_in4);
+  HDEBUG(D_transport|D_acl|D_v)
+    debug_printf("Transport port=%d replaced by host-specific port=%d\n", port,
+      host->port);
+  port = host->port;
   }
 
-HDEBUG(1) debug_printf("Connecting to %s [%s.%d] ... ", host->name,
-  host->address, (int)ntohs((unsigned short)port));
-
-/* In the old days of IPv4, failure to create a socket usually meant things
-were dire, and so Exim used to panic and die. However, with the arrival of IPv6
-there are circumstances where IPv6 sockets fail, but IPv4 ones work, and if a
-domain is routed to a mixture of IPv4 and IPv6 addresses, the right thing to do
-is to let it try them all. Consequently, this error no longer causes a panic,
-but instead gives an error return. */
-
-sock = socket(host_af, SOCK_STREAM, 0);
-if (sock < 0)
+HDEBUG(D_transport|D_acl|D_v)
   {
-  int save_errno = errno;
-  HDEBUG(1) debug_printf("socket creation failed: %s\n", strerror(errno));
-  errno = save_errno;
-  return -1;
+  if (interface == NULL)
+    debug_printf("Connecting to %s [%s]:%d ... ",host->name,host->address,port);
+  else
+    debug_printf("Connecting to %s [%s]:%d from %s ... ", host->name,
+      host->address, port, interface);
   }
 
-/* Bind to a specific interface if requested. On an IPv6 system, this has
-to be of the same family as the address we are calling. On an IPv4 system the
-test is redundant, but it keeps the code tidier. */
+/* Create the socket */
 
-if (interface != NULL)
+if ((sock = ip_socket(SOCK_STREAM, host_af)) < 0) return -1;
+
+/* Set TCP_NODELAY; Exim does its own buffering. */
+
+setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (uschar *)(&on), sizeof(on));
+
+/* Bind to a specific interface if requested. Caller must ensure the interface
+is the same type (IPv4 or IPv6) as the outgoing address. */
+
+if (interface != NULL && ip_bind(sock, host_af, interface, 0) < 0)
   {
-  int interface_af = (strchr(interface, ':') != NULL)? AF_INET6 : AF_INET;
-
-  if (interface_af == host_af)
-    {
-    #if HAVE_IPV6
-
-    /* Set up for IPv6 binding */
-
-    if (host_af == AF_INET6)
-      {
-      memset(&s_in6, 0, sizeof(s_in6));
-      s_in6.sin6_family = AF_INET6;
-      s_in6.sin6_port = 0;
-      if (inet_pton(AF_INET6, interface, &s_in6.sin6_addr) != 1)
-        log_write(0, LOG_PANIC_DIE, "unable to parse \"%s\"", interface);
-      }
-    else
-    #endif
-
-    /* Set up for IPv4 binding */
-
-      {
-      memset(&s_in4, 0, sizeof(s_in4));
-      s_in4.sin_family = AF_INET;
-      s_in4.sin_port = 0;
-      s_in4.sin_addr.s_addr = (S_ADDR_TYPE)inet_addr(interface);
-      }
-
-    /* Bind */
-
-    if (bind(sock, s_ptr, s_len) < 0)
-      log_write(0, LOG_PANIC_DIE, "unable to bind outgoing SMTP call to %s: %s",
-        interface, strerror(errno));
-    }
+  save_errno = errno;
+  HDEBUG(D_transport|D_acl|D_v)
+    debug_printf("unable to bind outgoing SMTP call to %s: %s", interface,
+    strerror(errno));
   }
 
-/* Set up a remote IPv6 address */
+/* Connect to the remote host, and add keepalive to the socket before returning
+it, if requested. */
 
-#if HAVE_IPV6
-if (host_af == AF_INET6)
+else if (ip_connect(sock, host_af, host->address, port, timeout) < 0)
+  save_errno = errno;
+
+/* Either bind() or connect() failed */
+
+if (save_errno != 0)
   {
-  memset(&s_in6, 0, sizeof(s_in6));
-  s_in6.sin6_family = AF_INET6;
-  s_in6.sin6_port = port;
-  if (inet_pton(host_af, host->address, &s_in6.sin6_addr) != 1)
-    log_write(0, LOG_PANIC_DIE, "unable to parse \"%s\"", host->address);
-  }
-else
-#endif
-
-/* Set up a remote IPv4 address */
-
-  {
-  memset(&s_in4, 0, sizeof(s_in4));
-  s_in4.sin_family = AF_INET;
-  s_in4.sin_port = port;
-  s_in4.sin_addr.s_addr = (S_ADDR_TYPE)inet_addr(host->address);
-  }
-
-/* If no connection timeout is set, just call connect() without setting
-a timer, thereby allowing the inbuilt timeout to operate. */
-
-sigalrm_seen = FALSE;
-
-if (timeout > 0)
-  {
-  os_non_restarting_signal(SIGALRM, sigalrm_handler);
-  alarm(timeout);
-  }
-
-rc = connect(sock, s_ptr, s_len);
-save_errno = errno;
-
-if (timeout > 0)
-  {
-  alarm(0);
-  signal(SIGALRM, SIG_IGN);
-  }
-
-/* A failure whose error code is "Interrupted system call" is in fact
-an externally applied timeout if the signal handler has been run. */
-
-if (rc < 0)
-  {
-  if (save_errno == EINTR && sigalrm_seen) save_errno = ETIMEDOUT;
+  HDEBUG(D_transport|D_acl|D_v) debug_printf("failed\n");
   close(sock);
-  HDEBUG(1) debug_printf("failed\n");
   errno = save_errno;
   return -1;
   }
 
-HDEBUG(1) debug_printf("connected\n");
+/* Both bind() and connect() succeeded */
 
-/* Add keepalive to the socket if requested, before returning it. */
-
-if (keepalive)
+else
   {
-  int fodder = 1;
-  if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
-      (char *)(&fodder), sizeof(fodder)) != 0)
-    log_write(4, LOG_MAIN, "send setsockopt(SO_KEEPALIVE) failed on "
-      "connection to %s: %s", host->address, strerror(errno));
+  HDEBUG(D_transport|D_acl|D_v) debug_printf("connected\n");
+  if (keepalive) ip_keepalive(sock, host->address, TRUE);
+  return sock;
   }
-
-return sock;
 }
 
 
@@ -222,7 +266,7 @@ else
 rc = send(outblock->sock, outblock->buffer, outblock->ptr - outblock->buffer, 0);
 if (rc <= 0)
   {
-  HDEBUG(1) debug_printf("send failed: %s\n", strerror(errno));
+  HDEBUG(D_transport|D_acl) debug_printf("send failed: %s\n", strerror(errno));
   return FALSE;
   }
 
@@ -260,11 +304,11 @@ int rc = 0;
 va_list ap;
 
 va_start(ap, format);
-if (!string_vformat(big_buffer, big_buffer_size, format, ap))
-  log_write(0, LOG_PANIC_DIE, "overlong write_command in outgoing SMTP");
+if (!string_vformat(big_buffer, big_buffer_size, CS format, ap))
+  log_write(0, LOG_MAIN|LOG_PANIC_DIE, "overlong write_command in outgoing "
+    "SMTP");
 va_end(ap);
-count = (int)strlen(big_buffer);
-HDEBUG(1) debug_printf("  SMTP>> %s", big_buffer);
+count = Ustrlen(big_buffer);
 
 if (count > outblock->buffersize - (outblock->ptr - outblock->buffer))
   {
@@ -272,10 +316,31 @@ if (count > outblock->buffersize - (outblock->ptr - outblock->buffer))
   if (!flush_buffer(outblock)) return -1;
   }
 
-strncpy(CS outblock->ptr, big_buffer, count);
+Ustrncpy(CS outblock->ptr, big_buffer, count);
 outblock->ptr += count;
 outblock->cmd_count++;
-big_buffer[count-2] = 0;     /* remove \r\n for error message */
+count -= 2;
+big_buffer[count] = 0;     /* remove \r\n for error message */
+
+/* We want to hide the actual data sent in AUTH transactions from reflections
+and logs. While authenticating, a flag is set in the outblock to enable this.
+The AUTH command itself gets any data flattened. Other lines are flattened
+completely. */
+
+if (outblock->authenticating)
+  {
+  uschar *p = big_buffer;
+  if (Ustrncmp(big_buffer, "AUTH ", 5) == 0)
+    {
+    p += 5;
+    while (isspace(*p)) p++;
+    while (!isspace(*p)) p++;
+    while (isspace(*p)) p++;
+    }
+  while (*p != 0) *p++ = '*';
+  }
+
+HDEBUG(D_transport|D_acl|D_v) debug_printf("  SMTP>> %s\n", big_buffer);
 
 if (!noflush)
   {
@@ -311,19 +376,18 @@ Returns:    length of a line that has been put in the buffer
 static int
 read_response_line(smtp_inblock *inblock, uschar *buffer, int size, int timeout)
 {
-fd_set select_inset;
-struct timeval tv;
 uschar *p = buffer;
 uschar *ptr = inblock->ptr;
 uschar *ptrend = inblock->ptrend;
 int sock = inblock->sock;
-int rc;
 
 /* Loop for reading multiple packets or reading another packet after emptying
 a previously-read one. */
 
 for (;;)
   {
+  int rc;
+
   /* If there is data in the input buffer left over from last time, copy
   characters from it until the end of a line, at which point we can return,
   having removed any whitespace (which will include CR) at the end of the line.
@@ -351,74 +415,22 @@ for (;;)
       }
     }
 
-  /* We need to read a new input packet. Loop to cover select() getting
-  interrupted, and the possibility of select() returning with a positive
-  result but no ready descriptor. Is this in fact possible? */
+  /* Need to read a new input packet. */
 
-  for (;;)
-    {
-    FD_ZERO (&select_inset);
-    FD_SET (sock, &select_inset);
-    tv.tv_sec = timeout;
-    tv.tv_usec = 0;
-    rc = select(sock + 1, (SELECT_ARG2_TYPE *)&select_inset, NULL, NULL, &tv);
-
-    /* If some interrupt arrived, just retry. We presume this to be rare,
-    but it can happen (e.g. the SIGUSR1 signal sent by exiwhat causes
-    select() to exit). */
-
-    if (rc < 0 && errno == EINTR)
-      {
-      HDEBUG(9) debug_printf("EINTR while selecting for SMTP response\n");
-      continue;
-      }
-
-    /* Handle a timeout, and treat any other error as a timeout. */
-
-    if (rc <= 0)
-      {
-      errno = ETIMEDOUT;
-      goto ERROR_RETURN;
-      }
-
-    /* If the socket is ready, read from it (via TLS if it's active), and break
-    out of this select/retry loop. */
-
-    if (FD_ISSET(sock, &select_inset))
-      {
-      #ifdef SUPPORT_TLS
-      if (tls_active == sock)
-        rc = tls_read(inblock->buffer, size-1);
-      else
-      #endif
-        rc = recv(sock, inblock->buffer, inblock->buffersize, 0);
-      break;
-      }
-    }
-
-  /* We get here when a read operation has been obeyed. First, handle errors.
-  On EOF (i.e. close down of the connection), set errno to zero; otherwise
-  leave it alone. */
-
-  if (rc <= 0)
-    {
-    if (rc == 0) errno = 0;
-    goto ERROR_RETURN;
-    }
+  rc = ip_recv(sock, inblock->buffer, inblock->buffersize, timeout);
+  if (rc <= 0) break;
 
   /* Another block of data has been successfully read. Set up the pointers
   and let the loop continue. */
 
   ptrend = inblock->ptrend = inblock->buffer + rc;
   ptr = inblock->buffer;
-  DEBUG(9) debug_printf("read response data: size=%d\n", rc);
+  DEBUG(D_transport|D_acl) debug_printf("read response data: size=%d\n", rc);
   }
 
-/* Get here if there has been some kind of error; errno is set, but we
+/* Get here if there has been some kind of recv() error; errno is set, but we
 ensure that the result buffer is empty before returning. */
 
-ERROR_RETURN:
-
 *buffer = 0;
 return -1;
 }
@@ -465,7 +477,7 @@ for (;;)
   if ((count = read_response_line(inblock, ptr, size, timeout)) < 0)
     return FALSE;
 
-  HDEBUG(1)
+  HDEBUG(D_transport|D_acl|D_v)
     debug_printf("  %s %s\n", (ptr == buffer)? "SMTP<<" : "      ", ptr);
 
   /* Check the format of the response: it must start with three digits; if
