@@ -2,16 +2,18 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
-/* Functions for reading spool files. */
+/* Functions for reading spool files. When compiling for a utility (eximon),
+not all are needed, and some functionality can be cut out. */
 
 
 #include "exim.h"
 
 
 
+#ifndef COMPILE_UTILITY
 /*************************************************
 *           Open and lock data file              *
 *************************************************/
@@ -30,12 +32,12 @@ Side effect: deliver_datafile is set to the fd of the open file.
 */
 
 BOOL
-spool_open_datafile(char *id)
+spool_open_datafile(uschar *id)
 {
 int i;
 struct stat statbuf;
 flock_t lock_data;
-char spoolname[256];
+uschar spoolname[256];
 
 /* If split_spool_directory is set, first look for the file in the appropriate
 sub-directory of the input directory. If it is not found there, try the input
@@ -48,8 +50,8 @@ for (i = 0; i < 2; i++)
   {
   int save_errno;
   message_subdir[0] = (split_spool_directory == (i == 0))? id[5] : 0;
-  sprintf(spoolname, "%s/input/%s/%s-D", spool_directory, message_subdir, id);
-  deliver_datafile = open(spoolname, O_RDWR | O_APPEND);
+  sprintf(CS spoolname, "%s/input/%s/%s-D", spool_directory, message_subdir, id);
+  deliver_datafile = Uopen(spoolname, O_RDWR | O_APPEND, 0);
   if (deliver_datafile >= 0) break;
   save_errno = errno;
   if (errno == ENOENT)
@@ -65,18 +67,26 @@ for (i = 0; i < 2; i++)
   }
 
 /* File is open and message_subdir is set. Set the close-on-exec flag, and lock
-the file. */
+the file. We lock only the first line of the file (containing the message ID)
+because this apparently is needed for running Exim under Cygwin. If the entire
+file is locked in one process, a sub-process cannot access it, even when passed
+an open file descriptor (at least, I think that's the Cygwin story). On real
+Unix systems it doesn't make any difference as long as Exim is consistent in
+what it locks. */
 
 fcntl(deliver_datafile, F_SETFD, fcntl(deliver_datafile, F_GETFD) |
   FD_CLOEXEC);
 
 lock_data.l_type = F_WRLCK;
-lock_data.l_whence = lock_data.l_start = lock_data.l_len = 0;
+lock_data.l_whence = SEEK_SET;
+lock_data.l_start = 0;
+lock_data.l_len = SPOOL_DATA_START_OFFSET;
 
 if (fcntl(deliver_datafile, F_SETLK, &lock_data) < 0)
   {
-  log_write(5, LOG_MAIN,
-    "Spool file is locked (another process is handling this message)");
+  log_write(L_skip_delivery,
+            LOG_MAIN,
+            "Spool file is locked (another process is handling this message)");
   close(deliver_datafile);
   deliver_datafile = -1;
   errno = 0;
@@ -88,12 +98,13 @@ in the count, but add one for the newline before the data. */
 
 if (fstat(deliver_datafile, &statbuf) == 0)
   {
-  message_body_size = statbuf.st_size - MESSAGE_ID_LENGTH - 3;
+  message_body_size = statbuf.st_size - SPOOL_DATA_START_OFFSET;
   message_size = message_body_size + 1;
   }
 
 return TRUE;
 }
+#endif  /* COMPILE_UTILITY */
 
 
 
@@ -145,23 +156,23 @@ Returns:       FALSE on format error
 */
 
 static BOOL
-read_nonrecipients_tree(tree_node **connect, FILE *f, char *buffer,
+read_nonrecipients_tree(tree_node **connect, FILE *f, uschar *buffer,
   int buffer_size)
 {
 tree_node *node;
-int n = (int)strlen(buffer);
+int n = Ustrlen(buffer);
 BOOL right = buffer[1] == 'Y';
 
 if (n < 5) return FALSE;    /* malformed line */
 buffer[n-1] = 0;            /* Remove \n */
 node = store_get(sizeof(tree_node) + n - 3);
 *connect = node;
-strcpy(node->name, buffer + 3);
+Ustrcpy(node->name, buffer + 3);
 node->data.ptr = NULL;
 
 if (buffer[0] == 'Y')
   {
-  if (fgets(buffer, buffer_size, f) == NULL ||
+  if (Ufgets(buffer, buffer_size, f) == NULL ||
     !read_nonrecipients_tree(&node->left, f, buffer, buffer_size))
       return FALSE;
   }
@@ -169,7 +180,7 @@ else node->left = NULL;
 
 if (right)
   {
-  if (fgets(buffer, buffer_size, f) == NULL ||
+  if (Ufgets(buffer, buffer_size, f) == NULL ||
     !read_nonrecipients_tree(&node->right, f, buffer, buffer_size))
       return FALSE;
   }
@@ -210,33 +221,53 @@ Returns:        spool_read_OK        success
 */
 
 int
-spool_read_header(char *name, BOOL read_headers, BOOL subdir_set)
+spool_read_header(uschar *name, BOOL read_headers, BOOL subdir_set)
 {
-FILE *f;
+FILE *f = NULL;
 int n;
+int rcount = 0;
 long int uid, gid;
 BOOL inheader = FALSE;
-char originator[64];
+uschar originator[64];
 
-/* Ensure all pointers to store that will be obtained are NULL, just
-in case. */
+/* Reset all the global variables to their default values. However, there is
+one exception. DO NOT change the default value of dont_deliver, because it may
+be forced by an external setting. */
+
+for (n = 0; n < ACL_C_MAX + ACL_M_MAX; n++) acl_var[n] = NULL;
 
 authenticated_id = NULL;
 authenticated_sender = NULL;
+body_linecount = 0;
+deliver_firsttime = FALSE;
+deliver_freeze = FALSE;
+deliver_frozen_at = 0;
+deliver_manual_thaw = FALSE;
+/* dont_deliver must NOT be reset */
 header_list = header_last = NULL;
+host_lookup_failed = FALSE;
+interface_address = NULL;
+interface_port = 0;
+local_error_message = FALSE;
+local_scan_data = NULL;
+message_linecount = 0;
+received_protocol = NULL;
+received_count = 0;
+recipients_list = NULL;
 sender_address = NULL;
+sender_fullhost = NULL;
+sender_helo_name = NULL;
 sender_host_address = NULL;
 sender_host_name = NULL;
 sender_host_port = 0;
 sender_host_authenticated = NULL;
-interface_address = NULL;
-sender_helo_name = NULL;
-sender_fullhost = NULL;
 sender_ident = NULL;
-recipients_list = NULL;
+sender_local = FALSE;
+sender_set_untrusted = FALSE;
 tree_nonrecipients = NULL;
 
 #ifdef SUPPORT_TLS
+tls_certificate_verified = FALSE;
 tls_cipher = NULL;
 tls_peerdn = NULL;
 #endif
@@ -249,22 +280,25 @@ for (n = 0; n < 2; n++)
   {
   if (!subdir_set)
     message_subdir[0] = (split_spool_directory == (n == 0))? name[5] : 0;
-  sprintf(big_buffer, "%s/input/%s/%s", spool_directory, message_subdir, name);
-  f = fopen(big_buffer, "r");
+  sprintf(CS big_buffer, "%s/input/%s/%s", spool_directory, message_subdir,
+    name);
+  f = Ufopen(big_buffer, "rb");
   if (f != NULL) break;
   if (n != 0 || subdir_set || errno != ENOENT) return spool_read_notopen;
   }
 
 errno = 0;
 
-DEBUG(5) debug_printf("Opened spool file %s\n", name);
+#ifndef COMPILE_UTILITY
+DEBUG(D_deliver) debug_printf("reading spool file %s\n", name);
+#endif  /* COMPILE_UTILITY */
 
 /* The first line of a spool file contains the message id followed by -H (i.e.
 the file name), in order to make the file self-identifying. */
 
-if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
-if ((int)strlen(big_buffer) != MESSAGE_ID_LENGTH + 3 ||
-    strncmp(big_buffer, name, MESSAGE_ID_LENGTH + 2) != 0)
+if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+if (Ustrlen(big_buffer) != MESSAGE_ID_LENGTH + 3 ||
+    Ustrncmp(big_buffer, name, MESSAGE_ID_LENGTH + 2) != 0)
   goto SPOOL_FORMAT_ERROR;
 
 /* The next three lines in the header file are in a fixed format. The first
@@ -273,125 +307,120 @@ The second contains the mail address of the message's sender, enclosed in <>.
 The third contains the time the message was received, and the number of warning
 messages for delivery delays that have been sent. */
 
-if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
 
-if (sscanf(big_buffer, "%s %ld %ld", originator, &uid, &gid) != 3)
+if (sscanf(CS big_buffer, "%s %ld %ld", originator, &uid, &gid) != 3)
   goto SPOOL_FORMAT_ERROR;
 originator_login = string_copy(originator);
 originator_uid = (uid_t)uid;
 originator_gid = (gid_t)gid;
 
-if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
-n = (int)strlen(big_buffer);
+if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+n = Ustrlen(big_buffer);
 if (n < 3 || big_buffer[0] != '<' || big_buffer[n-2] != '>')
   goto SPOOL_FORMAT_ERROR;
 
 sender_address = store_get(n-2);
-strncpy(sender_address, big_buffer+1, n-3);
+Ustrncpy(sender_address, big_buffer+1, n-3);
 sender_address[n-3] = 0;
 
-if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
-if (sscanf(big_buffer, "%d %d", &received_time, &warning_count) != 2)
+if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+if (sscanf(CS big_buffer, "%d %d", &received_time, &warning_count) != 2)
   goto SPOOL_FORMAT_ERROR;
 
 message_age = time(NULL) - received_time;
 
-DEBUG(5) debug_printf("user=%s uid=%ld gid=%ld sender=%s\n",
+#ifndef COMPILE_UTILITY
+DEBUG(D_deliver) debug_printf("user=%s uid=%ld gid=%ld sender=%s\n",
   originator_login, (long int)originator_uid, (long int)originator_gid,
   sender_address);
+#endif  /* COMPILE_UTILITY */
 
 /* Now there may be a number of optional lines, each starting with "-".
-First set default values for the options. */
-
-sender_local = FALSE;
-header_names = header_names_normal;
-body_linecount = 0;
-message_linecount = 0;
-deliver_freeze = FALSE;
-deliver_frozen_at = 0;
-host_lookup_failed = FALSE;
-local_error_message = FALSE;
-sender_set_untrusted = FALSE;
-
-#ifdef SUPPORT_DSN
-/* This partial attempt at doing DSN was abandoned. This obsolete code is
-left here just in case. Nothing is documented. */
-dsn_ret = 0;
-dsn_envid = NULL;
-#endif
+If you add a new setting here, make sure you set the default above. */
 
 for (;;)
   {
-  if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+  if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
   if (big_buffer[0] != '-') break;
 
-  big_buffer[(int)strlen(big_buffer) - 1] = 0;
-  if (strcmp(big_buffer, "-local") == 0) sender_local = TRUE;
-    else if (strcmp(big_buffer, "-localerror") == 0)
-      local_error_message = TRUE;
-    else if (strcmp(big_buffer, "-host_lookup_failed") == 0)
-      host_lookup_failed = TRUE;
-    else if (strcmp(big_buffer, "-resent") == 0)
-      header_names = header_names_resent;
-    else if (strncmp(big_buffer, "-body_linecount", 15) == 0)
-      body_linecount = atoi(big_buffer + 15);
-    else if (strncmp(big_buffer, "-frozen", 7) == 0)
+  big_buffer[Ustrlen(big_buffer) - 1] = 0;
+  if (Ustrncmp(big_buffer, "-acl ", 5) == 0)
+    {
+    int index, count;
+    if (sscanf(CS big_buffer + 5, "%d %d", &index, &count) != 2)
+      goto SPOOL_FORMAT_ERROR;
+    /* Ignore if index too big - might be if a later release with more
+    variables built this spool file. */
+    if (index < ACL_C_MAX + ACL_M_MAX)
       {
-      deliver_freeze = TRUE;
-      deliver_frozen_at = atoi(big_buffer + 7);
+      acl_var[index] = store_get(count + 1);
+      if (fread(acl_var[index], 1, count+1, f) < count) goto SPOOL_READ_ERROR;
+      acl_var[index][count] = 0;
       }
-    else if (strcmp(big_buffer, "-deliver_firsttime") == 0)
-      deliver_firsttime = TRUE;
-    else if (strcmp(big_buffer, "-manual_thaw") == 0)
-      deliver_manual_thaw = TRUE;
-    else if (strncmp(big_buffer, "-auth_id", 8) == 0)
-      authenticated_id = string_copy(big_buffer + 9);
-    else if (strncmp(big_buffer, "-auth_sender", 12) == 0)
-      authenticated_sender = string_copy(big_buffer + 13);
-    else if (strncmp(big_buffer, "-sender_set_untrusted", 21) == 0)
-      sender_set_untrusted = TRUE;
+    }
+  else if (Ustrcmp(big_buffer, "-local") == 0) sender_local = TRUE;
+  else if (Ustrcmp(big_buffer, "-localerror") == 0)
+    local_error_message = TRUE;
+  else if (Ustrncmp(big_buffer, "-local_scan ", 12) == 0)
+    local_scan_data = string_copy(big_buffer + 12);
+  else if (Ustrcmp(big_buffer, "-host_lookup_failed") == 0)
+    host_lookup_failed = TRUE;
+  else if (Ustrncmp(big_buffer, "-body_linecount", 15) == 0)
+    body_linecount = Uatoi(big_buffer + 15);
+  else if (Ustrncmp(big_buffer, "-frozen", 7) == 0)
+    {
+    deliver_freeze = TRUE;
+    deliver_frozen_at = Uatoi(big_buffer + 7);
+    }
+  else if (Ustrcmp(big_buffer, "-deliver_firsttime") == 0)
+    deliver_firsttime = TRUE;
+  else if (Ustrcmp(big_buffer, "-manual_thaw") == 0)
+    deliver_manual_thaw = TRUE;
+  else if (Ustrncmp(big_buffer, "-auth_id", 8) == 0)
+    authenticated_id = string_copy(big_buffer + 9);
+  else if (Ustrncmp(big_buffer, "-auth_sender", 12) == 0)
+    authenticated_sender = string_copy(big_buffer + 13);
+  else if (Ustrncmp(big_buffer, "-sender_set_untrusted", 21) == 0)
+    sender_set_untrusted = TRUE;
 
-    #ifdef SUPPORT_TLS
-    else if (strncmp(big_buffer, "-tls_cipher", 11) == 0)
-      tls_cipher = string_copy(big_buffer + 12);
-    else if (strncmp(big_buffer, "-tls_peerdn", 11) == 0)
-      tls_peerdn = string_copy(big_buffer + 12);
-    #endif
+  #ifdef SUPPORT_TLS
+  else if (Ustrncmp(big_buffer, "-tls_certificate_verified", 25) == 0)
+    tls_certificate_verified = TRUE;
+  else if (Ustrncmp(big_buffer, "-tls_cipher", 11) == 0)
+    tls_cipher = string_copy(big_buffer + 12);
+  else if (Ustrncmp(big_buffer, "-tls_peerdn", 11) == 0)
+    tls_peerdn = string_copy(big_buffer + 12);
+  #endif
 
-    #ifdef SUPPORT_DSN
-    /* This partial attempt at doing DSN was abandoned. This obsolete code is
-    left here just in case. Nothing is documented. */
-    else if (strncmp(big_buffer, "-dsn_ret", 8) == 0)
-      dsn_ret= atoi(big_buffer + 8);
+  /* We now record the port number after the address, separated by a
+  dot. For compatibility during upgrading, do nothing if there
+  isn't a value (it gets left at zero). */
 
-    else if (strncmp(big_buffer, "-dsn_envid", 10) == 0)
-      dsn_envid = string_copy(big_buffer + 11);
-    #endif
+  else if (Ustrncmp(big_buffer, "-host_address", 13) == 0)
+    {
+    sender_host_port = host_extract_port(big_buffer + 14);
+    sender_host_address = string_copy(big_buffer + 14);
+    }
 
-    /* We now record the port number after the address, separated by a
-    dot. For compatibility during upgrading, do nothing if there
-    isn't a value (it gets left at zero). */
+  else if (Ustrncmp(big_buffer, "-interface_address", 18) == 0)
+    {
+    interface_port = host_extract_port(big_buffer + 19);
+    interface_address = string_copy(big_buffer + 19);
+    }
 
-    else if (strncmp(big_buffer, "-host_address", 13) == 0)
-      {
-      sender_host_port = host_extract_port(big_buffer + 14);
-      sender_host_address = string_copy(big_buffer + 14);
-      }
-
-    else if (strncmp(big_buffer, "-host_auth", 10) == 0)
-      sender_host_authenticated = string_copy(big_buffer + 11);
-    else if (strncmp(big_buffer, "-host_name", 10) == 0)
-      sender_host_name = string_copy(big_buffer + 11);
-    else if (strncmp(big_buffer, "-interface_address", 18) == 0)
-      interface_address = string_copy(big_buffer + 19);
-    else if (strncmp(big_buffer, "-helo_name", 10) == 0)
-      sender_helo_name = string_copy(big_buffer + 11);
-    else if (strncmp(big_buffer, "-ident", 6) == 0)
-      sender_ident = string_copy(big_buffer + 7);
-    else if (strncmp(big_buffer, "-received_protocol", 18) == 0)
-      received_protocol = string_copy(big_buffer + 19);
-    else if (strncmp(big_buffer, "-N", 2) == 0)
-      dont_deliver = TRUE;
+  else if (Ustrncmp(big_buffer, "-host_auth", 10) == 0)
+    sender_host_authenticated = string_copy(big_buffer + 11);
+  else if (Ustrncmp(big_buffer, "-host_name", 10) == 0)
+    sender_host_name = string_copy(big_buffer + 11);
+  else if (Ustrncmp(big_buffer, "-helo_name", 10) == 0)
+    sender_helo_name = string_copy(big_buffer + 11);
+  else if (Ustrncmp(big_buffer, "-ident", 6) == 0)
+    sender_ident = string_copy(big_buffer + 7);
+  else if (Ustrncmp(big_buffer, "-received_protocol", 18) == 0)
+    received_protocol = string_copy(big_buffer + 19);
+  else if (Ustrncmp(big_buffer, "-N", 2) == 0)
+    dont_deliver = TRUE;
 
   /* To allow new versions of Exim that add additional flags to interwork
   with older versions that do not understand them, just ignore any flagged
@@ -404,106 +433,141 @@ for (;;)
 
 /* Build sender_fullhost if required */
 
+#ifndef COMPILE_UTILITY
 host_build_sender_fullhost();
+#endif  /* COMPILE_UTILITY */
 
-DEBUG(5)
-  debug_printf("sender_local=%d resent=%s ident=%s\n", sender_local,
-    (header_names == header_names_normal)? "no" : "yes",
-    (sender_ident == NULL)? "unset" : sender_ident);
+#ifndef COMPILE_UTILITY
+DEBUG(D_deliver)
+  debug_printf("sender_local=%d ident=%s\n", sender_local,
+    (sender_ident == NULL)? US"unset" : sender_ident);
+#endif  /* COMPILE_UTILITY */
 
 /* We now have the tree of addresses NOT to deliver to, or a line
 containing "XX", indicating no tree. */
 
-if (strncmp(big_buffer, "XX\n", 3) != 0 &&
+if (Ustrncmp(big_buffer, "XX\n", 3) != 0 &&
   !read_nonrecipients_tree(&tree_nonrecipients, f, big_buffer, big_buffer_size))
     goto SPOOL_FORMAT_ERROR;
 
-DEBUG(5)
+#ifndef COMPILE_UTILITY
+DEBUG(D_deliver)
   {
   debug_printf("Non-recipients:\n");
-  tree_print(tree_nonrecipients, debug_file);
+  debug_print_tree(tree_nonrecipients);
   }
+#endif  /* COMPILE_UTILITY */
 
 /* After reading the tree, the next line has not yet been read into the
 buffer. It contains the count of recipients which follow on separate lines. */
 
-if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
-if (sscanf(big_buffer, "%d", &recipients_count) != 1) goto SPOOL_FORMAT_ERROR;
+if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+if (sscanf(CS big_buffer, "%d", &rcount) != 1) goto SPOOL_FORMAT_ERROR;
 
-DEBUG(5) debug_printf("recipients_count=%d\n", recipients_count);
+#ifndef COMPILE_UTILITY
+DEBUG(D_deliver) debug_printf("recipients_count=%d\n", rcount);
+#endif  /* COMPILE_UTILITY */
 
-recipients_list_max = recipients_count;
-recipients_list = store_get(recipients_count * sizeof(recipient_item));
+recipients_list_max = rcount;
+recipients_list = store_get(rcount * sizeof(recipient_item));
 
-for (n = 0; n < recipients_count; n++)
+for (recipients_count = 0; recipients_count < rcount; recipients_count++)
   {
   int nn;
-  int flags = 0;
-  int pno = 0;
-  char *orcpt = NULL;
-  char *p;
+  int pno = -1;
+  uschar *errors_to = NULL;
+  uschar *p;
 
-  if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
-  nn = (int)strlen(big_buffer);
+  if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+  nn = Ustrlen(big_buffer);
   if (nn < 2) goto SPOOL_FORMAT_ERROR;
 
-  /* Look back from the end for digits/commas and a space, indicating the
-  presence of flags, pno data, and possibly the DSN orcpt field.
-  The partial attempt at doing DSN was abandoned. This obsolete code is
-  left here just in case. Nothing is documented.
+  /* Remove the newline; this terminates the address if there is no additional
+  data on the line. */
 
-  Jan, 2002: Added code to make it read the Exim 4 version of this, which
-  is just a single number. At least, it was at first. Later it was revised
-  to a more flexible format. Further code added: April 2002. */
+  p = big_buffer + nn - 1;
+  *p-- = 0;
 
-  p = big_buffer + nn - 2;
-  while (isdigit((uschar)*p) || *p == ',') p--;
+  /* Look back from the end of the line for digits and special terminators.
+  Since an address must end with a domain, we can tell that extra data is
+  present by the presence of the terminator, which is always some character
+  that cannot exist in a domain. (If I'd thought of the need for additional
+  data early on, I'd have put it at the start, with the address at the end. As
+  it is, we have to operate backwards. Addresses are permitted to contain
+  spaces, you see.)
 
-  /* New Exim 4 style */
+  This code has to cope with various versions of this data that have evolved
+  over time. In all cases, the line might just contain an address, with no
+  additional data. Otherwise, the possibilities are as follows:
 
-  if (*p == '#')
+  Exim 3 type:       <address><space><digits>,<digits>,<digits>
+
+    The second set of digits is the parent number for one_time addresses. The
+    other values were remnants of earlier experiments that were abandoned.
+
+  Exim 4 first type: <address><space><digits>
+
+    The digits are the parent number for one_time addresses.
+
+  Exim 4 new type:   <address><space><data>#<type bits>
+
+    The type bits indicate what the contents of the data are.
+
+    Bit 01 indicates that, reading from right to left, the data
+      ends with <errors_to address><space><len>,<pno> where pno is
+      the parent number for one_time addresses, and len is the length
+      of the errors_to address (zero meaning none).
+   */
+
+  while (isdigit(*p)) p--;
+
+  /* Handle Exim 3 spool files */
+
+  if (*p == ',')
     {
-    int len;
-    p--;
-    while (isdigit((uschar)*p) || *p == ',') p--;
-    sscanf(p+1, "%d,%d", &len, &pno);
-    flags = rf_onetime;
-    p -= len + 1;
-    *p = 0;
+    int dummy;
+    while (isdigit(*(--p)) || *p == ',');
+    if (*p == ' ')
+      {
+      *p++ = 0;
+      sscanf(CS p, "%d,%d", &dummy, &pno);
+      }
     }
 
-  /* Old Exim 4 style and Exim 3 style */
+  /* Handle early Exim 4 spool files */
 
   else if (*p == ' ')
     {
-    int orc;
     *p++ = 0;
-    if (sscanf(p, "%d,%d,%d", &flags, &pno, &orc) == 1)
-      {
-      pno = flags;                 /* Convert Exim 4 one-number */
-      flags = rf_onetime;          /* to Exim 3 setting plus a  */
-      orc = 0;                     /* bit in the flags */
-      }
-    if (orc > 0)
-      {
-      p -= orc + 1;
-      p[-1] = 0;
-      orcpt = string_copy(p);
-      }
+    sscanf(CS p, "%d", &pno);
     }
-  else big_buffer[nn-1] = 0;
 
-  recipients_list[n].address = string_copy(big_buffer);
-  recipients_list[n].flags = flags;
-  recipients_list[n].pno = pno;
+  /* Handle current format Exim 4 spool files */
 
-  #ifdef SUPPORT_DSN
-  /* This partial attempt at doing DSN was abandoned. This obsolete code is
-  left here just in case. Nothing is documented. */
-  recipients_list[n].orcpt = orcpt;
-  #else
-  orcpt = orcpt;   /* Keep picky compilers happy */
-  #endif
+  else if (*p == '#')
+    {
+    int flags;
+    sscanf(CS p+1, "%d", &flags);
+
+    if ((flags & 0x01) != 0)      /* one_time data exists */
+      {
+      int len;
+      while (isdigit(*(--p)) || *p == ',' || *p == '-');
+      sscanf(CS p+1, "%d,%d", &len, &pno);
+      *p = 0;
+      if (len > 0)
+        {
+        p -= len;
+        errors_to = string_copy(p);
+        }
+      }
+
+    *(--p) = 0;   /* Terminate address */
+    }
+
+  recipients_list[recipients_count].address = string_copy(big_buffer);
+  recipients_list[recipients_count].pno = pno;
+  recipients_list[recipients_count].errors_to = errors_to;
   }
 
 /* The remainder of the spool header file contains the headers for the message,
@@ -515,7 +579,7 @@ always, in order to check on the format of the file, but only create a header
 list if requested to do so. */
 
 inheader = TRUE;
-if (fgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
+if (Ufgets(big_buffer, big_buffer_size, f) == NULL) goto SPOOL_READ_ERROR;
 if (big_buffer[0] != '\n') goto SPOOL_FORMAT_ERROR;
 
 while ((n = fgetc(f)) != EOF)
@@ -527,7 +591,6 @@ while ((n = fgetc(f)) != EOF)
   if (!isdigit(n)) goto SPOOL_FORMAT_ERROR;
   ungetc(n, f);
   fscanf(f, "%d%c ", &n, flag);
-  if (!mac_isprint(flag[0])) goto SPOOL_FORMAT_ERROR;
   if (flag[0] != '*') message_size += n;  /* Omit non-transmitted headers */
 
   if (read_headers)
@@ -552,18 +615,6 @@ while ((n = fgetc(f)) != EOF)
       h->text[i] = c;
       }
     h->text[i] = 0;
-
-    /* Make the Precedence value available in a variable. This isn't
-    a standard RFC 822 header, but it is in widespread use. */
-
-    if (header_checkname(h, "Precedence", 10))
-      {
-      char *s = h->text + 10;
-      while (isspace((uschar)*s)) s++;
-      s++;
-      while (isspace((uschar)*s)) s++;
-      message_precedence = string_copy(s);
-      }
     }
 
   /* Not requiring header data, just skip through the bytes */
@@ -579,8 +630,10 @@ while ((n = fgetc(f)) != EOF)
 line count by adding the body linecount to the header linecount. Close the file
 and give a positive response. */
 
-DEBUG(5) debug_printf("body_linecount=%d message_linecount=%d\n",
+#ifndef COMPILE_UTILITY
+DEBUG(D_deliver) debug_printf("body_linecount=%d message_linecount=%d\n",
   body_linecount, message_linecount);
+#endif  /* COMPILE_UTILITY */
 
 message_linecount += body_linecount;
 
@@ -596,14 +649,22 @@ SPOOL_READ_ERROR:
 if (errno != 0)
   {
   n = errno;
-  DEBUG(1) debug_printf("Error while reading spool file %s\n", name);
+
+  #ifndef COMPILE_UTILITY
+  DEBUG(D_any) debug_printf("Error while reading spool file %s\n", name);
+  #endif  /* COMPILE_UTILITY */
+
   fclose(f);
   errno = n;
   return inheader? spool_read_hdrerror : spool_read_enverror;
   }
 
 SPOOL_FORMAT_ERROR:
-DEBUG(1) debug_printf("Format error in spool file %s\n", name);
+
+#ifndef COMPILE_UTILITY
+DEBUG(D_any) debug_printf("Format error in spool file %s\n", name);
+#endif  /* COMPILE_UTILITY */
+
 fclose(f);
 errno = ERRNO_SPOOLFORMAT;
 return inheader? spool_read_hdrerror : spool_read_enverror;
