@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Functions for writing spool files, and moving them about. */
@@ -33,9 +33,10 @@ Returns:      -1 if errmsg is not NULL; otherwise doesn't return
 */
 
 static int
-spool_write_error(int where, char **errmsg, char *s, char *temp_name, FILE *f)
+spool_write_error(int where, uschar **errmsg, uschar *s, uschar *temp_name,
+  FILE *f)
 {
-char *msg = (where == SW_RECEIVING)?
+uschar *msg = (where == SW_RECEIVING)?
   string_sprintf("spool file %s error while receiving from %s: %s", s,
     (sender_fullhost != NULL)? sender_fullhost : sender_ident,
     strerror(errno))
@@ -44,7 +45,7 @@ char *msg = (where == SW_RECEIVING)?
     (where == SW_DELIVERING)? "delivering" : "modifying",
     strerror(errno));
 
-if (temp_name != NULL) unlink(temp_name);
+if (temp_name != NULL) Uunlink(temp_name);
 if (f != NULL) fclose(f);
 
 if (errmsg == NULL)
@@ -69,9 +70,9 @@ Returns:  file descriptor of open file, or < 0 on failure, with errno unchanged
 */
 
 int
-spool_open_temp(char *temp_name)
+spool_open_temp(uschar *temp_name)
 {
-int fd = open(temp_name, O_RDWR|O_CREAT|O_EXCL, SPOOL_MODE);
+int fd = Uopen(temp_name, O_RDWR|O_CREAT|O_EXCL, SPOOL_MODE);
 
 /* If the file already exists, something has gone wrong. This process may well
 have previously created the file if it is delivering more than one address, but
@@ -82,16 +83,16 @@ have the same pid. We therefore have one go at unlinking it before giving up.
 
 if (fd < 0 && errno == EEXIST)
   {
-  DEBUG(2) debug_printf("%s exists: unlinking\n", temp_name);
-  unlink(temp_name);
-  fd = open(temp_name, O_RDWR|O_CREAT|O_EXCL, SPOOL_MODE);
+  DEBUG(D_any) debug_printf("%s exists: unlinking\n", temp_name);
+  Uunlink(temp_name);
+  fd = Uopen(temp_name, O_RDWR|O_CREAT|O_EXCL, SPOOL_MODE);
   }
 
-/* If the file has been opened, make sure the file's group is the Exim gid
-if exim_uid exists (can't have exim_uid set without exim_gid), and double-check
-the mode because the group setting doesn't always get set automatically. */
+/* If the file has been opened, make sure the file's group is the Exim gid, and
+double-check the mode because the group setting doesn't always get set
+automatically. */
 
-if (fd >= 0 && exim_uid_set)
+if (fd >= 0)
   {
   fchown(fd, exim_uid, exim_gid);
   fchmod(fd, SPOOL_MODE);
@@ -123,7 +124,7 @@ Returns:  the size of the header texts on success;
 */
 
 int
-spool_write_header(char *id, int where, char **errmsg)
+spool_write_header(uschar *id, int where, uschar **errmsg)
 {
 int fd;
 int i;
@@ -131,15 +132,15 @@ int size_correction;
 FILE *f;
 header_line *h;
 struct stat statbuf;
-char name[256];
-char temp_name[256];
+uschar name[256];
+uschar temp_name[256];
 
-sprintf(temp_name, "%s/input/%s/hdr.%d", spool_directory, message_subdir,
+sprintf(CS temp_name, "%s/input/%s/hdr.%d", spool_directory, message_subdir,
   (int)getpid());
 fd = spool_open_temp(temp_name);
-if (fd < 0) return spool_write_error(where, errmsg, "open", NULL, NULL);
-f = fdopen(fd, "w");
-DEBUG(9) debug_printf("Writing spool header file\n");
+if (fd < 0) return spool_write_error(where, errmsg, US"open", NULL, NULL);
+f = fdopen(fd, "wb");
+DEBUG(D_receive|D_deliver) debug_printf("Writing spool header file\n");
 
 /* We now have an open file to which the header data is to be written. Start
 with the file's leaf name, to make the file self-identifying. Continue with the
@@ -171,7 +172,7 @@ if (sender_host_address != NULL)
 /* Also about the interface a message came in on */
 
 if (interface_address != NULL)
-  fprintf(f, "-interface_address %s\n", interface_address);
+  fprintf(f, "-interface_address %s.%d\n", interface_address, interface_port);
 
 /* Likewise for any ident information; for local messages this is
 likely to be the same as originator_login, but will be different if
@@ -184,6 +185,15 @@ if (sender_ident != NULL) fprintf(f, "-ident %s\n", sender_ident);
 if (received_protocol != NULL)
   fprintf(f, "-received_protocol %s\n", received_protocol);
 
+/* Preserve any ACL variables that are set. Because the values may contain
+newlines, we use an explicit length. */
+
+for (i = 0; i < ACL_C_MAX + ACL_M_MAX; i++)
+  {
+  if (acl_var[i] != NULL)
+    fprintf(f, "-acl %d %d\n%s\n", i, Ustrlen(acl_var[i]), acl_var[i]);
+  }
+
 /* Now any other data that needs to be remembered. */
 
 fprintf(f, "-body_linecount %d\n", body_linecount);
@@ -198,47 +208,33 @@ if (dont_deliver) fprintf(f, "-N\n");
 if (host_lookup_failed) fprintf(f, "-host_lookup_failed\n");
 if (sender_local) fprintf(f, "-local\n");
 if (local_error_message) fprintf(f, "-localerror\n");
+if (local_scan_data != NULL) fprintf(f, "-local_scan %s\n", local_scan_data);
 if (deliver_manual_thaw) fprintf(f, "-manual_thaw\n");
-if (header_names != header_names_normal) fprintf(f, "-resent\n");
 if (sender_set_untrusted) fprintf(f, "-sender_set_untrusted\n");
 
 #ifdef SUPPORT_TLS
+if (tls_certificate_verified) fprintf(f, "-tls_certificate_verified\n");
 if (tls_cipher != NULL) fprintf(f, "-tls_cipher %s\n", tls_cipher);
 if (tls_peerdn != NULL) fprintf(f, "-tls_peerdn %s\n", tls_peerdn);
 #endif
 
-#ifdef SUPPORT_DSN
-/* This partial attempt at doing DSN was abandoned. This obsolete code is
-left here just in case. Nothing is documented. */
-if (dsn_envid != NULL) fprintf(f, "-dsn_envid %s\n", dsn_envid);
-if (dsn_ret != 0) fprintf(f, "-dsn_ret %d\n", dsn_ret);
-#endif
-
 /* To complete the envelope, write out the tree of non-recipients, followed by
 the list of recipients. These won't be disjoint the first time, when no
-checking has been done. */
+checking has been done. If a recipient is a "one-time" alias, it is followed by
+a space and its parent address number (pno). */
 
 tree_write(tree_nonrecipients, f);
 fprintf(f, "%d\n", recipients_count);
 for (i = 0; i < recipients_count; i++)
   {
   recipient_item *r = recipients_list + i;
-  if (r->flags == 0
-      #ifdef SUPPORT_DSN
-      && r->orcpt == NULL
-      #endif
-     ) fprintf(f, "%s\n", r->address);
-
+  if (r->pno < 0 && r->errors_to == NULL)
+    fprintf(f, "%s\n", r->address);
   else
     {
-    #ifdef SUPPORT_DSN
-    if (r->orcpt != NULL && r->orcpt[0] != 0)
-      fprintf(f, "%s %s %d,%d,%d\n", r->address, r->orcpt, r->flags, r->pno,
-        (int)strlen(r->orcpt));
-    else
-    #endif
-
-    fprintf(f, "%s %d,%d,0\n", r->address, r->flags, r->pno);
+    uschar *errors_to = (r->errors_to == NULL)? US"" : r->errors_to;
+    fprintf(f, "%s %s %d,%d#1\n", r->address, errors_to,
+      Ustrlen(errors_to), r->pno);
     }
   }
 
@@ -271,34 +267,63 @@ for (h = header_list; h != NULL; h = h->next)
 /* Flush and check for any errors while writing */
 
 if (fflush(f) != 0 || ferror(f))
-  return spool_write_error(where, errmsg, "write", temp_name, f);
+  return spool_write_error(where, errmsg, US"write", temp_name, f);
 
-/* Force the file's contents to be written to disc. Note that fflush()
+/* Force the file's contents to be written to disk. Note that fflush()
 just pushes it out of C, and fclose() doesn't guarantee to do the write
 either. That's just the way Unix works... */
 
 if (fsync(fileno(f)) < 0)
-  return spool_write_error(where, errmsg, "sync", temp_name, f);
+  return spool_write_error(where, errmsg, US"sync", temp_name, f);
 
 /* Get the size of the file, and close it. */
 
 fstat(fd, &statbuf);
 if (fclose(f) != 0)
-  return spool_write_error(where, errmsg, "close", temp_name, NULL);
+  return spool_write_error(where, errmsg, US"close", temp_name, NULL);
 
 /* Rename the file to its correct name, thereby replacing any previous
 incarnation. */
 
-sprintf(name, "%s/input/%s/%s-H", spool_directory, message_subdir, id);
+sprintf(CS name, "%s/input/%s/%s-H", spool_directory, message_subdir, id);
 
-if (rename(temp_name, name) < 0)
-  return spool_write_error(where, errmsg, "rename", temp_name, NULL);
+if (Urename(temp_name, name) < 0)
+  return spool_write_error(where, errmsg, US"rename", temp_name, NULL);
+
+/* Linux (and maybe other OS?) does not automatically sync a directory after
+an operation like rename. We therefore have to do it forcibly ourselves in
+these cases, to make sure the file is actually accessible on disk, as opposed
+to just the data being accessible from a file in lost+found. Linux also has
+O_DIRECTORY, for opening a directory.
+
+However, it turns out that some file systems (some versions of NFS?) do not
+support directory syncing. It seems safe enough to ignore EINVAL to cope with
+these cases. One hack on top of another... but that's life. */
+
+#ifdef NEED_SYNC_DIRECTORY
+
+sprintf(CS temp_name, "%s/input/%s/.", spool_directory, message_subdir);
+
+#ifndef O_DIRECTORY
+#define O_DIRECTORY 0
+#endif
+
+if ((fd = Uopen(temp_name, O_RDONLY|O_DIRECTORY, 0)) < 0)
+  return spool_write_error(where, errmsg, US"directory open", name, NULL);
+
+if (fsync(fd) < 0 && errno != EINVAL)
+  return spool_write_error(where, errmsg, US"directory sync", name, NULL);
+
+if (close(fd) < 0)
+  return spool_write_error(where, errmsg, US"directory close", name, NULL);
+
+#endif  /* NEED_SYNC_DIRECTORY */
 
 /* Return the number of characters in the headers, which is the file size, less
 the prelimary stuff, less the additional count fields on the headers. */
 
-DEBUG(2)
-  debug_printf("Size of headers = %d\n", statbuf.st_size - size_correction);
+DEBUG(D_receive) debug_printf("Size of headers = %d\n",
+  (int)(statbuf.st_size - size_correction));
 
 return statbuf.st_size - size_correction;
 }
@@ -328,13 +353,13 @@ Returns:     TRUE if all went well
 */
 
 static BOOL
-make_link(char *dir, char *subdir, char *id, char *suffix, char *from,
-  char *to, BOOL noentok)
+make_link(uschar *dir, uschar *subdir, uschar *id, uschar *suffix, uschar *from,
+  uschar *to, BOOL noentok)
 {
-char f[256], t[256];
-sprintf(f, "%s/%s%s/%s/%s%s", spool_directory, from, dir, subdir, id, suffix);
-sprintf(t, "%s/%s%s/%s/%s%s", spool_directory, to, dir, subdir, id, suffix);
-if (link(f, t) < 0 && (!noentok || errno != ENOENT))
+uschar f[256], t[256];
+sprintf(CS f, "%s/%s%s/%s/%s%s", spool_directory, from, dir, subdir, id, suffix);
+sprintf(CS t, "%s/%s%s/%s/%s%s", spool_directory, to, dir, subdir, id, suffix);
+if (Ulink(f, t) < 0 && (!noentok || errno != ENOENT))
   {
   log_write(0, LOG_MAIN|LOG_PANIC, "link(\"%s\", \"%s\") failed while moving "
     "message: %s", f, t, strerror(errno));
@@ -366,12 +391,12 @@ Returns:     TRUE if all went well
 */
 
 static BOOL
-break_link(char *dir, char *subdir, char *id, char *suffix, char *from,
+break_link(uschar *dir, uschar *subdir, uschar *id, uschar *suffix, uschar *from,
   BOOL noentok)
 {
-char f[256];
-sprintf(f, "%s/%s%s/%s/%s%s", spool_directory, from, dir, subdir, id, suffix);
-if (unlink(f) < 0 && (!noentok || errno != ENOENT))
+uschar f[256];
+sprintf(CS f, "%s/%s%s/%s/%s%s", spool_directory, from, dir, subdir, id, suffix);
+if (Uunlink(f) < 0 && (!noentok || errno != ENOENT))
   {
   log_write(0, LOG_MAIN|LOG_PANIC, "unlink(\"%s\") failed while moving "
     "message: %s", f, strerror(errno));
@@ -383,7 +408,7 @@ return TRUE;
 
 
 /************************************************
-*	     Move message files			*
+*            Move message files                 *
 ************************************************/
 
 /* Move the files for a message (-H, -D, and msglog) from one directory (or
@@ -402,13 +427,13 @@ Returns:      TRUE if all is well
 */
 
 BOOL
-spool_move_message(char *id, char *subdir, char *from, char *to)
+spool_move_message(uschar *id, uschar *subdir, uschar *from, uschar *to)
 {
 /* Create any output directories that do not exist. */
 
-sprintf(big_buffer, "%sinput/%s", to, subdir);
+sprintf(CS big_buffer, "%sinput/%s", to, subdir);
 (void)directory_make(spool_directory, big_buffer, INPUT_DIRECTORY_MODE, TRUE);
-sprintf(big_buffer, "%smsglog/%s", to, subdir);
+sprintf(CS big_buffer, "%smsglog/%s", to, subdir);
 (void)directory_make(spool_directory, big_buffer, INPUT_DIRECTORY_MODE, TRUE);
 
 /* Move the message by first creating new hard links for all the files, and
@@ -420,14 +445,14 @@ rule of waiting for a -H file before doing anything. When moving messsages off
 the mail spool, the -D file should be open and locked at the time, thus keeping
 Exim's hands off. */
 
-if (!make_link("msglog", subdir, id, "", from, to, TRUE) ||
-    !make_link("input", subdir, id, "-D", from, to, FALSE) ||
-    !make_link("input", subdir, id, "-H", from, to, FALSE))
+if (!make_link(US"msglog", subdir, id, US"", from, to, TRUE) ||
+    !make_link(US"input",  subdir, id, US"-D", from, to, FALSE) ||
+    !make_link(US"input",  subdir, id, US"-H", from, to, FALSE))
   return FALSE;
 
-if (!break_link("input", subdir, id, "-H", from, FALSE) ||
-    !break_link("input", subdir, id, "-D", from, FALSE) ||
-    !break_link("msglog", subdir, id, "", from, TRUE))
+if (!break_link(US"input",  subdir, id, US"-H", from, FALSE) ||
+    !break_link(US"input",  subdir, id, US"-D", from, FALSE) ||
+    !break_link(US"msglog", subdir, id, US"", from, TRUE))
   return FALSE;
 
 log_write(0, LOG_MAIN, "moved from %sinput, %smsglog to %sinput, %smsglog",
