@@ -2,47 +2,41 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Exim gets and frees all its store through these functions. In the original
-implementation there was a lot of mallocing and freeing of small bits of
-store. The philosophy has now changed to a scheme which includes two "stacking
-pools" of store. For the short-lived processes, there isn't any real need to do
-any garbage collection, but the stack concept allows quick resetting, which
-means some activities such as filtering can easily throw away the store they
-used - this reduces the total, which might get large if there were many
-adddresses.
+implementation there was a lot of mallocing and freeing of small bits of store.
+The philosophy has now changed to a scheme which includes the concept of
+"stacking pools" of store. For the short-lived processes, there isn't any real
+need to do any garbage collection, but the stack concept allows quick resetting
+in places where this seems sensible.
 
-Obviously the very long-running processes (the daemon, the queue runner, and
-eximon) must take care not to eat store.
+Obviously the long-running processes (the daemon, the queue runner, and eximon)
+must take care not to eat store.
 
-Three different kinds of store are recognized:
+The following different types of store are recognized:
 
-. Long-lived: This is implemented by retaining the original malloc/free
-  functions, and it used for permanent working buffers and for the caching of
-  lookup data.
+. Long-lived, large blocks: This is implemented by retaining the original
+  malloc/free functions, and it used for permanent working buffers and for
+  getting blocks to cut up for the other types.
 
-. Medium-term: Most of the dynamic store falls into this category. It is
-  implemented as a stacking pool (POOL_MAIN) which can be thrown away after
-  accepting a message, for example.
+. Long-lived, small blocks: This is used for blocks that have to survive until
+  the process exits. It is implemented as a stacking pool (POOL_PERM). This is
+  functionally the same as store_malloc(), except that the store can't be
+  freed, but I expect it to be more efficient for handling small blocks.
 
-. Short-term: This is another stacking pool (POOL_TEMP) which is used for
-  processing .forward (including filter) and alias files. The resulting new
-  addresses are then created in the medium pool so that the store used in this
-  processing can then be thrown away.
-
-The conversion to the new scheme happened for versions of Exim after 1.92. */
+. Short-lived, short blocks: Most of the dynamic store falls into this
+  category. It is implemented as a stacking pool (POOL_MAIN) which is reset
+  after accepting a message when multiple messages are received by a single
+  process. Resetting happens at some other times as well, usually fairly
+  locally after some specific processing that needs working store.
+*/
 
 
 #include "exim.h"
 
 
-/* Define this to turn on the availability of debugging output via the
--dm option. It could be turned off to get a little bit more performance. */
-
-#define DEBUG_STORE
-
 /* We need to know how to align blocks of data for general use. I'm not sure
 how to get an alignment factor in general. In the current world, a value of 8
 is probably right, and this is sizeof(double) on some systems and sizeof(void
@@ -57,7 +51,7 @@ appears (I checked that gcc does do this). */
 must be a multiple of the alignment. We assume that 8192 is going to be
 suitably aligned. */
 
-#define STORE_BLOCK_SIZE (8192 - alignment)
+#define STORE_BLOCK_SIZE 8192
 
 /* Structure describing the beginning of each big block. */
 
@@ -66,10 +60,19 @@ typedef struct storeblock {
   size_t length;
 } storeblock;
 
-/* Static variables holding data for the local pools of store. The current pool
-number is held in the global store_pool. Setting the initial length values to
--1 forces a malloc for the first call, even if the length is zero (which is
-used for getting a point to reset to). */
+/* Just in case we find ourselves on a system where the structure above has a
+length that is not a multiple of the alignment, set up a macro for the padded
+length. */
+
+#define ALIGNED_SIZEOF_STOREBLOCK \
+  (((sizeof(storeblock) + alignment - 1) / alignment) * alignment)
+
+/* Variables holding data for the local pools of store. The current pool number
+is held in store_pool, which is global so that it can be changed from outside.
+Setting the initial length values to -1 forces a malloc for the first call,
+even if the length is zero (which is used for getting a point to reset to). */
+
+int store_pool = POOL_PERM;
 
 static storeblock *chainbase[2] = { NULL, NULL };
 static storeblock *current_block[2] = { NULL, NULL };
@@ -84,15 +87,22 @@ pointer. */
 static int pool_malloc = 0;
 static int nonpool_malloc = 0;
 
+/* This variable is set by store_get() to its yield, and by store_reset() to
+NULL. This enables string_cat() to optimize its store handling for very long
+strings. That's why the variable is global. */
+
+void *store_last_get[2] = { NULL, NULL};
+
 
 
 /*************************************************
-*                   Get a block                  *
+*       Get a block from the current pool        *
 *************************************************/
 
 /* Running out of store is a total disaster. This function is called via the
 macro store_get(). It passes back a block of store within the current big
-block, getting a new one if necessary.
+block, getting a new one if necessary. The address is saved in
+store_last_was_get.
 
 Arguments:
   size        amount wanted
@@ -103,10 +113,8 @@ Returns:      pointer to store (panic on malloc failure)
 */
 
 void *
-store_get_3(int size, char *filename, int linenumber)
+store_get_3(int size, const char *filename, int linenumber)
 {
-void *yield;
-
 /* Round up the size to a multiple of the alignment. Although this looks a
 messy statement, because "alignment" is a constant expression, the compiler can
 do a reasonable job of optimizing, especially if the value of "alignment" is a
@@ -122,16 +130,11 @@ these functions are mostly called for small amounts of store. */
 if (size > yield_length[store_pool])
   {
   int length = (size <= STORE_BLOCK_SIZE)? STORE_BLOCK_SIZE : size;
-  int mlength = length + sizeof(storeblock);
+  int mlength = length + ALIGNED_SIZEOF_STOREBLOCK;
   storeblock *newblock;
 
-  #ifdef DEBUG_STORE
-  if (debug_trace_memory > 0)
-    {
-    pool_malloc += mlength;           /* Used in pools */
-    nonpool_malloc -= mlength;        /* Exclude from overall total */
-    }
-  #endif
+  pool_malloc += mlength;           /* Used in pools */
+  nonpool_malloc -= mlength;        /* Exclude from overall total */
 
   newblock = store_malloc(mlength);
   newblock->next = NULL;
@@ -143,30 +146,64 @@ if (size > yield_length[store_pool])
   current_block[store_pool] = newblock;
   yield_length[store_pool] = length;
   next_yield[store_pool] =
-    (void *)((char *)current_block[store_pool] + sizeof(storeblock));
+    (void *)((char *)current_block[store_pool] + ALIGNED_SIZEOF_STOREBLOCK);
   }
 
 /* There's (now) enough room in the current block; the yield is the next
 pointer. */
 
-yield = next_yield[store_pool];
+store_last_get[store_pool] = next_yield[store_pool];
 
-#ifdef DEBUG_STORE
-if (debug_trace_memory > 0)
+/* Cut out the debugging stuff for utilities, but stop picky compilers from
+giving warnings. */
+
+#ifdef COMPILE_UTILITY
+filename = filename;
+linenumber = linenumber;
+#else
+DEBUG(D_memory)
   {
-  if (debug_trace_memory == 1)
+  if (running_in_test_harness)
     debug_printf("---%d Get %5d\n", store_pool, size);
   else
-    debug_printf("---%d Get %6d %5d %-14s %4d\n", store_pool, yield, size,
-      filename, linenumber);
+    debug_printf("---%d Get %6p %5d %-14s %4d\n", store_pool,
+      store_last_get[store_pool], size, filename, linenumber);
   }
-#endif
+#endif  /* COMPILE_UTILITY */
 
 /* Update next pointer and number of bytes left in the current block. */
 
 next_yield[store_pool] = (void *)((char *)next_yield[store_pool] + size);
 yield_length[store_pool] -= size;
 
+return store_last_get[store_pool];
+}
+
+
+
+/*************************************************
+*       Get a block from the PERM pool           *
+*************************************************/
+
+/* This is just a convenience function, useful when just a single block is to
+be obtained.
+
+Arguments:
+  size        amount wanted
+  filename    source file from which called
+  linenumber  line number in source file.
+
+Returns:      pointer to store (panic on malloc failure)
+*/
+
+void *
+store_get_perm_3(int size, const char *filename, int linenumber)
+{
+void *yield;
+int old_pool = store_pool;
+store_pool = POOL_PERM;
+yield = store_get_3(size, filename, linenumber);
+store_pool = old_pool;
 return yield;
 }
 
@@ -195,7 +232,7 @@ Returns:     TRUE if the block is at the top of the stack and has been
 */
 
 BOOL
-store_extend_3(void *ptr, int oldsize, int newsize, char *filename,
+store_extend_3(void *ptr, int oldsize, int newsize, const char *filename,
   int linenumber)
 {
 int inc = newsize - oldsize;
@@ -208,16 +245,22 @@ if ((char *)ptr + rounded_oldsize != (char *)(next_yield[store_pool]) ||
     inc > yield_length[store_pool] + rounded_oldsize - oldsize)
   return FALSE;
 
-#ifdef DEBUG_STORE
-if (debug_trace_memory > 0)
+/* Cut out the debugging stuff for utilities, but stop picky compilers from
+giving warnings. */
+
+#ifdef COMPILE_UTILITY
+filename = filename;
+linenumber = linenumber;
+#else
+DEBUG(D_memory)
   {
-  if (debug_trace_memory == 1)
+  if (running_in_test_harness)
     debug_printf("---%d Ext %5d\n", store_pool, newsize);
   else
-    debug_printf("---%d Ext %6d %5d %-14s %4d\n", store_pool, ptr, newsize,
+    debug_printf("---%d Ext %6p %5d %-14s %4d\n", store_pool, ptr, newsize,
       filename, linenumber);
   }
-#endif
+#endif  /* COMPILE_UTILITY */
 
 if (newsize % alignment != 0) newsize += alignment - (newsize % alignment);
 next_yield[store_pool] = (char *)ptr + newsize;
@@ -247,13 +290,17 @@ Returns:      nothing
 */
 
 void
-store_reset_3(void *ptr, char *filename, int linenumber)
+store_reset_3(void *ptr, const char *filename, int linenumber)
 {
 storeblock *bb;
 storeblock *b = current_block[store_pool];
-char *bc = (char *)b + sizeof(storeblock);
+char *bc = (char *)b + ALIGNED_SIZEOF_STOREBLOCK;
 int newlength;
 
+/* Last store operation was not a get */
+
+store_last_get[store_pool] = NULL;
+
 /* See if the place is in the current block - as it often will be. Otherwise,
 search for the block in which it lies. */
 
@@ -261,7 +308,7 @@ if ((char *)ptr < bc || (char *)ptr > bc + b->length)
   {
   for (b = chainbase[store_pool]; b != NULL; b = b->next)
     {
-    bc = (char *)b + sizeof(storeblock);
+    bc = (char *)b + ALIGNED_SIZEOF_STOREBLOCK;
     if ((char *)ptr >= bc && (char *)ptr <= bc + b->length) break;
     }
   if (b == NULL)
@@ -269,9 +316,13 @@ if ((char *)ptr < bc || (char *)ptr > bc + b->length)
       "failed: pool=%d %-14s %4d", ptr, store_pool, filename, linenumber);
   }
 
-/* Back up, rounding to the alignment if necessary */
+/* Back up, rounding to the alignment if necessary. When testing, flatten
+the released memory. */
 
 newlength = bc + b->length - (char *)ptr;
+#ifndef COMPILE_UTILITY
+if (running_in_test_harness) memset(ptr, 0xF0, newlength);
+#endif
 yield_length[store_pool] = newlength - (newlength % alignment);
 next_yield[store_pool] = (char *)ptr + (newlength % alignment);
 current_block[store_pool] = b;
@@ -285,24 +336,26 @@ while (bb != NULL)
   {
   b = bb;
   bb = bb->next;
-
-  #ifdef DEBUG_STORE
-  if (debug_trace_memory > 0) pool_malloc -= b->length;
-  #endif
-
+  pool_malloc -= b->length + ALIGNED_SIZEOF_STOREBLOCK;
   store_free_3(b, filename, linenumber);
   }
 
-#ifdef DEBUG_STORE
-if (debug_trace_memory > 0)
+/* Cut out the debugging stuff for utilities, but stop picky compilers from
+giving warnings. */
+
+#ifdef COMPILE_UTILITY
+filename = filename;
+linenumber = linenumber;
+#else
+DEBUG(D_memory)
   {
-  if (debug_trace_memory == 1)
+  if (running_in_test_harness)
     debug_printf("---%d Rst    ** %d\n", store_pool, pool_malloc);
   else
-    debug_printf("---%d Rst %6d    ** %-14s %4d %d\n", store_pool, ptr,
+    debug_printf("---%d Rst %6p    ** %-14s %4d %d\n", store_pool, ptr,
       filename, linenumber, pool_malloc);
   }
-#endif
+#endif  /* COMPILE_UTILITY */
 }
 
 
@@ -331,7 +384,7 @@ Returns:      nothing
 */
 
 void
-store_release_3(void *block, char *filename, int linenumber)
+store_release_3(void *block, const char *filename, int linenumber)
 {
 storeblock *b;
 
@@ -340,20 +393,30 @@ storeblock *b;
 for (b = chainbase[store_pool]; b != NULL; b = b->next)
   {
   storeblock *bb = b->next;
-  if (bb != NULL && (char *)block == (char *)bb + sizeof(storeblock))
+  if (bb != NULL && (char *)block == (char *)bb + ALIGNED_SIZEOF_STOREBLOCK)
     {
     b->next = bb->next;
-    #ifdef DEBUG_STORE
-    if (debug_trace_memory > 0)
+    pool_malloc -= bb->length + ALIGNED_SIZEOF_STOREBLOCK;
+
+    /* Cut out the debugging stuff for utilities, but stop picky compilers
+    from giving warnings. */
+
+    #ifdef COMPILE_UTILITY
+    filename = filename;
+    linenumber = linenumber;
+    #else
+    DEBUG(D_memory)
       {
-      pool_malloc -= bb->length;
-      if (debug_trace_memory == 1)
+      if (running_in_test_harness)
         debug_printf("-Release       %d\n", pool_malloc);
       else
-        debug_printf("-Release %6d %-20s %4d %d\n", bb, filename,
+        debug_printf("-Release %6p %-20s %4d %d\n", (void *)bb, filename,
           linenumber, pool_malloc);
       }
-    #endif
+    if (running_in_test_harness)
+      memset(bb, 0xF0, bb->length+ALIGNED_SIZEOF_STOREBLOCK);
+    #endif  /* COMPILE_UTILITY */
+
     free(bb);
     return;
     }
@@ -380,7 +443,7 @@ Returns:      pointer to gotten store (panic on failure)
 */
 
 void *
-store_malloc_3(int size, char *filename, int linenumber)
+store_malloc_3(int size, const char *filename, int linenumber)
 {
 void *yield;
 
@@ -391,18 +454,31 @@ if (yield == NULL)
   log_write(0, LOG_MAIN|LOG_PANIC_DIE, "failed to malloc %d bytes of memory: "
     "called from line %d of %s", size, linenumber, filename);
 
-#ifdef DEBUG_STORE
-if (debug_trace_memory > 0)
+nonpool_malloc += size;
+
+/* Cut out the debugging stuff for utilities, but stop picky compilers from
+giving warnings. */
+
+#ifdef COMPILE_UTILITY
+filename = filename;
+linenumber = linenumber;
+#else
+
+/* If running in test harness, spend time making sure all the new store
+is not filled with zeros so as to catch problems. */
+
+if (running_in_test_harness)
   {
-  nonpool_malloc += size;
-  if (debug_trace_memory == 1)
-    debug_printf("--Malloc %5d %d %d\n", size, pool_malloc,
-      nonpool_malloc);
-  else
-    debug_printf("--Malloc %6d %5d %-14s %4d %d %d\n", yield, size,
-      filename, linenumber, pool_malloc, nonpool_malloc);
+  memset(yield, 0xF0, (size_t)size);
+  DEBUG(D_memory) debug_printf("--Malloc %5d %d %d\n", size, pool_malloc,
+    nonpool_malloc);
   }
-#endif
+else
+  {
+  DEBUG(D_memory) debug_printf("--Malloc %6p %5d %-14s %4d %d %d\n", yield,
+    size, filename, linenumber, pool_malloc, nonpool_malloc);
+  }
+#endif  /* COMPILE_UTILITY */
 
 return yield;
 }
@@ -423,17 +499,20 @@ Returns:      nothing
 */
 
 void
-store_free_3(void *block, char *filename, int linenumber)
+store_free_3(void *block, const char *filename, int linenumber)
 {
-#ifdef DEBUG_STORE
-if (debug_trace_memory > 0)
+#ifdef COMPILE_UTILITY
+filename = filename;
+linenumber = linenumber;
+#else
+DEBUG(D_memory)
   {
-  if (debug_trace_memory == 1)
+  if (running_in_test_harness)
     debug_printf("----Free\n");
   else
-    debug_printf("----Free %6d %-20s %4d\n", block, filename, linenumber);
+    debug_printf("----Free %6p %-20s %4d\n", block, filename, linenumber);
   }
-#endif
+#endif  /* COMPILE_UTILITY */
 free(block);
 }
 
