@@ -2,26 +2,26 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Private structure for the private options. */
 
 typedef struct {
-  char *from;
-  char *reply_to;
-  char *to;
-  char *cc;
-  char *bcc;
-  char *subject;
-  char *headers;
-  char *text;
-  char *file;
-  char *logfile;
-  char *oncelog;
+  uschar *from;
+  uschar *reply_to;
+  uschar *to;
+  uschar *cc;
+  uschar *bcc;
+  uschar *subject;
+  uschar *headers;
+  uschar *text;
+  uschar *file;
+  uschar *logfile;
+  uschar *oncelog;
+  uschar *once_repeat;
   int   mode;
   off_t once_file_size;
-  time_t once_repeat;
   BOOL  file_expand;
   BOOL  file_optional;
   BOOL  return_message;
