@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 
@@ -19,82 +19,58 @@ instance block so as to be publicly visible; these are flagged with opt_public.
 */
 
 optionlist pipe_transport_options[] = {
-  { "*expand_group",     opt_stringptr | opt_hidden | opt_public,
-      (void *)(offsetof(transport_instance, expand_gid)) },
-  { "*expand_user",      opt_stringptr | opt_hidden | opt_public,
-      (void *)(offsetof(transport_instance, expand_uid)) },
-  { "*set_group",         opt_bool | opt_hidden | opt_public,
-      (void *)(offsetof(transport_instance, gid_set)) },
-  { "*set_user",          opt_bool | opt_hidden | opt_public,
-      (void *)(offsetof(transport_instance, uid_set)) },
   { "allow_commands",    opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, allow_commands)) },
-  { "batch",             opt_local_batch | opt_public,
-      (void *)(offsetof(transport_instance, local_batch)) },
+      (void *)offsetof(pipe_transport_options_block, allow_commands) },
+  { "batch_id",          opt_stringptr | opt_public,
+      (void *)offsetof(transport_instance, batch_id) },
   { "batch_max",         opt_int | opt_public,
-      (void *)(offsetof(transport_instance, batch_max)) },
-  { "bsmtp",             opt_local_batch | opt_public,
-      (void *)(offsetof(transport_instance, local_smtp)) },
-  { "bsmtp_helo",        opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, bsmtp_helo)) },
+      (void *)offsetof(transport_instance, batch_max) },
   { "check_string",      opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, check_string)) },
+      (void *)offsetof(pipe_transport_options_block, check_string) },
   { "command",           opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, cmd)) },
-  { "current_directory", opt_stringptr | opt_public,
-      (void *)(offsetof(transport_instance, current_dir)) },
+      (void *)offsetof(pipe_transport_options_block, cmd) },
   { "environment",       opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, environment)) },
+      (void *)offsetof(pipe_transport_options_block, environment) },
   { "escape_string",     opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, escape_string)) },
+      (void *)offsetof(pipe_transport_options_block, escape_string) },
   { "freeze_exec_fail",  opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, freeze_exec_fail)) },
-  { "from_hack",         opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, from_hack)) },
-  { "group",             opt_expand_gid | opt_public,
-      (void *)(offsetof(transport_instance, gid)) },
-  { "home_directory",    opt_stringptr | opt_public,
-      (void *)(offsetof(transport_instance, home_dir)) },
+      (void *)offsetof(pipe_transport_options_block, freeze_exec_fail) },
   { "ignore_status",     opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, ignore_status)) },
-  { "initgroups",        opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, initgroups)) },
+      (void *)offsetof(pipe_transport_options_block, ignore_status) },
   { "log_defer_output",  opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, log_defer_output)) },
+      (void *)offsetof(transport_instance, log_defer_output) },
   { "log_fail_output",   opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, log_fail_output)) },
+      (void *)offsetof(transport_instance, log_fail_output) },
   { "log_output",        opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, log_output)) },
+      (void *)offsetof(transport_instance, log_output) },
   { "max_output",        opt_mkint,
-      (void *)(offsetof(pipe_transport_options_block, max_output)) },
+      (void *)offsetof(pipe_transport_options_block, max_output) },
+  { "message_prefix",    opt_stringptr,
+      (void *)offsetof(pipe_transport_options_block, message_prefix) },
+  { "message_suffix",    opt_stringptr,
+      (void *)offsetof(pipe_transport_options_block, message_suffix) },
   { "path",              opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, path)) },
+      (void *)offsetof(pipe_transport_options_block, path) },
   { "pipe_as_creator",   opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, deliver_as_creator)) },
-  { "prefix",            opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, prefix)) },
+      (void *)offsetof(transport_instance, deliver_as_creator) },
   { "restrict_to_path",  opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, restrict_to_path)) },
-  { "retry_use_local_part", opt_bool | opt_public,
-      (void *)offsetof(transport_instance, retry_use_local_part) },
+      (void *)offsetof(pipe_transport_options_block, restrict_to_path) },
   { "return_fail_output",opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, return_fail_output)) },
+      (void *)offsetof(transport_instance, return_fail_output) },
   { "return_output",     opt_bool | opt_public,
-      (void *)(offsetof(transport_instance, return_output)) },
-  { "suffix",            opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, suffix)) },
+      (void *)offsetof(transport_instance, return_output) },
   { "temp_errors",       opt_stringptr,
-      (void *)(offsetof(pipe_transport_options_block, temp_errors)) },
+      (void *)offsetof(pipe_transport_options_block, temp_errors) },
   { "timeout",           opt_time,
-      (void *)(offsetof(pipe_transport_options_block, timeout)) },
+      (void *)offsetof(pipe_transport_options_block, timeout) },
   { "umask",             opt_octint,
-      (void *)(offsetof(pipe_transport_options_block, umask)) },
+      (void *)offsetof(pipe_transport_options_block, umask) },
+  { "use_bsmtp",         opt_bool,
+      (void *)offsetof(pipe_transport_options_block, use_bsmtp) },
   { "use_crlf",          opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, use_crlf)) },
+      (void *)offsetof(pipe_transport_options_block, use_crlf) },
   { "use_shell",         opt_bool,
-      (void *)(offsetof(pipe_transport_options_block, use_shell)) },
-  { "user",              opt_expand_uid | opt_public,
-      (void *)(offsetof(transport_instance, uid)) },
+      (void *)offsetof(pipe_transport_options_block, use_shell) },
 };
 
 /* Size of the options list. An extern variable has to be used so that its
@@ -109,12 +85,11 @@ pipe_transport_options_block pipe_transport_option_defaults = {
   NULL,           /* cmd */
   NULL,           /* allow_commands */
   NULL,           /* environment */
-  "/usr/bin",     /* path */
-  "From ${if def:return_path{$return_path}{MAILER-DAEMON}} ${tod_bsdinbox}\n",
-                  /* prefix */
-  "\n",           /* suffix */
-  mac_expanded_string(EX_TEMPFAIL) ":"    /* temp_errors */
-  mac_expanded_string(EX_CANTCREAT),
+  US"/usr/bin",   /* path */
+  NULL,           /* message_prefix (reset in init if not bsmtp) */
+  NULL,           /* message_suffix (ditto) */
+  US mac_expanded_string(EX_TEMPFAIL) ":"    /* temp_errors */
+     mac_expanded_string(EX_CANTCREAT),
   NULL,           /* check_string */
   NULL,           /* escape_string */
   022,            /* umask */
@@ -122,10 +97,11 @@ pipe_transport_options_block pipe_transport_option_defaults = {
   60*60,          /* timeout */
   0,              /* options */
   FALSE,          /* freeze_exec_fail */
-  FALSE,          /* from_hack */
   FALSE,          /* ignore_status */
   FALSE,          /* restrict_to_path */
   FALSE,          /* use_shell */
+  FALSE,          /* use_bsmtp */
+  FALSE           /* use_crlf */
 };
 
 
@@ -144,10 +120,6 @@ pipe_transport_init(transport_instance *tblock)
 pipe_transport_options_block *ob =
   (pipe_transport_options_block *)(tblock->options_block);
 
-/* Retry_use_local_part defaults TRUE if unset */
-
-if (tblock->retry_use_local_part == 2) tblock->retry_use_local_part = TRUE;
-
 /* If pipe_as_creator is set, then uid/gid should not be set. */
 
 if (tblock->deliver_as_creator && (tblock->uid_set || tblock->gid_set ||
@@ -165,13 +137,13 @@ if (tblock->uid_set && !tblock->gid_set && tblock->expand_gid == NULL)
 /* Temp_errors must consist only of digits and colons, but there can be
 spaces round the colons, so allow them too. */
 
-if (ob->temp_errors != NULL)
+if (ob->temp_errors != NULL && Ustrcmp(ob->temp_errors, "*") != 0)
   {
-  size_t p = strspn(ob->temp_errors, "0123456789: ");
+  size_t p = Ustrspn(ob->temp_errors, "0123456789: ");
   if (ob->temp_errors[p] != 0)
     log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
-      "temp_errors option must be a list of numbers in the %s transport",
-        tblock->name);
+      "temp_errors must be a list of numbers or an asterisk for the %s "
+      "transport", tblock->name);
   }
 
 /* Only one of return_output/return_fail_output or log_output/log_fail_output
@@ -184,21 +156,29 @@ if (tblock->return_output && tblock->return_fail_output)
 
 if (tblock->log_output && tblock->log_fail_output)
   log_write(0, LOG_PANIC_DIE|LOG_CONFIG,
-    "both log_output and log_fail_output set for %s transport",
+    "both log_output and log_fail_output set for the %s transport",
     tblock->name);
 
-/* If batch SMTP is set, ensure the generic local batch option matches. They
-take the same set of values. Also force the check and escape strings, and
-arrange that headers are also escaped. */
+/* If batch SMTP is set, force the check and escape strings, and arrange that
+headers are also escaped. */
 
-if (tblock->local_smtp != local_smtp_off)
+if (ob->use_bsmtp)
   {
-  tblock->local_batch = tblock->local_smtp;
-  ob->check_string = ".";
-  ob->escape_string = "..";
+  ob->check_string = US".";
+  ob->escape_string = US"..";
   ob->options |= topt_escape_headers;
   }
 
+/* If not batch SMTP, and message_prefix or message_suffix are unset, insert
+default values for them. */
+
+else
+  {
+  if (ob->message_prefix == NULL) ob->message_prefix =
+    US"From ${if def:return_path{$return_path}{MAILER-DAEMON}} ${tod_bsdinbox}\n";
+  if (ob->message_suffix == NULL) ob->message_suffix = US"\n";
+  }
+
 /* The restrict_to_path  and use_shell options are incompatible */
 
 if (ob->restrict_to_path && ob->use_shell)
@@ -238,6 +218,7 @@ Arguments:
   argvptr            pointer to anchor for argv vector
   cmd                points to the command string
   expand_arguments   true if expansion is to occur
+  expand_fail        error if expansion fails
   addr               chain of addresses
   tname              the transport name
   ob                 the transport options block
@@ -247,21 +228,21 @@ Returns:             TRUE if all went well; otherwise an error will be
 */
 
 static BOOL
-set_up_direct_command(char ***argvptr, char *cmd, BOOL expand_arguments,
-  address_item *addr, char *tname, pipe_transport_options_block *ob)
+set_up_direct_command(uschar ***argvptr, uschar *cmd, BOOL expand_arguments,
+  int expand_fail, address_item *addr, uschar *tname,
+  pipe_transport_options_block *ob)
 {
 BOOL permitted = FALSE;
-char **argv;
-char buffer[64];
+uschar **argv;
+uschar buffer[64];
 
 /* Set up "transport <name>" to be put in any error messages, and then
 call the common function for creating an argument list and expanding
 the items if necessary. If it fails, this function fails (error information
 is in the addresses). */
 
-sprintf(buffer, "%.50s transport", tname);
-if (!transport_set_up_command(argvptr, cmd, expand_arguments,
-      (addr->local_part[0]== '|')? FAIL : PANIC,
+sprintf(CS buffer, "%.50s transport", tname);
+if (!transport_set_up_command(argvptr, cmd, expand_arguments, expand_fail,
       addr, buffer, NULL))
   return FALSE;
 
@@ -274,8 +255,8 @@ argv = *argvptr;
 if (ob->allow_commands != NULL)
   {
   int sep = 0;
-  char *s, *p;
-  char buffer[256];
+  uschar *s, *p;
+  uschar buffer[256];
 
   s = expand_string(ob->allow_commands);
   if (s == NULL)
@@ -288,7 +269,7 @@ if (ob->allow_commands != NULL)
 
   while ((p = string_nextinlist(&s, &sep, buffer, sizeof(buffer))) != NULL)
     {
-    if (strcmp(p, argv[0]) == 0) { permitted = TRUE; break; }
+    if (Ustrcmp(p, argv[0]) == 0) { permitted = TRUE; break; }
     }
   }
 
@@ -303,7 +284,7 @@ if (!permitted)
   {
   if (ob->restrict_to_path)
     {
-    if (strchr(argv[0], '/') != NULL)
+    if (Ustrchr(argv[0], '/') != NULL)
       {
       addr->transport_return = FAIL;
       addr->message = string_sprintf("\"/\" found in \"%s\" (command for %s "
@@ -327,15 +308,15 @@ for it. */
 if (argv[0][0] != '/')
   {
   int sep = 0;
-  char *p;
-  char *listptr = ob->path;
-  char buffer[1024];
+  uschar *p;
+  uschar *listptr = ob->path;
+  uschar buffer[1024];
 
   while ((p = string_nextinlist(&listptr, &sep, buffer, sizeof(buffer))) != NULL)
     {
     struct stat statbuf;
-    sprintf(big_buffer, "%.256s/%.256s", p, argv[0]);
-    if (stat(big_buffer, &statbuf) == 0)
+    sprintf(CS big_buffer, "%.256s/%.256s", p, argv[0]);
+    if (Ustat(big_buffer, &statbuf) == 0)
       {
       argv[0] = string_copy(big_buffer);
       break;
@@ -365,6 +346,7 @@ Arguments:
   argvptr            pointer to anchor for argv vector
   cmd                points to the command string
   expand_arguments   true if expansion is to occur
+  expand_fail        error if expansion fails
   addr               chain of addresses
   tname              the transport name
 
@@ -373,73 +355,75 @@ Returns:             TRUE if all went well; otherwise an error will be
 */
 
 static BOOL
-set_up_shell_command(char ***argvptr, char *cmd, BOOL expand_arguments,
-  address_item *addr, char *tname)
+set_up_shell_command(uschar ***argvptr, uschar *cmd, BOOL expand_arguments,
+  int expand_fail, address_item *addr, uschar *tname)
 {
-char **argv;
+uschar **argv;
 
-*argvptr = argv = store_get((4)*sizeof(char *));
+*argvptr = argv = store_get((4)*sizeof(uschar *));
 
-argv[0] = "/bin/sh";
-argv[1] = "-c";
+argv[0] = US"/bin/sh";
+argv[1] = US"-c";
 
 /* We have to take special action to handle the special "variable" called
 $pipe_addresses, which is not recognized by the normal expansion function. */
 
-DEBUG(7) debug_printf("shell pipe command before expansion:\n  %s\n", cmd);
+DEBUG(D_transport)
+  debug_printf("shell pipe command before expansion:\n  %s\n", cmd);
 
 if (expand_arguments)
   {
-  char *s = cmd;
-  char *p = strstr(cmd, "pipe_addresses");
+  uschar *s = cmd;
+  uschar *p = Ustrstr(cmd, "pipe_addresses");
 
   if (p != NULL && (
          (p > cmd && p[-1] == '$') ||
          (p > cmd + 1 && p[-2] == '$' && p[-1] == '{' && p[14] == '}')))
     {
     address_item *ad;
-    char *q = p + 14;
-    int size = (int)strlen(cmd) + 64;
+    uschar *q = p + 14;
+    int size = Ustrlen(cmd) + 64;
     int offset;
 
     if (p[-1] == '{') { q++; p--; }
 
     s = store_get(size);
     offset = p - cmd - 1;
-    strncpy(s, cmd, offset);
+    Ustrncpy(s, cmd, offset);
 
     for (ad = addr; ad != NULL; ad = ad->next)
       {
-      if (ad != addr) string_cat(s, &size, &offset, " ", 1);
-      string_cat(s, &size, &offset, ad->orig, (int)strlen(ad->orig));
+      if (ad != addr) string_cat(s, &size, &offset, US" ", 1);
+      string_cat(s, &size, &offset, ad->address, Ustrlen(ad->address));
       }
 
-    string_cat(s, &size, &offset, q, (int)strlen(q));
+    string_cat(s, &size, &offset, q, Ustrlen(q));
     s[offset] = 0;
     }
 
   /* Allow $recipients in the expansion iff it comes from a system filter */
 
   enable_dollar_recipients = addr != NULL &&
-    addr->parent != NULL && strcmp(addr->parent->orig, "message filter") == 0;
+    addr->parent != NULL &&
+    Ustrcmp(addr->parent->address, "system-filter") == 0;
   argv[2] = expand_string(s);
   enable_dollar_recipients = FALSE;
 
   if (argv[2] == NULL)
     {
-    addr->transport_return = search_find_defer? DEFER :
-      (addr->local_part[0] == '|')? FAIL : PANIC;
+    addr->transport_return = search_find_defer? DEFER : expand_fail;
     addr->message = string_sprintf("Expansion of command \"%s\" "
       "in %s transport failed: %s",
       cmd, tname, expand_string_message);
     return FALSE;
     }
 
-  DEBUG(7) debug_printf("shell pipe command after expansion:\n  %s\n", argv[2]);
+  DEBUG(D_transport)
+    debug_printf("shell pipe command after expansion:\n  %s\n", argv[2]);
   }
 else argv[2] = cmd;
 
-argv[3] = (char *)0;
+argv[3] = (uschar *)0;
 return TRUE;
 }
 
@@ -463,19 +447,19 @@ pid_t pid, outpid;
 int fd_in, fd_out, rc;
 int envcount = 0;
 int envsep = 0;
+int expand_fail;
 pipe_transport_options_block *ob =
   (pipe_transport_options_block *)(tblock->options_block);
 int timeout = ob->timeout;
 BOOL written_ok = FALSE;
 BOOL expand_arguments;
-char **argv;
-char *envp[40];
-char *envlist = ob->environment;
-char *env_local_part;
-char *cmd, *ss;
-char *eol = (ob->use_crlf)? "\r\n" : "\n";
+uschar **argv;
+uschar *envp[50];
+uschar *envlist = ob->environment;
+uschar *cmd, *ss;
+uschar *eol = (ob->use_crlf)? US"\r\n" : US"\n";
 
-DEBUG(2) debug_printf("%s transport entered\n", tblock->name);
+DEBUG(D_transport) debug_printf("%s transport entered\n", tblock->name);
 
 /* Set up for the good case */
 
@@ -483,33 +467,48 @@ addr->transport_return = OK;
 addr->basic_errno = 0;
 
 /* Pipes are not accepted as general addresses, but they can be generated from
-.forward files or alias files. In those cases, the command to be obeyed is
-pointed to by addr->local_part; it starts with the pipe symbol. In other cases,
-the command is supplied as one of the pipe transport's options. */
+.forward files or alias files. In those cases, the pfr flag is set, and the
+command to be obeyed is pointed to by addr->local_part; it starts with the pipe
+symbol. In other cases, the command is supplied as one of the pipe transport's
+options. */
 
-if (testflag(addr,af_pfr) && addr->local_part[0] == '|')
+if (testflag(addr, af_pfr) && addr->local_part[0] == '|')
   {
   cmd = addr->local_part + 1;
-  while (isspace((uschar)*cmd)) cmd++;
+  while (isspace(*cmd)) cmd++;
   expand_arguments = testflag(addr, af_expand_pipe);
+  expand_fail = FAIL;
   }
 else
   {
   cmd = ob->cmd;
   expand_arguments = TRUE;
+  expand_fail = PANIC;
   }
 
-/* When a pipe is set up by a filter file, there may be numerical variables
-in existence. These are passed in addr->pipe_expandn for use here. */
+/* If no command has been supplied, we are in trouble. */
+
+if (cmd == NULL)
+  {
+  addr->transport_return = DEFER;
+  addr->message = string_sprintf("no command specified for %s transport",
+    tblock->name);
+  return FALSE;
+  }
+
+/* When a pipe is set up by a filter file, there may be values for $thisaddress
+and numerical the variables in existence. These are passed in
+addr->pipe_expandn for use here. */
 
 if (expand_arguments && addr->pipe_expandn != NULL)
   {
-  char **ss = addr->pipe_expandn;
+  uschar **ss = addr->pipe_expandn;
   expand_nmax = -1;
+  if (*ss != NULL) filter_thisaddress = *ss++;
   while (*ss != NULL)
     {
     expand_nstring[++expand_nmax] = *ss;
-    expand_nlength[expand_nmax] = (int)strlen(*ss++);
+    expand_nlength[expand_nmax] = Ustrlen(*ss++);
     }
   }
 
@@ -520,37 +519,42 @@ there is an option to do that. */
 
 if (ob->use_shell)
   {
-  if (!set_up_shell_command(&argv, cmd, expand_arguments, addr, tblock->name))
-    return FALSE;
+  if (!set_up_shell_command(&argv, cmd, expand_arguments, expand_fail, addr,
+    tblock->name)) return FALSE;
   }
-else if (!set_up_direct_command(&argv, cmd, expand_arguments, addr,
+else if (!set_up_direct_command(&argv, cmd, expand_arguments, expand_fail, addr,
   tblock->name, ob)) return FALSE;
 
-expand_nmax = -1;   /* Reset */
+expand_nmax = -1;           /* Reset */
+filter_thisaddress = NULL;
 
 /* Set up the environment for the command. */
 
-env_local_part = (deliver_localpart == NULL)? "" : deliver_localpart;
-
-envp[envcount++] = string_sprintf("LOCAL_PART=%s", env_local_part);
-envp[envcount++] = string_sprintf("LOGNAME=%s", env_local_part);
-envp[envcount++] = string_sprintf("USER=%s", env_local_part);
-
-envp[envcount++] = string_sprintf("DOMAIN=%s", (deliver_domain == NULL)?
-  "" : deliver_domain);
-
-envp[envcount++] = string_sprintf("HOME=%s", (deliver_home == NULL)?
-  "" : deliver_home);
-
-envp[envcount++] = string_sprintf("MESSAGE_ID=%s", message_id_external);
+envp[envcount++] = string_sprintf("LOCAL_PART=%s", deliver_localpart);
+envp[envcount++] = string_sprintf("LOGNAME=%s", deliver_localpart);
+envp[envcount++] = string_sprintf("USER=%s", deliver_localpart);
+envp[envcount++] = string_sprintf("LOCAL_PART_PREFIX=%#s",
+  deliver_localpart_prefix);
+envp[envcount++] = string_sprintf("LOCAL_PART_SUFFIX=%#s",
+  deliver_localpart_suffix);
+envp[envcount++] = string_sprintf("DOMAIN=%s", deliver_domain);
+envp[envcount++] = string_sprintf("HOME=%#s", deliver_home);
+envp[envcount++] = string_sprintf("MESSAGE_ID=%s", message_id);
 envp[envcount++] = string_sprintf("PATH=%s", ob->path);
+envp[envcount++] = string_sprintf("RECIPIENT=%#s%#s%#s@%#s",
+  deliver_localpart_prefix, deliver_localpart, deliver_localpart_suffix,
+  deliver_domain);
 envp[envcount++] = string_sprintf("QUALIFY_DOMAIN=%s", qualify_domain_sender);
 envp[envcount++] = string_sprintf("SENDER=%s", sender_address);
-envp[envcount++] = "SHELL=/bin/sh";
+envp[envcount++] = US"SHELL=/bin/sh";
 
 if (addr->host_list != NULL)
   envp[envcount++] = string_sprintf("HOST=%s", addr->host_list->name);
 
+if (timestamps_utc) envp[envcount++] = US"TZ=UTC";
+else if (timezone_string != NULL && timezone_string[0] != 0)
+  envp[envcount++] = string_sprintf("TZ=%s", timezone_string);
+
 /* Add any requested items */
 
 if (envlist != NULL)
@@ -569,21 +573,25 @@ if (envlist != NULL)
 while ((ss = string_nextinlist(&envlist, &envsep, big_buffer, big_buffer_size))
        != NULL)
    {
-   if (envcount > sizeof(envp)/sizeof(char *) - 2)
-     log_write(0, LOG_MAIN|LOG_PANIC_DIE, "Too many environment settings for "
+   if (envcount > sizeof(envp)/sizeof(uschar *) - 2)
+     {
+     addr->transport_return = DEFER;
+     addr->message = string_sprintf("too many environment settings for "
        "%s transport", tblock->name);
+     return FALSE;
+     }
    envp[envcount++] = string_copy(ss);
    }
 
 envp[envcount] = NULL;
 
-
 /* If the -N option is set, can't do any more. */
 
 if (dont_deliver)
   {
-  debug_printf("*** delivery by %s transport bypassed by -N option",
-    tblock->name);
+  DEBUG(D_transport)
+    debug_printf("*** delivery by %s transport bypassed by -N option",
+      tblock->name);
   return FALSE;
   }
 
@@ -606,12 +614,10 @@ reading of the output pipe. */
 
 
 /* As this is a local transport, we are already running with the required
-uid/gid and current directory, so pass NULL to child_open to indicate no
-change. Request that the new process be a process group leader, so we
-can kill it and all its children on a timeout. */
+uid/gid and current directory. Request that the new process be a process group
+leader, so we can kill it and all its children on a timeout. */
 
-if ((pid = child_open(argv, envp, ob->umask, NULL, NULL, &fd_in, &fd_out,
-    NULL, TRUE)) < 0)
+if ((pid = child_open(argv, envp, ob->umask, &fd_in, &fd_out, TRUE)) < 0)
   {
   addr->transport_return = DEFER;
   addr->message = string_sprintf(
@@ -652,10 +658,10 @@ if (outpid == 0)
     count += rc;
     if (count > ob->max_output)
       {
-      char *message = "\n\n*** Too much output - remainder discarded ***\n";
-      DEBUG(2) debug_printf("Too much output from pipe - killed\n");
+      uschar *message = US"\n\n*** Too much output - remainder discarded ***\n";
+      DEBUG(D_transport) debug_printf("Too much output from pipe - killed\n");
       if (addr->return_file >= 0)
-        write(addr->return_file, message, (int)strlen(message));
+        write(addr->return_file, message, Ustrlen(message));
       killpg(pid, SIGKILL);
       break;
       }
@@ -672,61 +678,66 @@ to it down the pipe. It is a fallacy to think that you can detect write errors
 when the sub-process fails to read the pipe. The parent process may complete
 writing and close the pipe before the sub-process completes. We could sleep a
 bit here to let the sub-process get going, but it may still not complete. So we
-ignore all writing errors. */
+ignore all writing errors. (When in the test harness, we do do a short sleep so
+any debugging output is likely to be in the same order.) */
 
-DEBUG(2) debug_printf("Writing message to pipe\n");
+if (running_in_test_harness) millisleep(500);
+
+DEBUG(D_transport) debug_printf("Writing message to pipe\n");
 
 /* Arrange to time out writes if there is a timeout set. */
 
 if (timeout > 0)
   {
   sigalrm_seen = FALSE;
-  os_non_restarting_signal(SIGALRM, sigalrm_handler);
   transport_write_timeout = timeout;
   }
 
-/* If the local_smtp option is not unset, we need to write SMTP prefix
-information. The various different values for batching are handled outside; if
-there is more than one address available here, all must be included. Force
-SMTP dot-handling. */
+/* Reset the counter of bytes written */
 
-if (tblock->local_smtp != local_smtp_off)
+transport_count = 0;
+
+/* First write any configured prefix information */
+
+if (ob->message_prefix != NULL)
+  {
+  uschar *prefix = expand_string(ob->message_prefix);
+  if (prefix == NULL)
+    {
+    addr->transport_return = search_find_defer? DEFER : PANIC;
+    addr->message = string_sprintf("Expansion of \"%s\" (prefix for %s "
+      "transport) failed: %s", ob->message_prefix, tblock->name,
+      expand_string_message);
+    return FALSE;
+    }
+  if (!transport_write_block(fd_in, prefix, Ustrlen(prefix)))
+    goto END_WRITE;
+  }
+
+/* If the use_bsmtp option is set, we need to write SMTP prefix information.
+The various different values for batching are handled outside; if there is more
+than one address available here, all must be included. Force SMTP dot-handling.
+*/
+
+if (ob->use_bsmtp)
   {
   address_item *a;
 
-  if (tblock->bsmtp_helo &&
-      !transport_write_string(fd_in, "HELO %s%s", primary_hostname, eol))
-    goto END_WRITE;
-
   if (!transport_write_string(fd_in, "MAIL FROM:<%s>%s", return_path, eol))
     goto END_WRITE;
 
   for (a = addr; a != NULL; a = a->next)
     {
     if (!transport_write_string(fd_in,
-        "RCPT TO:<%s@%s>%s", a->local_part, a->domain, eol))
+        "RCPT TO:<%s>%s",
+        transport_rcpt_address(a, tblock->rcpt_include_affixes),
+        eol))
       goto END_WRITE;
     }
 
   if (!transport_write_string(fd_in, "DATA%s", eol)) goto END_WRITE;
   }
 
-/* Now any other configured prefix. */
-
-if (ob->prefix != NULL)
-  {
-  char *prefix = expand_string(ob->prefix);
-  if (prefix == NULL)
-    {
-    addr->transport_return = search_find_defer? DEFER : PANIC;
-    addr->message = string_sprintf("Expansion of \"%s\" (prefix for %s "
-      "transport) failed: %s", ob->prefix, tblock->name, expand_string_message);
-    return FALSE;
-    }
-  if (!transport_write_block(fd_in, prefix, (int)strlen(prefix)))
-    goto END_WRITE;
-  }
-
 /* Now the actual message - the options were set at initialization time */
 
 if (!transport_write_message(addr, fd_in, ob->options, 0, tblock->add_headers,
@@ -736,24 +747,24 @@ if (!transport_write_message(addr, fd_in, ob->options, 0, tblock->add_headers,
 
 /* Now any configured suffix */
 
-if (ob->suffix != NULL)
+if (ob->message_suffix != NULL)
   {
-  char *suffix = expand_string(ob->suffix);
+  uschar *suffix = expand_string(ob->message_suffix);
   if (suffix == NULL)
     {
     addr->transport_return = search_find_defer? DEFER : PANIC;
     addr->message = string_sprintf("Expansion of \"%s\" (suffix for %s "
-      "transport) failed: %s", ob->suffix, tblock->name, expand_string_message);
+      "transport) failed: %s", ob->message_suffix, tblock->name,
+      expand_string_message);
     return FALSE;
     }
-  if (!transport_write_block(fd_in, suffix, (int)strlen(suffix)))
+  if (!transport_write_block(fd_in, suffix, Ustrlen(suffix)))
     goto END_WRITE;
   }
 
 /* If local_smtp, write the terminating dot. */
 
-if (tblock->local_smtp != local_smtp_off &&
-    !transport_write_string(fd_in, ".%s", eol))
+if (ob->use_bsmtp && !transport_write_string(fd_in, ".%s", eol))
   goto END_WRITE;
 
 /* Flag all writing completed successfully. */
@@ -764,10 +775,8 @@ written_ok = TRUE;
 
 END_WRITE:
 
-/* OK, the writing is now all done. Turn off the alarm handler and close the
-pipe. */
+/* OK, the writing is now all done. Close the pipe. */
 
-signal(SIGALRM, SIG_IGN);
 (void) close(fd_in);
 
 /* Handle errors during writing. For timeouts, set the timeout for waiting for
@@ -860,9 +869,9 @@ if ((rc = child_close(pid, timeout)) != 0)
   with the description
 
       EX_TEMPFAIL -- temporary failure, indicating something that
-	 is not really an error.  In sendmail, this means
-	 that a mailer (e.g.) could not create a connection,
-	 and the request should be reattempted later.
+         is not really an error.  In sendmail, this means
+         that a mailer (e.g.) could not create a connection,
+         and the request should be reattempted later.
 
   Based on this, we use exit code EX_TEMPFAIL as a default to mean "defer" when
   not ignoring the returned status. However, there is now an option that
@@ -871,27 +880,31 @@ if ((rc = child_close(pid, timeout)) != 0)
   Another case that needs special treatment is if execve() failed (typically
   the command that was given is a non-existent path). By default this is
   treated as just another failure, but if freeze_exec_fail is set, the reaction
-  is to freeze the message rather than bounce the address. This failure is
-  signalled by EX_UNAVAILABLE which is definined in many systems as
+  is to freeze the message rather than bounce the address. Exim used to signal
+  this failure with EX_UNAVAILABLE, which is definined in many systems as
 
       #define EX_UNAVAILABLE  69
 
   with the description
 
       EX_UNAVAILABLE -- A service is unavailable.  This can occur
-	    if a support program or file does not exist.  This
-	    can also be used as a catchall message when something
-	    you wanted to do doesn't work, but you don't know why.
+            if a support program or file does not exist.  This
+            can also be used as a catchall message when something
+            you wanted to do doesn't work, but you don't know why.
 
-  so this is used by the code in child.c to signal execve() failure and
-  other unexpected failures such as setuid() not working - though that won't
-  be the case here because we aren't changing uid. */
+  However, this can be confused with a command that actually returns 69 because
+  something *it* wanted is unavailable. At release 4.21, Exim was changed to
+  use return code 127 instead, because this is what the shell returns when it
+  is unable to exec a command. We define it as EX_EXECFAILED, and use it in
+  child.c to signal execve() failure and other unexpected failures such as
+  setuid() not working - though that won't be the case here because we aren't
+  changing uid. */
 
   else
     {
     /* Always handle execve() failure specially if requested to */
 
-    if (ob->freeze_exec_fail && (rc == EX_UNAVAILABLE))
+    if (ob->freeze_exec_fail && (rc == EX_EXECFAILED))
       {
       addr->transport_return = DEFER;
       addr->special_action = SPECIAL_FREEZE;
@@ -903,20 +916,28 @@ if ((rc = child_close(pid, timeout)) != 0)
 
     else if (!ob->ignore_status)
       {
-      char *s = ob->temp_errors;
-      char *ss, *p;
+      uschar *ss;
       int size, ptr, i;
-      int sep = 0;
-      char buffer[64];
 
-      addr->transport_return = FAIL;
+      /* If temp_errors is "*" all codes are temporary. Initializion checks
+      that it's either "*" or a list of numbers. If not "*", scan the list of
+      temporary failure codes; if any match, the result is DEFER. */
 
-      /* Scan the list of temporary failure codes; if any match, the
-      result becomes DEFER. */
+      if (ob->temp_errors[0] == '*')
+        addr->transport_return = DEFER;
 
-      while ((p = string_nextinlist(&s, &sep, buffer, sizeof(buffer))) != NULL)
+      else
         {
-        if (rc == atoi(p)) { addr->transport_return = DEFER; break; }
+        uschar *s = ob->temp_errors;
+        uschar *p;
+        uschar buffer[64];
+        int sep = 0;
+
+        addr->transport_return = FAIL;
+        while ((p = string_nextinlist(&s,&sep,buffer,sizeof(buffer))) != NULL)
+          {
+          if (rc == Uatoi(p)) { addr->transport_return = DEFER; break; }
+          }
         }
 
       /* Ensure the message contains the expanded command and arguments. This
@@ -925,7 +946,7 @@ if ((rc = child_close(pid, timeout)) != 0)
       addr->message = string_sprintf("Child process of %s transport returned "
         "%d", tblock->name, rc);
 
-      ptr = (int)strlen(addr->message);
+      ptr = Ustrlen(addr->message);
       size = ptr + 1;
 
       /* If the return code is > 128, it often means that a shell command
@@ -934,33 +955,33 @@ if ((rc = child_close(pid, timeout)) != 0)
       ss = (rc > 128)?
         string_sprintf("(could mean shell command ended by signal %d (%s))",
           rc-128, os_strsignal(rc-128)) :
-        os_strexit(rc);
+        US os_strexit(rc);
 
       if (*ss != 0)
         {
-        addr->message = string_cat(addr->message, &size, &ptr, " ", 1);
+        addr->message = string_cat(addr->message, &size, &ptr, US" ", 1);
         addr->message = string_cat(addr->message, &size, &ptr,
-          ss, (int)strlen(ss));
+          ss, Ustrlen(ss));
         }
 
       /* Now add the command and arguments */
 
       addr->message = string_cat(addr->message, &size, &ptr,
-        " from command:", 14);
+        US" from command:", 14);
 
       for (i = 0; i < sizeof(argv)/sizeof(int *) && argv[i] != NULL; i++)
         {
         BOOL quote = FALSE;
-        addr->message = string_cat(addr->message, &size, &ptr, " ", 1);
-        if (strpbrk(argv[i], " \t") != NULL)
+        addr->message = string_cat(addr->message, &size, &ptr, US" ", 1);
+        if (Ustrpbrk(argv[i], " \t") != NULL)
           {
           quote = TRUE;
-          addr->message = string_cat(addr->message, &size, &ptr, "\"", 1);
+          addr->message = string_cat(addr->message, &size, &ptr, US"\"", 1);
           }
         addr->message = string_cat(addr->message, &size, &ptr, argv[i],
-          (int)strlen(argv[i]));
+          Ustrlen(argv[i]));
         if (quote)
-          addr->message = string_cat(addr->message, &size, &ptr, "\"", 1);
+          addr->message = string_cat(addr->message, &size, &ptr, US"\"", 1);
         }
       addr->message[ptr] = 0;  /* Ensure concatenated string terminated */
       }
@@ -972,9 +993,16 @@ are complete before we pass this point. */
 
 while (wait(&rc) >= 0);
 
-DEBUG(2) debug_printf("%s transport yielded %d\n", tblock->name,
+DEBUG(D_transport) debug_printf("%s transport yielded %d\n", tblock->name,
   addr->transport_return);
 
+/* If there has been a problem, the message in addr->message contains details
+of the pipe command. We don't want to expose these to the world, so we set up
+something bland to return to the sender. */
+
+if (addr->transport_return != OK)
+  addr->user_message = US"local delivery failed";
+
 return FALSE;
 }
 
