@@ -2,7 +2,7 @@
 *     Exim - an Internet mail transport agent    *
 *************************************************/
 
-/* Copyright (c) University of Cambridge 1995 - 2002 */
+/* Copyright (c) University of Cambridge 1995 - 2004 */
 /* See the file NOTICE for conditions of use and distribution. */
 
 /* Function for setting up the version string. */
@@ -10,7 +10,7 @@
 #include "exim.h"
 
 
-#define THIS_VERSION  "3.36"
+#define THIS_VERSION  "4.34"
 
 
 /* The header file cnumber.h contains a single line containing the
@@ -21,8 +21,11 @@ we can't use #include inside a macro argument list */
 void
 version_init(void)
 {
-char today[20];
-char *version_cnumber_format;
+static uschar cnumber_buffer[24];
+static uschar date_buffer[32];
+
+uschar today[20];
+uschar *version_cnumber_format;
 
 int cnumber =
 #include "cnumber.h"
@@ -33,22 +36,22 @@ for automatic patching to standard values when running regression tests.
 The reason that version_cnumber_format isn't just written inline in the
 sprintf() call is the gcc -Wall warns about a \0 in a format string. */
 
-version_cnumber = store_malloc(24);
-version_cnumber_format = "%d\0<<eximcnumber>>";
-sprintf(version_cnumber, version_cnumber_format, cnumber);
-version_string = THIS_VERSION "\0<<eximversion>>";
+version_cnumber = cnumber_buffer;
+version_cnumber_format = US"%d\0<<eximcnumber>>";
+sprintf(CS version_cnumber, CS version_cnumber_format, cnumber);
+version_string = US THIS_VERSION "\0<<eximversion>>";
 
-strcpy(today, __DATE__);
+Ustrcpy(today, __DATE__);
 if (today[4] == ' ') today[4] = '0';
 today[3] = today[6] = '-';
 
-version_date = (char *)store_malloc(32);
+version_date = date_buffer;
 version_date[0] = 0;
-strncat(version_date, today+4, 3);
-strncat(version_date, today, 4);
-strncat(version_date, today+7, 4);
-strcat(version_date, " ");
-strcat(version_date, __TIME__);
+Ustrncat(version_date, today+4, 3);
+Ustrncat(version_date, today, 4);
+Ustrncat(version_date, today+7, 4);
+Ustrcat(version_date, " ");
+Ustrcat(version_date, __TIME__);
 }
 
 /* End of version.c */
